! omdetect_mod
!
! Authors:  Tim Harries - MSSL (orig) 
!           almost completely rewritten by J. Mittaz MSSL (Version 3)
!           Then almost completely rewritten by Chris Brindle - MSSL (Version 4)
!          
! Description: Uses a simple algorithm to detect sources in an OM OSW image
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History
!
!    Created: Jan 1998 (TJH)
! 
!*********************************************************************************************************** 
module omdetect_mod

    use dal
    use cal
    use param  
    use sasdal
    use OmAllocation_mod         ! Functions for allocating memory to arrays.
    use OmSourceParam_mod        ! Source definitions, etc
    use OmNumerics_mod           ! Various numerical subroutines and functions
    use Parameters_mod           ! Computes source parameters
    use OmConstants_mod
    use backgroundimage_mod      ! Computes background image
    use OmdetectDetection_mod    ! Does source detection      
    use OmUtility_mod            ! For storing columns in fits files
    use OmSkyBackground_mod      ! Computes sky mask1ackground
    use OmPhotometry_mod         ! Does aperture photometry
    use OmSmoothImage_mod        ! For smoothing the image
    use getFilterID_mod
    use errorHandling
    use omSrcTable_mod           ! For creating the source-list table
    use frametime_mod, only : getFrameTime ! mssllib
    use sourceanalysis_mod
    use omsourceflags_mod        ! For setting source flags
    use stardecomp_mod
    use ominterpolate_mod
    implicit none 
 
  private
  public  omdetect


   
real(double) :: sn, bck
logical(bool), pointer   :: maskImage(:, :) => Null() ! The image-pixel quality array
logical(bool), pointer   :: emaskImage(:, :) => Null() ! The image-pixel quality array


contains 

!**************************************************************
! The main program
!**************************************************************
subroutine omdetect
  
    ! DAL variables
    type(DataSetT)       :: oswDataSet ! their associated handles 
  
    ! binning variables
    logical                 :: binBPE

    character(6)            :: omFilter   ! the OM filter
    character(8)            :: omMode     ! FAST or IMAGING mode 

    ! input image variables
  
    real(single)                       :: exposure
    real(single), pointer   :: image(:, :) => Null()
    real(single), pointer   :: tempImage(:, :) => Null()
  
    real(single), pointer   :: backgroundImage(:, :) => null()
     
    ! source searching variables
    real(single)                  :: nSigma

    integer(int16), pointer  :: levelImage(:, :) => Null()   
    integer(int32)              :: allocErr
    integer(int16)              :: nSources
   
    real(double) :: lambda
    real(double) :: plateScale
    real(single) :: apertureRadius     ! Radius (pixels) of circular aperture to be used for photometry      
    real(single)   :: sigma, modeBackground      
 
    integer(int32) :: istat
 
    integer(int32) :: windowX0 = 0     ! WindowX0
    integer(int32) :: windowY0 = 0     ! WindowY0
    integer(int32) :: windowDX = 0     ! WindowDX
    integer(int32) :: windowDY = 0     ! WindowDY

    logical(bool)  :: fastMode = .false.
  
    integer(int32) :: filterIdInt
                  
    type(sourceParameters), pointer :: sourceData(:) => null()
   
    

    ! get the task parameters
    write(errstr,'(a)') "Loading task parameters"
    call message(AppMsg, Verbose)
    
    nSigma = realParameter("nsigma")
  

    ! get handle on osw image

    oswDataSet = DataSet( trim(stringParameter("set")), modify)
    

  
    ! Get the exposure time in secs
    exposure = real32Attribute(oswDataSet, "EXPOSURE")
    omFilter = stringAttribute(oswDataSet, "FILTER")

    ! Get the OM mode (IMAGING or FAST)
    omMode = trim(stringAttribute(oswDataSet, "DATAMODE"))

  
    if (trim(omMode) .eq. "FAST") then
        fastMode = .true.
    else
        if (omFilter .eq. "Grism1" .or. omFilter .eq. "Grism2" &
                     .or. omFilter .eq. "GRISM1" .or. &
                     omFilter .eq. "GRISM2") then
         
        end if
        fastMode = .false.
       
    end if 
 
    ! Get the image
    write(errstr,'(a)') "Loading image"
    call message(AppMsg, Verbose)
    

    call GetImage(oswDataSet, nx, ny, image)

!    call smoothimagesavgol(image, 2, 0, 3)
!     call SaveImage( image, nx, ny, "smoothedimage.fits", "level image")

!call fatal("stopped at line 150 displaying smoothedimage.fits")
 

    if (nx .gt. 3000 .or. ny .gt. 3000) then
        nx = min(3000_int16, nx)
        ny = min(3000_int16, ny)

        tempImage => allocate2dSingle(0_int16, nx - 1_int16, 0_int16, ny - 1_int16, "tempimage")
        tempImage = image(0_int16 : nx - 1_int16, 0_int16 : ny - 1_int16)
        deallocate(image)
        image => tempImage
   
        write(errstr,'("Image too big! ny or ny > 3000 - reset to 3000")')  
        call message(AppMsg, Verbose)      ! Verbose was 2

    end if


    ! compute binning
    binx = int32Attribute(oswDataSet, "BINAX1")
    biny = int32Attribute(oswDataSet, "BINAX2")
    binBPE = booleanAttribute(oswDataSet, "BINBPE")


    binx = 2 ** binx
    biny = 2 ** biny

    if (binBPE) then
        binx = binx * 2
        biny = biny * 2
    end if

    if (binx .ne. biny) then
        call fatal("diffBin","Version 4 cannot deal with different binning factors in x and y")
    end if

    ! Need the frame-time for the photomtery coincidence-loss corrections
    ! NB value is in milli-seconds
    call GetCCDConstants(oswDataSet, binBPE)

    ! Get the window parameters
    call GetImageSize(oswDataSet, windowX0,  windowY0,  windowDX, windowDY)
   

    !  Get the filter Id integer value
    filterIdInt = filterIdToInt(omFilter)

    call CAL_setState(OM, filterId = filterIdInt)

 
    ! Get the Plate scale (arcsecs/pixel)
    call CAL_omGetPlateScale(lambda, plateScale)
    
 
    write(errstr,'(a)') "Running source detection algorithm"
    call message(AppMsg, Verbose)

    !******************************************************************
    ! Detect "point" sources 
    !******************************************************************
   

    if (fastMode) then
        !*************************************************************
        ! Fast mode detection
        !************************************************************
        levelImage => Allocate2Dint16(0, nx - 1, 0, ny - 1, "levelImage")
        levelImage = 0
        
  
       call FastDetection(oswDataSet, image, backgroundimage, levelImage, nSigma, exposure, nSources, sourceData)



    else
        if (omFilter .ne. "Grism1" .and. omFilter .ne. "Grism2" &
                .and. omFilter .ne. "GRISM1" .and. omFilter .ne. "GRISM2") then
            !*************************************************************
            ! Imaging mode detection
            !*************************************************************


            call ImagingDetection(oswDataSet, image, backgroundimage, levelImage, &
               nSigma, nSources, sourceData, windowX0, windowY0, exposure)



        else
            !**************************************************************
            ! Grism detection
            !**************************************************************
            call GrismDetection(image, backgroundimage, levelImage, nsigma, nSources, sourceData)
!            call GrismDetection1(image, levelImage, nsigma, nSources, sourceData)
        

      
        
        end if
    end if     
 

    write(errstr,'(a, i5)') &
             "Finished running source detection algorithm - total no of sources=", nSources

    call message(AppMsg, Verbose)
  

    !********************************************************************************
    ! In order to convert the source counts to magnitudes have to use a 12 pixel
    ! aperture radius to sum the counts
    ! subroutine AperturePhotometry conducts aperture photometry on each source
    !
    !
    ! Need the quality array
    !*********************************
    modeBackground = ComputeBackgroundMode(image, levelImage, nx, ny, 2.0, sigma)


    write(errstr,'(a, f7.1, a, f7.1)') "Image background mode=", modeBackground, &
                "+/-", sigma


    call message(AppMsg, Verbose)
   

    apertureRadius = real(12 / binx)
 

    if(stringParameter("levelimage") .ne. "")  then

        call SaveImage( levelimage, nx, ny,          &
               trim(stringParameter("levelimage")), "level image")
    end if

    ! Save it if required
    if(trim(stringParameter("backgroundimage")) .ne. "")  then
        if (associated(backgroundImage)) then
            call SaveImage(backgroundImage, nx, ny, &
                trim(stringParameter("backgroundimage")), "Background image")
        end if
    end if


    deallocate(levelImage, stat = istat)
    call CheckDeallocation(istat, "levelImage")

    !*************************************
    ! Create the region file if required
    !***********************************

    if (stringParameter("regionfile") .ne. "") &
           call CreateRegionFile1(int(nSources), sourceData, trim(stringParameter("regionfile")) )

    !*****************************************
    ! Create the source list
    !*****************************************
    nSigma = 5.0
    call CreateSourceListFile(trim(stringParameter("outset")),    &
                          oswDataSet, int(nSources),              &
                          sourceData, plateScale, apertureRadius, &
                          modeBackground, nSigma, exposure,       &
                          windowX0, windowY0, windowDX, windowDY) 




    if (associated(backgroundImage)) deallocate(backgroundImage)    
    if (associated(sourceData)) then
        deallocate (sourceData, stat = allocErr)
        call CheckDeallocation(allocErr, "sourceData")
    end if


    call release(oswDataSet)
    write(errstr,'(a)') "Exiting..."
    call message(AppMsg, Verbose)



end subroutine omdetect


!*********************************************
! Subroutine GetImageSize
!********************************************
subroutine GetImageSize(oswDataSet, windowX0,  windowY0,  windowDX, windowDY)
    type(DataSetT), intent(in) :: oswDataSet 
    integer(int32), intent(out) ::  windowX0,  windowY0,  windowDX, windowDY
    
    type(TableT) :: inTableH
    integer, parameter :: noColumns = 4
    type(ColumnT) :: inColumnH(noColumns)
    integer(int32) :: i
    integer(int32), pointer ::  windowX0P(:),  windowY0P(:),  windowDXP(:), windowDYP(:)
   
    
    if (hasAttribute(oswDataSet, "WINDOWX0") .and. &
        hasAttribute(oswDataSet, "WINDOWY0") .and. &
        hasAttribute(oswDataSet, "WINDOWDX") .and. &
        hasAttribute(oswDataSet, "WINDOWDY")) then

        windowx0 = int32Attribute(oswdataset, "WINDOWX0")
        windowy0 = int32Attribute(oswdataset, "WINDOWY0")
        windowdx = int32Attribute(oswdataset, "WINDOWDX")
        windowdY = int32Attribute(oswdataset, "WINDOWDY")
        return
    end if 

    !**************************************************
    ! Get a handle on the modes binary table extension.
    !**************************************************


    inTableH = table(oswDataSet, "MODES")

    !**************************************************
    ! Get handles and pointers on the modes binary table 
    ! extension columns and extract the window data.
    !**************************************************
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)


    !************************************
    ! Release the handles
    !***********************************
    do i = 1, noColumns
        call release(inColumnH(i))
    end do
    call release(inTableH)

end subroutine GetImageSize

!**************************************************
! CheckSourceSignificances
!**************************************************
subroutine CheckSourceSignificances(noSources, source, nx, ny, levelImage)
    integer(int16), intent(inout) :: noSources
    type(sourceParameters), pointer :: source(:)

    integer(int16), intent(in) :: nx, ny
    integer(int16), pointer :: levelImage(:, :)

    real(single) :: minSignificance
    integer(int16) :: i, num
   
    integer(int16) :: noRejects1, noRejects2
    integer(int16), pointer :: tempLevelImage(:, :)

  
    ! Check source signal-to-noise ratios
    minSignificance = realParameter("minsignificance")
    if (minSignificance .lt. 0.001) minsignificance = -100.0
 


    num = 0 ! Source number
    
    noRejects1 = 0
    noRejects2 = 0

    write(errstr,'(a)') "Checking source significances"
    call message(AppMsg, Verbose)


    tempLevelImage => Allocate2Dint16(0, nx - 1, 0, ny - 1, "tempLevelImage")
    tempLevelImage = 0


    do i = 1, noSources

        if (source(i)%significance .ge. minSignificance) then !  .and. &
            
            num = num + 1           

            if (num .ne. i) then
                source(num) = source(i)

                call ReNumberLevelImagePixels(source(i), levelImage, tempLevelImage, num)

                source(num)%number = num
            end if 
                          
        else
              

            write(errstr,'(a, i5, a, f7.2, a, f7.2, a, f7.1, a)') "Rejecting source ", i, &
                          " (x=", source(i)%xpos, " y=", source(i)%ypos, &
                  " significance=", source(i)%significance, ")"
            call message(AppMsg, Verbose)

  
            if (source(i)%number .ge. int(0, int16)) then     
               noRejects1 = noRejects1 + 1
                              
            else                            
               
              noRejects2 = noRejects2 + 1

            end if
              
        end if
        
    end do

    if ( num .ne. nosources) then
        levelImage = tempLevelImage
        noSources = num
        if (noRejects1 .gt. 0) then
            write(errstr,'(a, i6)') "No of point sources rejected=", noRejects1
            call message(AppMsg, Verbose)
        end if
        if (noRejects2 .gt. 0) then
            write(errstr,'(a, i6)') "No of extended sources rejected=", noRejects2
            call message(AppMsg, Verbose)
        end if
    end if
    deallocate(tempLevelImage)

end subroutine CheckSourceSignificances

!********************************************************************
!
!
!*******************************************************************
subroutine reNumberLevelImagePixels(source, levelImage, tempLevelImage, number)

    type(sourceParameters), intent(inout) :: source

    integer(int16), pointer :: levelImage(:, :)
    integer(int16), pointer :: tempLevelImage(:, :)

    integer(int16), intent(in) :: number
    integer(int16) :: i, j


             
    do j = source%jmin, source%jmax
        do i = source%imin, source%imax
   
            if (levelImage(i, j) .eq. source%number) then 
                tempLevelImage(i, j) = number


            end if
        end do
    end do
     
end subroutine reNumberLevelImagePixels


!***********************************************************
! GetCCDConstants
! CCD frame-time and dead-fraction
!***********************************************************
subroutine GetCCDConstants(oswDataSet, binBpe)
    type(DataSetT), intent(inout) :: oswDataSet 
    logical, intent(in) :: binBpe    
    type(DataSetT) :: wdxDataSet

    !********************************************************************
    ! Need the frame-time for the photomtery coincidence-loss corrections
    ! NB value is in milli-seconds
    !********************************************************************
    if (HasAttribute(oswDataSet, "FRAMTIME") .and. HasAttribute(oswDataSet, "DEADFRAC")) then
        frameTime = Real32Attribute(oswDataSet, "FRAMTIME") / 1000.0   ! Converted to seconds
    
        deadFrac = Real32Attribute(oswDataSet, "DEADFRAC")   

    else
        wdxDataSet = dataSet(stringParameter("wdxset"), read)


        call getFrameTime(wdxDataSet, frameTime, deadFrac, binBpe)
        !************************************************
        ! Add CCD frame-time and dead-fraction to header
        !************************************************
        call setAttribute(oswDataSet, "FRAMTIME", frameTime, "millisecs", "CCD frame time")

        call setAttribute(oswDataSet, "DEADFRAC", deadfrac, "", "CCD dead fraction")


        frameTime = frameTime / 1000.0
        call release(wdxDataSet)
    end if     

end subroutine GetCCDConstants
!*****************************************************
!
!****************************************************

subroutine FastDetection(oswDataSet, image, backgroundimage, &
          levelImage, nSigma, exposure, nSources, sourceData)
    type(DataSetT), intent(in) :: oswDataSet 
 
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    integer(int16), pointer  :: levelImage(:, :)
    real(single), intent(in) :: exposure
    real(single), intent(in) :: nSigma
    integer(int16), intent(out) :: nSources
    type(sourceParameters), pointer :: sourceData(:)

    type(SourceLinkedList) :: sourceList 
 
   
    integer(int16) :: nx, ny
    integer(int32) :: minNoPixels    

    real(single), pointer   :: backgroundSigmaImage(:, :) => Null()
   
    integer(int8) :: minCounts, extensionFlag
    integer(int16) :: nPSources
    integer(int16), pointer :: pLevelImage(:, :) => null()
    real(single) :: height
    logical(bool), pointer   :: qualArray(:, :) => Null() ! The image-pixel quality array
    

    nSources = 0    
       
    nx = size(image, 1)
    ny = size(image, 2)

    backgroundImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage") 
    backgroundSigmaImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage") 
    backgroundSigmaImage = sqrt(backgroundImage)
    qualArray => GetQualityArray(oswDataSet, nx, ny)
  
 
 

    call createBackgroundImage(image, qualArray, int(50 / binx, int16), backgroundimage, &
             backgroundsigmaimage, maxheight = 10000000.0) ! ws 10000

    ! Main point-source detection call
    call DetectPointSources(nx, ny, image,  image,     &
                            backgroundImage,           &
                            backgroundSigmaImage,      &
                            levelimage=levelImage,     &
                            nsigma=nsigma,             &
                            exposure=exposure,         & 
                            sourcelist=sourceList,     &
                            nosources=nSources)  


    !***************************************************
    ! Check to see at least one source has been detected
    !***************************************************
    if (nSources .eq.  0) then

        call InitialiseDetection(nx, ny, 0.75, 3, 2.5, 2.5, 1.5) ! was 2.0   
        call SearchForPointSources(image, image,     & 
                               backgroundImage,      &  
                               backgroundSigmaImage, &                         
                               LevelImage,           &
                               sourceList,           & 
                               nSources,             &
                               6.0,                  & ! The height above the background
                               IsPointSource,        & 
                               heightfunction2) 
                         
    end if

 

    ! Check that at least one source has been detected
    if (nSources .eq. 0) then

           
        minNoPixels  = 10 / binX      
        extensionFlag = 0

          
        
        height = nSigma
      
        minCounts = 1    

                
        title = "point-sources"
        extensionFlag = 0

        pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "pLevelImage") 
        pLevelImage = 0 
        levelImage = 0

        npSources = 0
        call DetectExtendedSources(nx, ny, image, image, &
                       backgroundImage, maxval(image), plevelImage, levelImage,    &  
                       sourceList, npSources, nSources, 6 / binX,                  &
                       5000 / binX,                                                &      
                       minCounts, height, heightFunction1, NoChecks,               &
                       extensionFlag, .false.)   


        deallocate(plevelImage)
        nullify(plevelImage)


        where(levelImage .gt. 0) levelImage = 0     
        where(levelImage .lt. 0) levelImage = abs(levelImage)    

    end if

    if (nSources .gt. 0) then
        write(errstr,'(a)') "Conducting photometry on the point sources"
        call message(AppMsg, Verbose)

 
        sourceData => SourceDataArray(sourceList, nSources)
        call DeleteSourceList(sourceList)
        apertureRadius = real(12 / binx)
        sourceData%exposure = exposure

   
        call AperturePhotometry(nx, ny, image, backgroundImage, levelImage, qualArray,  & 
             apertureRadius, nSources, sourceData)
    end if
    call delete ! Delete some arrays

 
    deallocate(qualArray)
    nullify(qualArray)



    deallocate(backgroundsigmaimage)
    nullify(backgroundsigmaimage)

 
  
end subroutine FastDetection

!*****************************************************
!
! Subroutine for imaging mode detection
!
!****************************************************
subroutine ImagingDetection(oswDataSet, image, backgroundimage, levelImage, &
         nSigma, nSources, sourceData, x0, y0, exposure)
    type(DataSetT), intent(in) :: oswDataSet
 
    real(single), pointer :: image(:, :), backgroundimage(:, :)
  
    integer(int16), pointer  :: levelImage(:, :)
    real(single), intent(in) :: nSigma
    integer(int16), intent(out) :: nSources
    type(sourceParameters), pointer :: sourceData(:)  
    integer, intent(in) :: x0, y0
    real(single), intent(in) :: exposure 
    


    !*************************************************************************
    ! Local variables
    !************************************************************************
    integer(int16) :: nx, ny
    integer(int16)              :: nPSources = 0, nESources = 0
    integer(int8) :: minCounts, extensionFlag, kMax
    integer(int16) :: dx
    real :: t1, t2
    logical :: detectExtended, detectLargeextended
   
 
    real(single), pointer   :: backgroundsigmaImage(:, :) => null() 

    integer(int16), pointer  :: pLevelImage(:, :) => Null()  ! For point sources
    integer(int16), pointer  :: eLevelImage(:, :) => Null()  ! For extended sources
    integer(int16) :: i, k    
    real(single) :: height
    real(single)   :: neSigma        
    logical(bool), pointer   :: qualArray(:, :) => Null() ! The image-pixel quality array
    type(SourceLinkedList)  :: sourceList 
    real(single) :: maxCounts 
  
    type(SourceLinkedList) :: psourceList ! For point sources 
    type(SourceLinkedList) :: esourceList ! For extended sources 
    integer(int32) :: minNoPixels



    neSigma = nSigma + 1.0 
    nPSources = 0 ! The number of point-sources
    nESources = 0 ! The number of extended-sources


    detectExtended = booleanparameter("detectextended") ! if FALSE don't look for extended sources 

         
    nx = size(image, 1) ! The number of image columns  
    ny = size(image, 2) ! The number of image rows


    call CPU_TIME(t1)             

    ! Need the quality array from the image fits file
    qualArray => GetQualityArray(oswDataSet, nx, ny)
  

    !*************************************************************************************
    ! Create the background image
    !************************************************************************************

    dx = 50 / binx
    if (dx .gt. nx) dx = max(nx, 20_int16)
    if (dx .gt. ny) dx = max(ny, 20_int16)

    call createBackgroundImage(image, qualArray, dx,        &
                    backgroundimage, backgroundsigmaimage, &
                    maxheight = 10000000.0)

 



    pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "pLevelImage") ! Point-source level image

    !*************************************************
    ! Main point-source detection call
    !************************************************
    call DetectPointSources(nx, ny, image,  image,      &
                            backgroundImage,            & 
                            backgroundsigmaimage,       &
                            levelimage=plevelImage,     &
                            nsigma=nsigma,              &
                            exposure=exposure,          &
                            sourcelist=pSourceList,     &
                            nosources=nPSources)  




    call CPU_TIME(t2)  ! Check how much cpu time it has taken           

    if (real(npSources) / real(nx * ny) .gt. 0.0005 .or. (t2 - t1) .gt. 20000.0 ) then

!        write(*, *) "omdetect1 Line 920 npspurces=", npsources, " ratio=", (real(npSources) / real(nx * ny))
!        write(*, *) "Setting detectextended to false" 


        detectExtended = .false.      
        detectLargeExtended = .false.      

        write(errstr,'(a)') "Crowded field- extended-source detection switched off"
        call message(AppMsg, Verbose)
 
    end if


    maskImage => getMaskImage() ! Need the pixel=mask images
    emaskImage => getEMaskImage()
                             

    !*****************************************************
    ! Extended-source detection
    !*****************************************************

    eLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "eLevelImage")
     

    if (detectExtended) then

        where(emaskImage) elevelImage = 1
      


        !******************************************************************
        ! Detect extended sources - number them negatively from the 
        ! number of the last source found 
        !******************************************************************
        minCounts = 3    
             
        title = "large extended-sources"      
        extensionFlag = 1

        write(errstr,'(a)') "Looking for extended sources"
        call message(AppMsg, Verbose)
  

        kMax = 2  
    
        maxCounts = maxval(image)
        do k = 0, kmax              

            height = max(1.0, 1.5 + real(k) + nsigma - 2.0) 


            ! Extended-source detection
            call DetectExtendedSources(nx, ny, image, image, backgroundImage,   &
                     maxcounts, pLevelImage, eLevelImage, eSourceList, 1_int16,   &
                     nESources, 500 / binX, 12000 / binX,            &  ! Was 500 / bimx                                
                     minCounts, height, heightFunction1, IsValidExtendedSource, &
                     extensionFlag, .false.) ! NB FRIDAY 4th MAY WAS FALSE 


            where(elevelimage .gt. 1) elevelimage = 0

        end do

        write(errstr,'(a, i6)') "Number detected so far=", nesources
        call message(AppMsg, Verbose)
  

        if (nESources .gt. 0) then 

            where(eLevelImage .lt. 0) 
                maskImage = .true.
                emaskImage = .true.
            end where

                  
            where(emaskImage .and. elevelimage .eq. 0) elevelimage = 1
 
        end if


        !*************************************************************************
        ! Mask out large scattered-light regions
        !************************************************************************
        call MaskScatteredLightRegions(image,                   &
                                   eLevelImage, plevelImage,    &
                                   real(x0), real(y0))


 

    end if ! (detectextended)


    call CPU_TIME(t2)             



    if (npSources .gt. 0) then

        where(plevelimage .ne. 0) 
            maskimage = .true.
            emaskImage = .true.
        end where
       
    end if



    where (emaskImage .and. elevelimage .eq. 0) elevelImage = 1


    !*********************************************************
    ! If any point sources have been picked up as extended transfer
    ! them to the extended source-list, etc
    !*********************************************************
    call transferSources(nPsources, pSourceList, plevelImage, &
                              nesources, eSourceList, elevelImage)



    !*************************************************************************
    ! May have overlapping extended sources that should be merged into one
    ! source
    !*************************************************************************     
    call MergeExtendedSources(nesources, eSourceList, image, &
           backgroundimage, elevelImage, real(10 / binx))

   
    where (emaskImage .and. elevelimage .eq. 0) elevelImage = 1


    !**********************************************
    ! Search for "small" extended sources
    !**********************************************
    if (detectExtended .and. (t2 - t1) .lt. 72000.0) then
     
   
        minCounts = 3  ! NB was 3    

        title = "small extended-sources"
        extensionFlag = 1            
        minNoPixels = 20 / binx
        kmax = 2 
        maxcounts = maxval(image)
        do k = 0, kmax              

           
            height = max(1.0, 1.5 + real(k) + nsigma - 2.0) 

      
            call DetectExtendedSources(nx, ny, image, image,                                   &
                       backgroundImage, maxCounts, pLevelImage, eLevelImage,                   &  
                       eSourceList, 1_int16, nESources, minNoPixels, (2000 / (binX * binx)),   &     
                       minCounts,  height, heightFunction1, IsValidSmallExtendedSource,        &
                       extensionFlag, .false.)           

            where(eLevelImage .gt. 1 .and. .not. emaskImage) eLevelImage = 0
        end do  



        write(errstr,'(a, i6)') "Number detected so far=", nesources
        call message(AppMsg, Verbose)
  


        where(eLevelImage .gt. 1) eLevelImage = 0


        where(eLevelImage .lt. 0) 
            maskImage = .true.
            emaskimage = .true.
        end where 
   

    end if  ! (detectExtended)

    where (emaskImage .and. elevelimage .eq. 0) eLevelImage = 1

    nSources = nPSources + neSources
 
  
         
    write(errstr,'(a)') "Continuing search for point-sources"
    call message(AppMsg, Verbose)


    
    call InitialiseDetection(nx, ny, 0.8, 3, 2.0, 1.5, 1.1)  ! 0.75, 3, 2.0 
    call setMinimumNoPixels(int(4 / binx, int16))


    height = max(3.0, 3.0  + nsigma - 2.0) 


    call SearchForPointSources(image, image,                  & 
                               backgroundImage,               &  
                               backgroundsigmaImage,          &
                               pLevelImage,                   &
                               pSourceList,                   & 
                               npSources,                     &
                               height,                        & ! The height above the background
                               IsPointSource2,                &   
                               heightfunction1)                  
                         

    !*************************************************************
    ! Added for G153 images
    !*************************************************************

    call InitialiseDetection(nx, ny, 1.0, 2, 2.0, 1.0, 1.1)  ! 0.75, 3, 2.0 

    height = max(4.0, 4.0  + nsigma - 2.0) 

    call SearchForPointSources(image,                         &
                               image,                         & 
                               backgroundImage,               &  
                               backgroundsigmaImage,          &
                               plevelImage,                   &
                               psourceList,                   & 
                               npSources,                     &
                               height,                        & 
                               IsPointSource22,               &  
                               HeightFunction1)             
                         


    !**********************************
    !
    !*********************************
    call InitialiseDetection(nx, ny, 0.9, 2, 2.0, 1.0, 1.1)  ! 0.75, 3, 2.0 

    height = max(5.0, 5.0  + nsigma - 2.0) 

    call SearchForPointSources(image,                         &
                               image,                         & 
                               backgroundImage,               &  
                               backgroundsigmaImage,          &
                               plevelImage,                   &
                               psourceList,                   & 
                               npSources,                     &
                               height,                        &  
                               IsPointSource22,               & 
                               HeightFunction1)             


    title = 'faint sources'
    extensionFlag = 0

    where(eLevelImage .gt. 0) elevelimage = 0
         
        
    where (emaskImage .and. elevelImage .eq. 0) 
        eLevelImage = 1
        
    endwhere 

    write(errstr,'(a)') "Looking for extended sources"
    call message(AppMsg, Verbose)
  

    extensionFlag = 1
    minCounts = 1

    maxCounts = exposure * 5.0
    height = max(2.5, 2.5  + nsigma - 2.0) 


    call DetectExtendedSources(nx, ny, image, image,                               &
                       backgroundImage, maxcounts, pLevelImage, eLevelImage,       &  
                       eSourceList, 1_int16, nESources, 8 / binx,                  &    
                       100 / (binX * binx),                                        &       
                       minCounts, height, extendedheightFunction, IsValidESource3, &  
                       extensionFlag, .false.)    ! Was heightfunction11



    where(eLevelImage .gt. 0) elevelImage = 0
    where(elevelimage .lt. 0)
        maskimage = .true.
        emaskimage = .true. 
    end where



    where(emaskImage .and. elevelimage .eq. 0) elevelimage = 1      

 
    height = max(2.5, 2.5  + nsigma - 2.0) 


    call DetectExtendedSources(nx, ny, image, image,                                 &
                       backgroundImage, maxcounts, pLevelImage, eLevelImage,         &  
                       eSourceList, 1_int16, nESources, 12 / (binx * binx),          &
                       100 / (binX * binx),                                          &      
                       minCounts, height, extendedheightFunction, IsValidESource3,   &  
                       extensionFlag, .false.)    



    where(eLevelImage .gt. 0) elevelImage = 0
    where(elevelimage .lt. 0)
        maskimage = .true.
        emaskimage = .true. 
    end where



    where(emaskImage .and. elevelimage .eq. 0) elevelimage = 1      



    height = max(3.0, 3.0  + nsigma - 2.0) 

    call DetectExtendedSources(nx, ny, image, image,                                 &
                       backgroundImage, maxcounts, pLevelImage, eLevelImage,         &  
                       eSourceList, 1_int16, nESources, 6 / binx,                    &
                       100 / (binX * binx),                                          &      
                       minCounts, height, extendedheightFunction, IsValidESource3,   & 
                       extensionFlag, .false.)    


    where(eLevelImage .gt. 1) elevelImage= 0
   
    where(elevelimage .lt. 0)
        maskimage = .true.
        emaskimage = .true. 
    end where
  
    where(emaskImage .and. elevelimage .eq. 0) elevelimage = 1      


 
    height = max(5.0, 5.0  + nsigma - 2.0) 

    call DetectExtendedSources(nx, ny, image, image,                                 &
                       backgroundImage, maxcounts, pLevelImage, eLevelImage,         &  
                       eSourceList, 1_int16, nESources, 4 / binx,                    &
                       100 / (binX * binx),                                          &     
                       minCounts, height, extendedheightFunction, IsValidESource3,   & 
                       extensionFlag, .false.)   



    write(errstr,'(a, i6)') "Number detected=", nesources
    call message(AppMsg, Verbose)
  


 
    where(elevelimage .gt. 0) elevelimage = 0      





    call MergePointSources(npsources, pSourceList, image, &
          backgroundimage, plevelImage, real(4 / binx))



    call transferSources(nPsources, pSourceList, plevelImage, &
                              nesources, eSourceList, elevelImage)




    call MergeExtendedSources(nesources, eSourceList, image, &
           backgroundimage, elevelImage, real(10 / binx))



    nSources = nPSources + neSources ! The total number of sources
    
    apertureRadius = real(12 / binx)



    if (nPSources .gt. 0) then ! Aperture photometry

        !****************************************************
        ! Load the point sources from the linked list
        !****************************************************    
        sourceData => SourceDataArray(pSourceList, npSources)
    

        !**************************************************************************
        ! Remove bumps due to point-sources from background image
        !**************************************************************************
        call removeSources(backgroundimage, plevelimage, elevelimage, sourceData, npSources)


        !****************************************************
        ! Perform aperture photometry on the point-sources
        !****************************************************

        sourceData%exposure = exposure

        write(errstr,'(a)') "Conducting photometry on the point sources"
        call message(AppMsg, Verbose)

        call AperturePhotometry(nx, ny, image, backgroundImage, plevelImage, qualArray,  & 
             apertureRadius, npSources, sourceData)


        !***********************************************************************
        ! Recompute photometry for sources with a small signal-to-noise (10.0)
        !**********************************************************************
        call AperturePhotometry(sourceData, int(npSources), image, plevelImage, qualArray, real(6 / binx), 20.0)




        !*********************************************************
        ! Add the point sources to the final soource-list
        !*********************************************************


        do i = 1, nPsources  ! Add the sources to the final list

       
            call addSourceToList(sourceList, sourceData(i))

        end do


        deallocate(sourceData)
        nullify(sourceData)
        call DeleteSourceList(pSourceList)  

    end if ! (aperturephotometry)



    !*****************************************************************
    ! Construct the final source map
    !****************************************************************
    levelImage => allocate2dint16(0, nx -1 , 0, ny - 1, "levelImage")

    levelImage = pLevelImage
    if (associated(elevelImage)) then
        where(levelImage .eq. 0 .and. elevelImage .lt. 0) levelImage = elevelImage
    end if

 


    if (nESources .gt. 0) then
 


        !*******************************************************
        ! See if any source can be split into 2 or more sources
        !******************************************************* 
        call decomposeExtendedSources(image,             &
                                      plevelimage,       &
                                      elevelImage,       &
                                      eSourceList,       &
                                      nESources,         &
                                      isValidSplitSource)





        ! Load the extended sources
        sourceData => SourceDataArray(eSourceList, neSources)


        sourceData%exposure = exposure ! The exposure length (seconds)




 !       call RefineExtendedSourceMap(neSources, sourceData, image, backgroundImage, &
 !            backgroundSigmaImage, elevelImage, 1.0)


        !***********************************************************************************
        ! Create a new background image, using the old one
        !***********************************************************************************
        if (maxval(sourceData%noSourcePixels) .gt. 100) then 

            dx = 60 / binx
            if (dx .gt. nx) dx = max(nx, 20_int16)
            if (dx .gt. ny) dx = max(ny, 20_int16)


            call createBackgroundImage(backgroundimage, levelimage, dx, maxheight = 2.0) 

        end if
    



        call ReclassifyExtendedSources(sourceData, image, backgroundImage, elevelImage)
       

  
        !****************************************************
        ! Perform aperture photometry on the extended-sources
        !****************************************************
        write(errstr,'(a)') "Conducting photometry on the extended sources"
        call message(AppMsg, Verbose)



        call AperturePhotometry(nx, ny, image, backgroundImage, elevelImage, qualArray,  & 
             apertureRadius, neSources, sourceData)



!call GaussFitPhotometry(sourceData, image, backgroundimage, elevelimage)


    
        do i = 1, nEsources ! Add the extended sources to the master source-list
            call addSourceToList(sourceList, sourceData(i))
        end do     

        deallocate(sourceData)
        nullify(sourceData)
        call DeleteSourceList(eSourceList)  ! Finished with extended source-list

    end if ! (nesources > 0)

    

    nSources = nPSources + nESources ! The total number of sources


    if (nSources .gt. 0) then 

        ! Load the combined source-list 
        sourceData => SourceDataArray(SourceList, nSources)


!        call RefineSourceMap(nSources, sourceData, image, backgroundImage, levelImage)


!********************************************
! TUESDAY TEST
!call PSFPhotometry1(image,psfParams%psfimage, levelImage, backgroundimage, sourceData)

 
!        if (booleanParameter("deconvolve")) then

!write(*, *) "calling testdeconvolution"

            
            sourceData%exposure = exposure

!            call testdeconvolution(sourceData, image, backgroundimage, frameTime, binX)

!        end if

 

        ! Check source signal-to-noise ratios 
        call CheckSourceSignificances(nSources, sourceData, nx, ny, levelImage)
     
!        call RefineSourceMap(nSources, sourceData, image, backgroundImage, levelImage)





        ! Set the quality flags
        if (associated(plevelimage) .and. associated(elevelimage)) then
            call SetSourceFlags(nSources, sourceData, x0, y0, image, backgroundimage, &
                                levelimage, qualArray, plevelimage, elevelimage)
        else 
            call SetSourceFlags(nSources, sourceData, x0, y0, image, backgroundimage, levelimage, qualArray)
        end if

    
    end if 



    call delete ! Delete the detection arrays



    if (associated(qualArray)) then
        deallocate(qualArray)
        nullify(qualArray)
    end if
          
    if (associated(pLevelImage)) then
     
        deallocate(pLevelImage)
        nullify(pLevelImage)
    end if
 
    if (associated(eLevelImage)) then
        deallocate(eLevelImage)
        nullify(eLevelImage)
    end if


 
    if (associated(backgroundSigmaImage)) then
      deallocate(backgroundSigmaImage)
        nullify(backgroundSigmaImage)
    end if


    nullify(maskImage, eMaskImage)



end subroutine ImagingDetection


!************************************************
!
!**********************************************
subroutine GrismDetection(image,  backgroundimage, &
                          levelImage, nsigma,  &
                          nSources, sourceData)
 
    real(single), pointer :: image(:, :), backgroundimage(:, :)
    integer(int16), pointer  :: levelImage(:, :)
    real(single), intent(in) :: nSigma
    integer(int16), intent(out) :: nSources
    type(sourceParameters), pointer :: sourceData(:)
 
    type(SourceLinkedList) :: sourceList 
   
    integer(int16) :: i, nx, ny

    real(single), pointer   :: tempImage(:, :) => Null()

    real(single), pointer   :: backgroundSigmaImage(:, :) => Null()
 
   
    integer(int8) :: minCounts, extensionFlag
    integer(int16) :: nSources1
    integer(int8) :: iter
    integer(int16), pointer :: pLevelImage(:, :) => null()  
 
    integer(int32) :: minNoPixels

    real(single) :: height ! , modeBackground, sigma 
       
    nx = size(image, 1)
    ny = size(image, 2)

    minCounts = 3    
    tempImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "tempImage")
    tempImage = image


    BackgroundImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage")
    
    pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "plevelImage")
    pLevelImage = 0

    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "levelImage")
    levelImage = 0
    backgroundSigmaImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage") 
  

    call createGrismBackgroundImage(image, int(40 / binX, int16), int(10 / binY, int16) , &
               backgroundImage, backgroundsigmaImage,         &
               levelimage, 10000.0)


    !***********************************************************
    !
    !
    !***********************************************************
    call setBackgroundSigmaImagePtr(backgroundsigmaImage)
    

    call SmoothGrismImage(tempImage, int(4 / binX, int16), int(12 / binY, int16))

    where(image .le. 0.0) 
        levelimage = 1_int16
    end where

                       
    extensionFlag = 1
  
    iter = 6
 

    title = "first-order grism-sources"
   
    write(errstr,'(a)') "Looking for first-order spectra"
    call message(AppMsg, Verbose)

    do i = 1, iter
 

!        write(errstr,'(a, i2)') "Iteration ", i
!        call message(AppMsg, Verbose)
   
     
        height = nsigma + real(i - 1) * 0.5

        call DetectExtendedSources(nx, ny, tempImage, image, backgroundImage, &
                     maxval(image), pLevelImage, levelImage, sourceList, 1_int16, &
                     nSources, 200 / binX, &  ! was 400
                     100000 / binX, &  !            was 50 / binx then 20 / ninx, then 6                                  
                     minCounts, height, extendedheightFunction, IsValidFirstOrderSource, &
                     extensionFlag, .false.)  

                 

        where(levelImage .gt. 1_int16) levelImage = 0    

     
    end do



    !*******************************************************
    ! See if any source can be split into 2 or more sources
    !******************************************************* 
!    do i = 1, 2, -1
!    call decomposeExtendedSources(image,             &
!                                  plevelimage,       &
!                                  levelImage,       &
!                                  sourceList,       &
!                                  nSources,         &
!                                  isValidSplitGrismSource)
!    end do


    nSources1 = nSources
    write(errstr,'(a, i6)') "Number detected =", nSources
    call message(AppMsg, Verbose)
  

    title = "zeroth-order grism-sources"

    call SmoothLevelImage3(levelImage, 2 / binx, 20 / binX)
    tempImage = image

    call SmoothGrismImage(tempImage, int(6 / binX, int16), int(8 / binY, int16))

    where(image .le. 0.0) 
        levelimage = 1_int16
    end where

 



    iter = 6
    write(errstr,'(a)') "Looking for zeroth-order spectra"
    call message(AppMsg, Verbose)



    minNoPixels = 20 / binx
    do i = 1, iter

!       write(errstr,'(a, i2)') "Iteration ", i
!       call message(AppMsg, Verbose)
   
        height = nsigma + real(i - 1) * 0.5


        call DetectExtendedSources(nx, ny, tempImage, image, backgroundImage,               &
                     maxval(image), pLevelImage, levelImage, sourceList, 1_int16, nSources, &
                     minNoPixels,                                                           &
                     5000 / binX,                                                           &                                   
                     minCounts, height, extendedheightFunction, IsValidZeroOrderSource,     &
                     extensionFlag, .false.) 

   
        where(levelImage .gt. 1_int16) levelImage = 0    


    end do

    where(levelimage .gt. 0) levelimage= 0 
    write(errstr,'(a, i6)') "Number detected =", (nsources - nSources1)
    call message(AppMsg, Verbose)









    deallocate(backgroundsigmaimage, pLevelImage, tempImage)
    nullify(backgroundsigmaimage, pLevelImage, tempImage)

    if (nsources .gt. 0) then
        sourceData => SourceDataArray(sourceList, nSources)
        call DeleteSourceList(sourceList)

    end if

  


end subroutine GrismDetection


!**********************************************
!
!**********************************************
subroutine SmoothGrismImage(image, dx, dy)
    real(single), pointer :: image(:, :)
    integer(int16), intent(in) :: dx, dy
    real(single), pointer :: tempImage(:, :) => null()
    real(single) :: fMax
    integer(int16) :: i, j, k, l, nx, ny
    
    
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    tempImage => Allocate2DSingle(0_int16, nx, 0_int16, ny, "tempImage")

    tempImage = image


    do j = 0, ny
        do i = 0, nx
            if (image(i, j) .lt. 0.0) cycle
            fMax = 0.0
            do k = max(0_int16, i - dx), min(nx, i + dx)
                do l = max(0_int16, j - dy), min(ny, j + dy)
                    
                    fMax = max(fMax, tempImage(k, l))
                end do
            end do
            image(i, j) = fMax


        end do
    end do


    deallocate(tempImage)
    nullify(tempImage)


end subroutine SmoothGrismImage

!*******************************************
!  Given the total-counts, x, the 
! background-counts, bck, the signal-to-noise,
! sn, the function returns the difference between 
! the required signa-to-noise and the computed 
! one from the counts
!*******************************************
real(double) Function SZero(x)
    real(double), intent(in) :: x

    SZero = sn - (x - bck) / sqrt(x)

end function SZero

!************************************************************
!
!
!************************************************************
subroutine MergeSourceLists(inputList, nSourcesIn, outputList)
    type(SourceLinkedList), intent(inout) :: inputList 
    integer(int16), intent(inout) :: nSourcesIn
    type(SourceLinkedList), intent(inout) :: outputList 
  
    type(sourceParameters), pointer :: sourceData(:) => null()
    integer(int16) :: i

    ! Load the combined source-list 
    sourceData => SourceDataArray(inputList, nSourcesIn)
    do i = 1, nSourcesIn
        call addSourceToList(outputList, sourceData(i))   
    end do

    deallocate(sourceData)
    nullify(sourceData)
  
end subroutine MergeSourceLists

!**********************************************************************
!
!
!**********************************************************************
subroutine ReclassifyExtendedSources(source, image, backgroundImage, levelImage)
    type(sourceParameters), pointer :: source(:)
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: levelImage
   

    integer(int16) :: i, n, npSources
    integer(int16) :: nx, ny
    real(single), parameter :: nstd = 3.0

    if (.not. associated(source)) return
    write(errstr,'(a)') "Checking classification of extended sources"
    call message(AppMsg, Verbose)
 

    n = size(source)
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    npSources = maxval(levelImage) ! The number of point sources
  
   
  

    do i = 1, n


!if (i .lt. 100) then
!    write(*, *) i, " no=", source(i)%number, " x=", source(i)%ypos, " imin=", source(i)%imin, &
!   " imax=", source(i)%imax, " jmin=", source(i)%jmin, " jmax=", source(i)%jmax, &
!   " ext=", source(i)%extensionFlag, " no pixels=", source(i)%noSourcePixels
!end if




         call ComputeExtendedSrcParameters2(int(source(i)%number, int16), nx, ny, image, backgroundImage, &
                levelImage, source(i), .true.)  ! True for Monte-carlo errors

!if (i .lt. 50) then
!    write(*, *) i, " no=", source(i)%number, " x=", source(i)%xpos, &
!    " y=", source(i)%ypos, " error=", source(i)%poserr, &
!    " a=", source(i)%semia, " +/-", source(i)%semiaerr,  " b=", source(i)%semib, " +/-", source(i)%semiaerr, &
!    " theta=", source(i)%theta, " +/-", source(i)%thetaerr 
!end if

        if ( source(i)%semia .le. 0.0) cycle

        ! Check if originally classified as an extended source
        if (source(i)%extensionFlag .eq. 1) then

 
            if ((source(i)%semia - psfparams%psfsemia) .le. (nstd * source(i)%semiaErr)) then


                source(i)%extensionFlag = 0
                source(i)%noSourcePixels = 0

                
                npSources = nPSources + 1
                
                write(errstr,'(a, i6, a, f9.2, a, f9.2, a)') "Reclassified source", &
                source(i)%number, " (", source(i)%xpos, ",", source(i)%ypos, ")"
                call message(AppMsg, Verbose)
        
            end if
        end if 

    end do
    
end subroutine reclassifyExtendedSources

!**********************************************************************
! ReclassifyPointSources
!
!
!**********************************************************************
subroutine ReclassifyPointSources(source, image, backgroundImage, &
         plevelImage, elevelImage, npSources, eSourceList, neSources)
    type(sourceParameters), pointer :: source(:)
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer, dimension(:, :) :: plevelImage, elevelImage
    type(SourceLinkedList), intent(inout) :: esourceList ! For extended sources 
 
    integer, intent(inout) :: npSources, neSources
    real(single) :: meanSemia
    real(single), pointer :: data(:) => null()
    integer(int16) :: i, j
    integer(int16) :: n, nx, ny
    integer :: i0, j0
    integer(int16) :: k, dx
   
    integer(int16) :: oldSourceNumber
    real(single), parameter :: nstd = 3.0
  
    if (.not. associated(source) .or. .not. associated(elevelimage)) return
    write(errstr,'(a)') "Checking classification of point sources"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    dx = 20 / binx



    
    !**************************************************************
    ! Compute the median semi-a values
    !*************************************************************
    n = npsources

!write(*, *) "npsources=", npsources, " nesources=", nesources
    write(*, *) "psf semia=", psfparams%psfsemia,  " psf semib=", psfparams%psfsemib 
  
    if (n .gt. 10) then
        data => allocateSingle(1_int16, n, "data")
        data = source%semia
        meanSemia = MedianValue1(data, int(n))
        write(errstr,'(a, f6.3)') "Initial median source semi-a=", meanSemia
        call message(AppMsg, Verbose)


        write(errstr,'(a, f6.3, a, f6.3)') "Minimum semi-a=", minval(data), " max=", maxval(data)
        call message(AppMsg, Verbose)
 

        deallocate(data)
        nullify(data)
    else
        meanSemia = psfsemia
    end if

  

    npsources = 0

    do i = 1, n



   

        ! Check if originally classified as a point source but now probably 
        if (source(i)%extensionFlag .eq. 1 .or. source(i)%confusionFlag .gt. 1) cycle

           
        !***********************************************************************************
        ! Recompute the source parameters     
        !***********************************************************************************
        call ComputeExtendedSrcParameters2(int(source(i)%number, int16), nx, ny, image, backgroundImage, &
                plevelImage, source(i), .true.)  ! True for Monte-carlo errors


        source(i)%semiaErr = max(0.1,  source(i)%semiaErr) ! Was 0.02
        source(i)%semibErr = max(0.1,  source(i)%semibErr)

         
        ! Check classification


        i0 = int(source(i)%xpos - 0.5)
        j0 = int(source(i)%ypos - 0.5)

       
        if ((image(i0, j0) - backgroundImage(i0, j0)) .lt. 5000.0 .and. & 
            (source(i)%semia - psfparams%psfsemia) .gt. (nstd * source(i)%semiaErr)) then


 
            source(i)%extensionFlag = 1

            write(errstr,'(a, i4, a, f9.2, a, f9.2)') "Reclassified source", source(i)%number, &
                    " x=", source(i)%xpos, " y=", source(i)%ypos
            call message(AppMsg, Verbose)
            
            oldSourceNumber = source(i)%number
            source(i)%number = -neSources 


            do j = source(i)%imin, source(i)%imax                                          
                do k = source(i)%jmin, source(i)%jmax
                    if (plevelImage(j, k) .ne. oldSourcenumber) cycle
                    pLevelImage(j, k) = 0
                    if (elevelImage(j, k) .ne. 0) cycle
                    eLevelImage(j, k) = source(i)%number
                end do
            end do

            ! Add to the extended-source list 
            call addSourceToList(esourceList, source(i))
            nesources = nesources + 1

 
        else
            ! Add to point-source array
            npSources = npSources + 1
            source(npSources) = source(i)

        end if 
        



    end do

    n = count(source%extensionFlag .eq. 0)

    if (n .gt. 10) then
        data => allocateSingle(1_int16, n, "data")
        k = 0
        do i = 1, n
            if (source(i)%extensionFlag .eq. 0) then
                k = k + 1 
                data(k) = source(i)%semia
            end if 
        end do
        meanSemia = MedianValue1(data, int(n))
        write(errstr,'(a, f6.3)') "Recomputed median source semi-a=", meanSemia
        call message(AppMsg, Verbose)

        write(errstr,'(a, f6.3, a, f6.3)') "Minimum semi-a=", minval(data), " max=", maxval(data)
        call message(AppMsg, Verbose)
        deallocate(data)
        nullify(data)
    end if 


end subroutine reclassifyPointSources

!****************************************************
! Subroutine RealImage
!****************************************************
subroutine SaveBackgroundImage(image1, image2, nx, ny, fileName)
    integer (int16), intent(in) :: nx, ny    
 
    real (single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: image1, image2
    character(*), intent(in) :: fileName
    type(dataSetT) :: imageSet     
 
    imageSet = DataSet(fileName, Create)

    
    call AddImageArray(imageset, "BACKGROUND", image1)


    call AddImageArray(imageset, "SIGMA", image2)

    call release(imageSet)



end subroutine SaveBackgroundImage

!****************************************
! Subroutine AddImageArray
!****************************************
subroutine AddImageArray(set, name, image)
    type (DataSetT), intent(in) :: set
 
    character(*), intent(in) :: name
 
    real(single), intent(in) :: image(:, :)
  
    type (ArrayT) :: arr
    real (single), dimension(:, :), pointer :: optr
 

    integer :: dim(0 : 1)
    dim(0) = size(image, 1)
    dim(1) = size(image, 2)
 

    arr = addArray(set, name, REAL32, dimensions = dim)
    optr=>real32Array2Data(arr)
 
    optr = image
   
    ! Release dataset
    call release(arr)

end subroutine AddImageArray

!*********************************************************************
! Should be called after all large extended-sources have been found
! Finds large scattered light features
!*********************************************************************
subroutine MaskScatteredLightRegions(image,             &
                                     tempLevelImagePtr, &
                                     plevelImage,       &
                                     x0, y0)
 
    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: tempLevelimagePtr, pLevelImage
    real(single), intent(in) :: x0, y0

    type(SourceLinkedList) :: tempesourceList ! For extended sources 

    integer(int32) :: minNoPixels, maxNoPixels, nList
    real(single) :: height
    
    integer(int8) :: minCounts, extensionFlag
    real(single) :: xcentre, ycentre, xc, yc
    integer(int16) :: nesources, nesources1
    integer(int16) :: i, j, i0, j0, il, iu, jl, ju, dx
    integer(int16) :: nx, ny
    real(single), pointer :: tempBackgroundImagePtr(:, :) => null()
    real(single) :: mode, sigma, maxHeight   
    real(single), pointer :: pixelList(:) => null()

    xc = size(image, 1) / real(2 * binx)  ! The x-centre of the image
    yc = size(image, 2) / real(2 * binx)  ! The y-centre of the image



    xCentre = (992.0 - x0)  / real(binX) ! The x-centre of central-background scattered-light feature
    yCentre = (1063.0 - y0) / real(binY) 

 
    if ((xcentre - xc) .lt. 10.0 .or. (xcentre - xc) .gt. size(image, 1) - 10.0) return
    if ((ycentre - yc) .lt. 10.0 .or. (ycentre - yc) .gt. size(image, 2) - 10.0) return

    nx = size(image, 1)

    ny = size(image, 2)

    i0 = int(xcentre - 0.5)
    j0 = int(ycentre - 0.5)



    dx = 300 / binx ! Was 300


    il = max(int(0, int16), i0 - dx)
    iu = min(int(nx - 1_int16, int16), i0 + dx)
    jl = max(int(0, int16), j0 - dx)
    ju = min(int(ny - 1_int16, int16), j0 + dx)
   
 
 
    !*********************************************************************
    !
    !*******************************************************************
    height = 3.0 ! was 3
    

    minnopixels = 1000 / binx
    maxNoPixels = 100000 / binx
    mincounts = 1
    nesources1 = 0
    nesources = abs(minval(templevelimageptr))
    title = " "



    where(templevelimageptr(il : iu, jl : ju) .gt. 0) templevelimageptr(il : iu, jl : ju) = 0
 
    tempBackgroundImagePtr => allocate2dSingle(0_int16, nx - 1_int16, 0_int16, ny - 1_int16, "tempBackgroundImagePtr")



    ! Compute the background mode in this central region
    pixelList => AllocateSingle(1, (iu - il  + 1) * (ju - jl + 1), "list")

    nlist = 0

    do j = jl, ju
        do i = il, iu

            if (image(i, j) .ge. 0.0) then
                nList = nList + 1
                pixelList(nlist) = image(i, j)
            end if
        end do
    end do
 
    
    mode = BackgroundMode( nList, pixelList, 0.2, 2.0, sigma )  !

    deallocate( pixelList )
    nullify( pixelList )
    
    tempbackgroundImagePtr = mode
   
    maxHeight = mode + 3.0 * sigma



    do i = 1, 2
        call DetectExtendedSources(nx, ny, il, iu, jl, ju, image, image, &
                       tempbackgroundImagePtr, maxHeight, pLevelImage, tempLevelImagePtr,     &   
                       tempeSourceList, 0_int16, nESources1, minNoPixels, maxNoPixels,        &      
                       minCounts,  height, heightFunction5, IsExtendedSourceNoChecks1,        &
                       extensionFlag, .false.)

       
 
        where(tempLevelImagePtr(il : iu, jl : ju) .gt. 0) templevelImagePtr(il : iu, jl : ju) = 0
        if (nesources1 .gt. 0) exit

        height = 2.5
    end do

    deallocate(tempBackgroundImagePtr)
    nullify(tempBackgroundImagePtr)



    if (nesources1 .gt. nesources) then

        call DeleteSourceList(tempesourcelist)



        dx = 300 / binx ! Was 300


        il = max(int(0, int16), i0 - dx)
        iu = min(int(nx - 1_int16, int16), i0 + dx)
        jl = max(int(0, int16), j0 - dx)
        ju = min(int(ny - 1_int16, int16), j0 + dx)
   

        where(tempLevelImagePtr(il : iu, jl : ju) .lt. -nesources) 
            maskImage(il : iu, jl : ju) = .true.    
            emaskImage(il : iu, jl : ju) = .true.     
            tempLevelImagePtr(il : iu, jl : ju) = 0    
 
        end where


    

    end if


    where(emaskImage(il : iu, jl : ju) .and. templevelimagePtr(il : iu, jl : ju) .eq. 0) templevelimagePtr(il : iu, jl : ju) = 1


end subroutine MaskScatteredLightRegions


!*********************************************************************
! Should be called after all large extended-sources have been found
! Finds large scattered light features
!*********************************************************************

subroutine FindSmokeRings(sourceList, npSources, x0, y0, &
                          image, plevelimage, tempLevelImagePtr)
    type(SourceLinkedList) :: sourceList 
 
    integer(int16), intent(in) :: npsources
    integer, intent(in) :: x0, y0

    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: plevelimage, tempLevelimagePtr

    type(SourceLinkedList) :: tempesourceList ! For extended sources 

    integer(int32) :: minNoPixels, maxNoPixels
    real(single) :: height
    
    integer(int8) :: minCounts, extensionFlag
 
    integer(int16) :: neSources = 0, neSources1 = 0
    integer(int16) :: i, j, i0, j0, il, iu, jl, ju, nx, ny
    real(single) ::  xSmokeRing, ySmokeRing, rRing, x, y, rStar, theta
    real(single) :: dx, dx1, xl1, yl1, xl2, yl2, xl3, yl3
    real(single), pointer, dimension(:, :) :: tempbackgroundImagePtr => null()


    type(sourceParameters), pointer :: source(:) => null()
 return

    if (npSources .lt. 1) return 
 
    write(errstr,'(a)') "Looking for smoke rings"
    call message(AppMsg, Verbose)
 



    ! Bottom-left image full-frame coordinates
    xl1 = 0.5
    yl1 = 0.5



    call ConvertCoordsToFullFrame(xl1, yl1, x0, y0, binx, biny)


    ! Bottom-right image full-frame coordinates
    xl2 = real(size(image, 1)) + 0.5
    yl2 = 0.5
    call ConvertCoordsToFullFrame(xl2, yl2, x0, y0, Binx, Biny)


    ! Upper-right image full-frame coordinates
    xl3 = real(size(image, 1)) + 0.5
    yl3 = real(size(image, 2)) + 0.5
    call ConvertCoordsToFullFrame(xl3, yl3, x0, y0, Binx, Biny)

    dx = real(40 / binx)

    !*********************************************************************
    !
    !********************************************************************
    height = 2.5

    minNoPixels = 20 / binx
    minNoPixels = minNoPixels * minNoPixels * (int(PI) + 1)
    maxNoPixels = 100 / binx
    maxNoPixels = maxNoPixels * maxNoPixels * (int(PI) + 1)


    
    mincounts = 1
    nesources1 = 0

    nx= ubound(image, 1)
    ny= ubound(image, 2)


    tempBackgroundImagePtr => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempbackgroundimagePtr")


    title = " "


    where(eMaskImage .and. tempLevelImagePtr .eq. 0) tempLevelImagePtr = 1

    !****************************************************
    ! Load the point sources from the linked list
    !****************************************************    
    source => SourceDataArray(sourceList, npSources)
    
    dx1 = 100 / binx
    ! Loop through the sources to find any that may produce a smoke-ring
    do i = 1, npsources  

write(*, *) "source ", i, " quality=", source(i)%qualityFlag, " x=", &
             source(i)%xpos, " y=", source(i)%ypos
        if (.not. btest(source(i)%qualityFlag, 5)) cycle      

        ! The source is predicted to produce a smoke-ring 
        ! Predict where smoke-ring is
        ! Radial distance of star from centre of full-frame image
        ! xCentre and yCentre are defined in omconstants_mod (1024.5. 1024.5)
        x = source(i)%xpos
        y = source(i)%ypos    
        call ConvertCoordsToFullFrame(x, y, x0, y0, Binx, Biny)


    
        rStar = RadialDistance(real(xCentre), real(yCentre), x, y)
        theta = PositionAngle(real(xCentre), real(yCentre), x, y)

 
        rRing = 1.206 * rStar

        xSmokeRing = xCentre + rRing * cos(theta)
        ySmokeRing = yCentre + rRing * sin(theta)
              

       
        if ( (xSmokeRing - xl1) .lt. dx) cycle
        if ( (ySmokeRing - yl1) .lt. dx) cycle

        if ( (xl2 - xSmokeRing) .lt. dx) cycle
        if ( (yl3 - ySmokeRing) .lt. dx) cycle


        !***********************************************************
        ! Set up detection bounds
        !***********************************************************

        call ConvertCoordsFromFullFrame(xsmokering, ysmokering, x0, y0, Binx, Biny)

        i0 = int(xSmokeRing - 0.5)
        j0 = int(ySmokeRing - 0.5)

write(*, *) "smoke ring at ", i0, ", ", j0
        il = i0 - dx1
        if (il .lt. 0) then
            il = 0
            iu = dx1   
        end if
        iu = i0 + dx1
        if (iu .gt. nx) then
            iu = nx
            il = iu - dx1   
        end if

        jl = j0 - dx1
        if (jl .lt. 0) then
            jl = 0
            ju = dx1   
        end if
        ju = j0 + dx1
        if (ju .gt. ny) then
            ju = ny
            jl = ju - dx1   
        end if
        tempBackgroundImagePtr = minval(image(il : iu, jl : ju), image(il : iu, jl : ju) .ge. 0.0)
        
write(*, *) "Looking for smoke ring based on source ", i, "(", source(i)%xpos, ", ", source(i)%ypos, ")"
write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju

        do j = 1, 3
            height = real(j)
            call DetectExtendedSources(nx + 1_int16, ny + 1_int16, il, iu, jl, ju, image, image, &
                       tempbackgroundImagePtr, maxval(image), pLevelImage, tempLevelImagePtr,    &   
                       tempeSourceList, 1_int16, nESources1, minNoPixels, maxNoPixels,      &  ! was 20 / binx    
                       minCounts,  height, heightFunction5, IsSmokeRing,        &
                       extensionFlag, .true.)
            if (neSources1 .lt. neSources) exit
        end do 

if(nesources1 .lt. nesources) write(*, *) "no of sources found=", (nesources - nesources1)



        where(tempLevelImagePtr(il : iu, jl : ju) .gt. 1) templevelImagePtr(il : iu, jl : ju) = 0




   
    end do 
  
    deallocate(source)
    nullify(source)
  
 
 
    where(tempLevelImagePtr .lt. nesources) 
        maskImage = .true.    
        emaskImage = .true.     
        templevelimagePtr = 1
    end where


     deallocate(tempBackgroundimagePtr)
     nullify(tempBackgroundimagePtr)

! call saveImage(templevelimagePtr, nx, ny, "templevelimage.fits", "background image")
!call fatal("stopped line 3206 displaying templevelimage.fits")



    if (nesources1 .lt. nesources) call DeleteSourceList(tempesourcelist)



end subroutine FindSmokeRings


!*********************************************************
! If any sources have been picked up as extended in 
! point-source list, transfer
! them to the extended source-list, etc
!*********************************************************
subroutine TransferSources(nPsources, pSourceList, &
                           plevelImage, nesources, eSourceList, elevelImage)

    ! Point-source variables
    integer(int16), intent(inout) :: nPsources
    type(SourceLinkedList), intent(inout) :: psourceList 
     integer(int16), pointer :: plevelImage(:, :)

    ! Extended-source variables
    integer(int16), intent(inout) :: nesources
    type(SourceLinkedList), intent(inout) :: esourceList  
    integer(int16), pointer :: elevelImage(:, :)
    integer(int16), pointer :: templevelImage(:, :) => null()
 
    integer :: nx, ny, n

    type(sourceParameters), pointer :: source(:) => null()
    integer :: i

 
    if (npSources .eq. 0) return

    ! Load the point-sources
    source => SourceDataArray(pSourceList, npSources)
    


    if (count(source%extensionFlag .eq. 0) .eq. 0) then
        deallocate(source)
        nullify(source)
        return
    end if

    call DeleteSourceList(psourcelist)

    nx = ubound(plevelimage, 1)
    ny = ubound(plevelimage, 2)


    tempLevelImage => Allocate2DInt16(0, nx, 0, ny, "tempeLevelImage")
    n = npsources
    nPSources = 0

    !*******************************************************************************
    ! Loop through all the sources in the point-source list and
    ! transfer them to the extended source-list if the extension flag is 1
    !******************************************************************************
    do i = 1, n

        if (source(i)%extensionFlag .eq. 0) then ! Add to new point-source linked-list
          
            nPsources = npSources + 1
            
            templevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = nPSources
            source(i)%number = npSources
            call addSourceToList(pSourceList, source(i))

        else ! Add to extended-source linked-list
            neSources = neSources + 1 
            where(plevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                  elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = -neSources
   
                  plevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = 0
            end where
   
            source(i)%number = -neSources
            source(i)%noSourcePixels = count(elevelImage(source(i)%imin : source(i)%imax, &
                       source(i)%jmin : source(i)%jmax) .eq. source(i)%number)
   
            call addSourceToList(eSourceList, source(i))   
 
        end if

    end do
    plevelimage = templevelimage


    if (n .ne. npsources) then
        write(errstr,'(a, i6, a, i6)') "New number of likely point-sources=", npsources, " likely extended-sources=", nesources
        call message(AppMsg, Verbose)
    end if

   

    deallocate(templevelimage, source)
    nullify(templevelimage, source)
  


end subroutine transferSources

!*********************************************************
! If any 2 extended sources overlap, merge them
! into a single source
!*********************************************************
subroutine MergeExtendedSources(nesources, eSourceList, image, &
          backgroundimage, elevelImage, tolerance)

    ! Extended-source variables

    integer(int16), intent(inout) :: nesources
    type(SourceLinkedList), intent(inout) :: esourceList  

    real(single), pointer, dimension(:, :) :: image, backgroundimage 
 
    integer(int16), pointer :: elevelImage(:, :)
    real(single), intent(in) :: tolerance
    integer(int16), pointer :: templevelImage(:, :) => null()
    real(single) :: dx, dy, radius, error, error1, error2
    real(single) :: xStart, yStart
    integer(int16) :: nx, ny

    type(sourceParameters), pointer :: source(:) => null()
   

    integer :: i, j, k, n
    logical, pointer :: merged(:) => null()
    logical :: mergedP    



    if (neSources .lt. 2) return

    mergedP = .false.
    nx = ubound(elevelimage, 1)
    ny = ubound(elevelimage, 2)

    tempLevelImage => Allocate2DInt16(0_int16, nx, 0_int16, ny, "tempeLevelImage")
    allocate(merged(1 : nesources))
  

    do k = 1, 3 

        where(elevelimage .gt. 0_int16) 
            templevelimage = elevelimage
        elsewhere
            tempLevelImage = 0_int16
        end where


        n = neSources
        neSources = 0
   
        ! Load the extended-sources
        source => SourceDataArray(eSourceList, n)
        call DeleteSourceList(esourcelist)  ! Delete source-list and start a new one
        merged = .false.


        do i = 1, n
            if (merged(i)) cycle
            neSources = neSources + 1

   
            where(elevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                    templevelImage(source(i)%imin : source(i)%imax, source(i)%jmin : source(i)%jmax) = -neSources
                                              
            end where
        


            error1 = source(i)%poserr

            !*****************************************************************************
            ! Check to see if source near an edge
            !***************************************************************************** 
            if (source(i)%imin .eq. 0 .or. source(i)%jmin .eq. 0 .or.& 
                source(i)%imax .eq. nx .or. source(i)%jmax .eq. ny) error1 = max(real(10 / binx), error1) ! was 2 / 


      jLoop :      do j = i + 1, n


                if (i .eq. j .or. merged(j)) cycle
                error2 = source(j)%poserr
                if (source(j)%imin .eq. 0 .or. source(j)%jmin .eq. 0 .or.& 
                    source(j)%imax .eq. nx .or. source(j)%jmax .eq. ny) error2 = max(real(10 / binx), error2) ! was 2 / 

               




                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)
                error = 2.0 * sqrt(error1 * error1 + error2 * error2)
                error = max(error, tolerance)


                if (error .gt. radius) then ! Merge sources

                    if (radius .gt. real(6 / binx)) then
                        dx = (abs(dx) - real(8 / binx)) / 9.0
                        dy = (abs(dy) - real(8 / binx)) / 9.0
                      
                        if (source(j)%xpos .gt. source(i)%xpos) then
                            xStart = source(i)%xpos + real(4 / binx) - 0.5
                        else
                            xStart = source(j)%xpos + real(4 / binx) - 0.5
                        end if
                        if (source(j)%ypos .gt. source(i)%ypos) then
                            yStart = source(i)%ypos + real(4 / binx) - 0.5
                        else
                            yStart = source(j)%ypos + real(4 / binx) - 0.5
                        end if

                            
                    end if

                    merged(i) = .true.
                    merged(j) = .true.
                    write(errstr,'(a, i6, a, i6)') "Merging extended sources ", i, " and ", j
                    call message(AppMsg, Verbose)
                    
                 
                
                    where(elevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax) .eq. source(j)%number) 
                      
                        templevelImage(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax) = -neSources
                         
                    end where
                         
                    source(i)%number = -neSources
                    source(i)%imin = min(source(i)%imin, source(j)%imin)
                    source(i)%imax = max(source(i)%imax, source(j)%imax)
                    source(i)%jmin = min(source(i)%jmin, source(j)%jmin)
                    source(i)%jmax = max(source(i)%jmax, source(j)%jmax)
                    source(i)%noSourcePixels = source(i)%noSourcePixels + source(j)%noSourcePixels 

                end if        
            end do jLoop
  
            if (merged(i)) then
                !*****************************************************************
                ! Compute extended-source parameters
                !****************************************************************
                call ComputeExtendedSrcParameters2(int(source(i)%number, int16), &
                          nx, ny, image, backgroundImage, &
                        templevelImage, source(i), .true.)  ! True for Monte-carlo errors
                mergedP = .true.


            else

                source(i)%number = -nesources 


            
            end if
            ! Add source to list
              
            call addSourceToList(eSourceList, source(i))   
           
        end do
   
     
        deallocate(source)
        nullify(source)


        if (n .eq. nesources) exit

        elevelimage = templevelimage
 

    end do

    if (mergedP) then

        write(errstr,'(a, i6)') "New number of likely extended-sources=", nesources
        call message(AppMsg, Verbose)
    end if


    deallocate(merged, tempLevelImage)
    nullify(merged, tempLevelImage)
  
end subroutine mergeextendedSources



!*********************************************************
! If any 2 extended sources overlap, merge them
! into a single source
!*********************************************************
subroutine MergePointSources(npsources, pSourceList, image, &
          backgroundimage, plevelImage, tolerance)

    ! Extended-source variables

    integer(int16), intent(inout) :: npsources
    type(SourceLinkedList), intent(inout) :: psourceList  

    real(single), pointer, dimension(:, :) :: image, backgroundimage 
 
    integer(int16), pointer :: plevelImage(:, :)
    real(single), intent(in) :: tolerance
    integer(int16), pointer :: templevelImage(:, :) => null()
    real(single) :: dx, dy, radius, error, error1, error2
   
    integer(int16) :: nx, ny

    type(sourceParameters), pointer :: source(:) => null()
    

    integer :: i, j, k, n
    logical, pointer :: merged(:) => null()
    logical :: mergedP    

    if (npSources .lt. 2) return

    mergedP = .false.
    nx = ubound(plevelimage, 1)
    ny = ubound(plevelimage, 2)

    tempLevelImage => Allocate2DInt16(0_int16, nx, 0_int16, ny, "tempeLevelImage")
    allocate(merged(1 : npsources))
  


    do k = 1, 3 ! was 3

        tempLevelImage = 0_int16
        
        n = npSources
        npSources = 0
   
        ! Load the extended-sources
        source => SourceDataArray(pSourceList, n)
        call DeleteSourceList(psourcelist)
        merged = .false.


        do i = 1, n
            if (merged(i)) cycle
            npSources = npSources + 1

   
            where(plevelImage(source(i)%imin : source(i)%imax,      &
                              source(i)%jmin : source(i)%jmax) .eq. source(i)%number) 
                    templevelImage(source(i)%imin : source(i)%imax, &
                                   source(i)%jmin : source(i)%jmax) = npSources
                                              
            end where
        

            error1 = source(i)%poserr
         

            do j = i + 1, n
                if (merged(j)) cycle
                error2 = source(j)%poserr
             


                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)
                error = 2.0 * sqrt(error1 * error1 + error2 * error2)
                error = max(error, tolerance)

                if (error .gt. radius) then ! Merge sources

                    merged(i) = .true.
                    merged(j) = .true.
                    write(errstr,'(a, i6, a, i6)') "Merging point sources ", i, " and ", j
                    call message(AppMsg, Verbose)
                    

                 
                
                    where(plevelImage(source(j)%imin : source(j)%imax, &
                                     source(j)%jmin : source(j)%jmax) .eq. source(j)%number) 
                      
                      templevelImage(source(j)%imin : source(j)%imax, &
                                     source(j)%jmin : source(j)%jmax) = npSources
                         
                    end where
                    
                    source(i)%number = npSources
                    source(i)%xpos = (source(i)%xpos + source(j)%xpos) / 2.0
                    source(i)%ypos = (source(i)%ypos + source(j)%ypos) / 2.0

                    source(i)%imin = min(source(i)%imin, source(j)%imin)
                    source(i)%imax = max(source(i)%imax, source(j)%imax)
                    source(i)%jmin = min(source(i)%jmin, source(j)%jmin)
                    source(i)%jmax = max(source(i)%jmax, source(j)%jmax)
                    source(i)%noSourcePixels = source(i)%noSourcePixels + source(j)%noSourcePixels
                end if        
            end do 
  
            if (merged(i)) then
                
               
                call ComputeExtendedSrcParameters2(int(source(i)%number, int16), nx, ny, image, backgroundImage, &
                    templevelImage, source(i), .true.)  ! True for Monte-carlo errors

                
                !**********************************************************************
                ! Classify the source - point (0) or extended (1)
                !*********************************************************************
                call ClassifySource(source(i), image, backgroundImage)
                if (source(i)%extensionFlag .eq. 0_int16) then
                   

                    source(i)%noSourcePixels = 0_int16

                end if


                mergedP = .true.

            else
                source(i)%number = npsources 
            end if 

            ! Add source to list
               
            call addSourceToList(pSourceList, source(i))   
           

        end do
   
        plevelimage = templevelimage
   
        deallocate(source)
        nullify(source)


        if (n .eq. npsources) exit
    end do

    if (mergedP) then

        write(errstr,'(a, i6)') "New number of likely point-sources=", npsources
        call message(AppMsg, Verbose)
    end if

    deallocate(merged, tempLevelImage)
    nullify(merged, tempLevelImage)
  
end subroutine mergePointSources





!*******************************************************************
!
!
!******************************************************************
subroutine GaussFitPhotometry(source, image, backgroundimage, levelimage)
    type(sourceParameters), pointer :: source(:) 
    real(single), pointer :: image(:, :)
    real(single), pointer :: backgroundimage(:, :)
    integer(int16), pointer :: levelImage(:, :)
    integer :: i
    type(GaussFitParameters) :: fitParameters

return
    if (.not. associated(source)) return

    do i = 1, size(source)
!if (i .ne. 21) cycle
        if (source(i)%noSourcePixels .lt. 20) cycle
        
        call fit2dGaussian3(source(i), image, binx, levelimage, &
                  backgroundimage, fitparameters=fitParameters)

!   call fit2dGaussian2(source(i), image, binx, levelimage, &
!                  fitparameters=fitParameters)



        
        ! Get the new raw counts
        write(*, *) "Source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos, " raw counts=", &
            source(i)%rawCountRate

        source(i)%rawCountRate = TWO_PI * fitParameters%height * fitParameters%sigmax * &
                                 fitParameters%sigmay / source(i)%exposure
        source(i)%xpos = fitParameters%xcentre
        source(i)%ypos = fitParameters%ycentre
        source(i)%corrCountRate = source(i)%rawCountRate

!        source(i)%semia = fitParameters%sigmax
!        source(i)%semib = fitParameters%sigmay
!        source(i)%theta = fitParameters%positionangle


        write(*, *) "New x=", source(i)%xpos, " y=", source(i)%ypos, " raw counts=", &
            source(i)%rawCountRate


    end do

end subroutine GaussFitPhotometry

!*****************************************************
!
!****************************************************
!subroutine GrismDetection1(image, levelImage, nSigma, nSources, sourceData)
!    real(single), pointer :: image(:, :)
!    integer(int16), pointer  :: levelImage(:, :)
!    real(single), intent(in) :: nSigma
!    integer(int16), intent(out) :: nSources
!    type(sourceParameters), pointer :: sourceData(:) 

!    type(SourceLinkedList) :: sourceList

!    integer(int16) :: i, nx, ny

!    real(single), pointer   :: backgroundImage(:, :) => Null()

!    real(single) :: modeBackground, sigma
!    integer(int8) :: minCounts, extensionFlag
!    integer(int16) :: nPSources
!    integer(int8) :: iter
!    real(single), pointer :: tempImage(:, :) => null()
!    integer(int16), pointer :: pLevelImage(:, :) => null()

!    nPSources = 0

!    nx = size(image, 1)
!    ny = size(image, 2)

!    tempImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "tempImage")

!    tempImage = image

!    minCounts = 2

!    backgroundImage => Allocate2DSingle(0, nx - 1, 0, ny - 1, "backgroundImage")

 


!    call SmoothGrismImage(tempImage, int(2 / binX, int16), int(12 / binY, int16))

!    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "levelImage")
!    levelImage = 0



!    modeBackground = ComputeBackgroundMode(tempImage, levelImage, int(nx, int16), int(ny, int16), 2.0, sigma)

!    sigma  = max(2.0, sqrt(modeBackground))

!    extensionFlag = 1

!    iter = 3

!    title = "first-order grism-sources"

!    pLevelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "plevelImage")
!    pLevelImage = 0

  


!    do i = iter, 1,  -1
!        write(errstr,'(a, i2)') "Iteration ", (iter - i + 1)
!        call message(AppMsg, Verbose)

!        backgroundImage = modeBackground + (nSigma + real(i)) * sigma

 

    

!        call GrismDetectExtendedSources(nx, ny, tempimage,    & 
!                        image, backgroundIMage,             &
!                        maxval(image), levelImage, &
!                        sourceList, nPSources, nSources,   &
!                        400 / binx, 40000 / binx,           &
!                        minCounts, isValidGrismSource, 1_int8)



     

 
!        where(levelImage .gt. 0) levelImage = 0
!        call SmoothLevelImage(levelImage, 4 / binX, 60 / binX)

!    end do
!    title = "zeroth-order grism-sources"
!    call SmoothLevelImage(levelImage, 4 / binX, 20 / binX)
!    tempImage = image
!    call SmoothGrismImage(tempImage, int(6 / binX, int16), int(8 / binY, int16))

!    npSources = 0
!    do i = iter, 1,  -1

!        write(errstr,'(a, i2)') "Iteration ", (iter - i + 1)
!        call message(AppMsg, Verbose)

!        backgroundImage = modeBackground + (nSigma + real(i)) * sigma



!        call GrismDetectExtendedSources(nx, ny, tempimage,    & 
!                        image, backgroundIMage,             &
!                        maxval(image), levelImage, &
!                        sourceList, nPSources, nSources,   &
!                        20 / binx, 5000 / binx,           &
!                        minCounts, isValidGrismSource, 1_int8)






!        where(levelImage .gt. 0) levelImage = 0
!        call SmoothLevelImage(levelImage, 4 / binX, 8 / binX)

!    end do

!    deallocate(backGroundImage)
!    nullify(backGroundImage)

!    deallocate(tempImage)
!    nullify(tempImage)
!    if (associated(plevelImage)) then
!        deallocate(plevelImage)
!        nullify(plevelImage)
!    end if
!    if (nSources .gt. 0) then
!        sourceData => SourceDataArray(sourceList, nSources)
!        call DeleteSourceList(sourceList)
!    end if

!end subroutine GrismDetection1



end module omdetect_mod









