!*************************************************************
! Module sourcedetection_mod
! Created 15th April 2002- C.Brindle
! Contains a subroutine for detecting "point" sources and one 
! for detecting any other kind of source
!****************************************************************
module omdetectdetection_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSkyBackground_mod  ! For computing sky background mode
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use parameters_mod       ! For computing source moments
    use OmSmoothImage_mod    ! For smoothing image
    use OmConstants_mod      ! For binx, biny, etc
    use omnumerics_mod 
    use stardecomp_mod
    use omphotometry_mod
    use leastsquares_mod
    use ominterpolate_mod
    use nonlinls_mod
    use brent_mod
    use cubicshepard_mod
    use splinefit_mod
    use omsourcemap_mod
    use psffunctions_mod   
    use omutility_mod
    use gridinterpolation_mod
!    use centroid_mod
    use zeroin_mod

public DetectPointSources, DetectExtendedSources
public IsPointSource, IsPointSource2




interface DetectExtendedSources 
    module procedure DetectExtendedSources1, DetectExtendedSources2, &
                     DetectExtendedSources3,  DetectExtendedSources4, &
                     DetectExtendedSources5
end interface DetectExtendedSources 




interface NumberOfStars
    module procedure NumberOfStars1, numberofstars2, &
                     numberOfStars3
end interface NumberOfStars

interface NumberOfNeighbours
    module procedure NumberOfNeighbours1, numberofNeighbours2, &
                     NumberOfNeighbours3
end interface NumberOfNeighbours

interface BackgroundGradientOkay
    module procedure  BackgroundGradientOkay1,  BackgroundGradientOkay2, &
                      BackgroundGradientOkay3
end interface BackgroundGradientOkay


!****************************************
! A structure to store detection parameters, 
! etc!
!****************************************
type DetectionParameters
    real(single) :: nStd(3) = (/5.0, 2.0, 1.0/)   ! The standard-deviation constant
    integer      :: nCircles = 4 ! The number of circles to compute mean counts for
    integer      :: nSize = 16   ! The number of circle intervals
    integer      :: kmin = 8     ! The minimum number of data points round circle
    real(single) :: minSigNoise = 5.0 ! 
    real(single) :: c2 = 1.0
    real(single) :: minSemib = 0.5
    real(single) :: semiAConstant = 1.0 
    real(single) :: meanSemia = 1.0_double
    real(double) :: semiaSum = 0.0_double
    integer :: n = 0
    real(single) :: exposure   
    integer(int16) :: minNoPixels
end type DetectionParameters


!***************************************************
! Global variables
!***************************************************
real(single), private, pointer :: FWHMImagePtr(:, :) => null()
real(single), private :: psfXcentre, psfYCentre
real(double), private :: fCentre
real(single), private, pointer, dimension(:) :: mean => null(), sigma => null()
real(single), private, pointer :: rCosTheta(:, :) => null(), rSinTheta(:, :) => null()
real(single), private, pointer :: imagePtr(:, :) => null() 
real(single), private, pointer :: smoothedImagePtr(:, :) => null() 
real(single), private, pointer :: backgroundImagePtr(:, :) => null() 
real(single), private, pointer :: backgroundSigmaImagePtr(:, :) => null() 
real(single), private :: xLower, yLower, xUpper, yUpper
real(single), private :: minimumCounts  ! For isValidSplitSource
real(single), private, pointer :: fCircle(:, :) => null()
integer(int16), private :: nx1, ny1, lbx, lby, ratio1, ratio2
integer(int16), private, pointer :: pLevelImagePtr(:, :) => null() ! level image map for point-sources
integer(int16), private, pointer :: eLevelImagePtr(:, :) => null() ! level image map for extended-sources
integer(int16), private :: number = 0 ! , sourceNumber = 0

type(DetectionParameters), private, save :: detectParams 
logical(boolean), pointer, private :: maskImage(:, :) => null()
logical(boolean), pointer, private :: emaskImage(:, :) => null()
integer(int8), private :: extensionFlag = 0_int8
real(single), private :: minHeight = 7.0  ! was 7.0
real(single), private, parameter :: xOffset(5) = (/0.0, -0.25, 0.25, -0.5, 0.5/)
real(single), private, parameter :: yOffset(5) = (/0.0, -0.25, 0.25, -0.5, 0.5/)

character(100) :: comment
real(single)  :: psfSemia1 = -1.0, psfSemiB1 = -1.0
real(single) :: psfSemia = -1.0, psfSemiB = -1.0

character(100) :: title

!*********************************************************
! The following is for test purposes to find out why a
! particular source hasn't been detected
!*********************************************************
integer, private, parameter :: ipixel=67, jPixel=396
logical, private :: checkPixel = .false.
contains

!*********************************************************
! Access functions
! for retrieving/setting private data from elsewhere
!*******************************************************
function GetMaskImage() result(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    maskImage1 => maskImage

end function GetMaskImage
!*********************************************************
!
!********************************************************
function GetEMaskImage() result(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    maskImage1 => emaskImage

end function GetEMaskImage
!*********************************************************
!
!
!********************************************************
subroutine setMaskImage(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    maskImage => maskImage1
     
end subroutine setMaskImage

!*********************************************************
!
!
!********************************************************
subroutine setEMaskImage(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    emaskImage => maskImage1
     
end subroutine setEMaskImage

!********************************************************
!
!*******************************************************
subroutine setMinimumCounts(minCounts)
    real(single), intent(in) :: minCounts

    minimumCounts = minCounts

end subroutine setMinimumCounts


!********************************************************
!
!*******************************************************
subroutine setMinimumNoPixels(minNoPixels)
    integer(int16), intent(in) :: minNoPixels

    detectParams%minNoPixels = minNoPixels - 1
 
end subroutine setMinimumNoPixels
!********************************************************
!
!*******************************************************
subroutine setMinimumHeight(height)
    real(single), intent(in) :: height

    minheight = height
 
end subroutine setMinimumHeight


!***********************************************************
! 
!
!***********************************************************
subroutine setBackgroundSigmaImagePtr(sigmaImagePtr)
    real(single), pointer :: sigmaImagePtr(:, :)
    backgroundsigmaImagePtr => sigmaImagePtr    

end subroutine setBackgroundSigmaImagePtr



!********************************************************
! Subroutine DetectPointSources
! This routine is intended to be the first
! search for point-sources- hence maskimage
! is initialised to true so that all pixels
! will be checked.
!
!
!********************************************************
subroutine DetectPointSources(nx,              &  ! x image dimension
                              ny,              &  ! y - image dimesnion
                              image,           &  ! The image
                              smoothedImage,   &  ! The smoothed image
                              backGroundImage, &  ! The background image
                              backgroundSigmaImage, & !
                              levelImage,      &  ! The level-image map
                              nSigma,          &
                              exposure,        &
                              sourceList,      &  ! The linked-list source-list
                              noSources)          ! The number of point-sources detected

    integer(int16), intent(in) :: nx, ny
    real(single), pointer, dimension(:, :) :: image, smoothedImage
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)


    integer(int16), pointer :: levelImage(:, :)
    real(single), intent(in) :: nSigma, exposure
    type(SourceLinkedList), intent(inout):: sourceList  ! The linked-list source-list
    integer(int16), intent(out) :: noSources  
    integer(int16) :: iStart   
    real(single) :: height

    write(errstr,'(a)') "Looking for point sources"
    call message(AppMsg, Verbose)

  
    detectParams%exposure = exposure 
    detectParams%minNoPixels = 2 / binx - 1
 
    ! Initialise some global variables
    ! Need nstd=0.1 to detect extreme strong mod-8 sources


!call SourceSearch(image, backgroundImage, levelImage, 2, 100, 3.0, isvalidesource3, heightFunction1)



 
    call GetPSFWidths(psfsemiA1, psfsemiB1)

    psfParams%psfSemiA = psfSemiA1 !* 1.25  ! Was2.0
    psfParams%psfSemiB = psfSemiB1 !* 1.25  ! Was 2.0

    lbx = lbound(image, 1)
    lby = lbound(image, 2)
    

 !   call GetPSfParameters




    ! Find very bright sources (badly affected by mod-8 sources)  
    call InitialiseDetection(nx, ny, 1.0, 3, 0.1, 0.1, 1.5) ! was 1.5
 
 


!write(*, *) "line 271 maskimage(", ipixel, ",", jpixel, ")=", maskimage(ipixel, jpixel)



    where(image .lt. 0.0 .or. image .lt. (backgroundImage - 5.0 * backgroundSigmaImage)) ! Was 2.0
        maskImage = .true.
    elsewhere 
        emaskImage = .false.
        maskImage = .false. ! Test all pixels for a possible source
    end where 

!write(*, *) "line 282 maskimage(", ipixel, ",", jpixel, ")=", maskimage(ipixel, jpixel)

!write(*, *) " line 256 maskimage ", ipixel, " , ", jpixel, " =", maskimage(ipixel, jpixel), &
!" pixelvalue=", image(ipixel, jpixel)


   detectParams%minNoPixels = 2 / binx - 1

! write(*, *) "line 290 maskimage(", ipixel, ",", jpixel, ")=", maskimage(ipixel, jpixel)


   call SearchForPointSources(image,                 &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               2500.0,               &  ! was 2500  Minimum height above background 
                               IsPointSourceMod8,    & 
                               HeightFunction2)  

!write(*, *) "line 304 maskimage(", ipixel, ",", jpixel, ")=", maskimage(ipixel, jpixel)




    detectParams%minNoPixels = 2 / binx - 1


    ! Find other very bright sources (near mod-8 sources)
    call InitialiseDetection(nx, ny, 1.0, 3, 2.0, 2.0, 1.5) ! Was 2.0, 2.0, 1.5


    call SearchForPointSources(image,                &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &                          
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               50.0,                 & ! was 50 Minimum height above background
                               IsPointSource,        &
                               HeightFunction2)




!write(*, *) " line 293 maskimage ", ipixel, " , ", jpixel, " =", maskimage(ipixel, jpixel)


    detectParams%minNoPixels = 4 / binx - 1
 

!call fatal("stopped")
    call InitialiseDetection(nx, ny, 0.8, 3, 2.0, 2.0, 1.5) ! Was 2.0, 2.0, 1.5
    call SearchForPointSources(image,                &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               50.0,                 & ! height above background
                               IsPointSource2,       &
                               HeightFunction2) 

!write(*, *) " line 358 maskimage ", ipixel, " , ", jpixel, " =", maskimage(ipixel, jpixel)


    !*************************************************************
    ! Added for G153 images
    !*************************************************************                         
    call InitialiseDetection(nx, ny, 1.0, 2, 3.0, 1.0, 1.1) 

    height = max(10.0, 10.0 + nSigma - 2.0)

    call SearchForPointSources(image,                  &
                               smoothedImage,          & 
                               backgroundImage,        &  
                               backgroundsigmaImage,   &
                               levelImage,             &
                               sourceList,             & 
                               noSources,              &
                               height,                 & ! Was 5.0 The height above the background
                               IsPointSource3,         & ! Was IsPointSource2 
                               HeightFunction1)          ! 
         

!write(*, *) " line 385 maskimage ", ipixel, " , ", jpixel, " =", maskimage(ipixel, jpixel)


    !****************************************************************************
    ! For removing extended light round bright sources (mod-8 patterns, etc
    !****************************************************************************
    
    call BrightSourceMask(noSources, 1_int16, sourceList, smoothedImage, backgroundimage)
    iStart = noSources + 1   
 

    detectParams%minNoPixels = 4 / binx - 1
 
                       

    call InitialiseDetection(nx, ny, 1.0, 2, 4.0, 1.0, 1.1)  ! 1.0, 2, 3.0, 2.0   ! 0.9 was 1.0

    height = max(3.0, 3.0 + nSigma - 2.0) ! was 3

    call SearchForPointSources(image,                  &
                               smoothedImage,          & 
                               backgroundImage,        &  
                               backgroundsigmaImage,   &
                               levelImage,             &
                               sourceList,             & 
                               noSources,              &
                               height,                 & ! Was 5.0 The height above the background
                               IsPointSource33,        & ! Was IsPointSource2 
                               HeightFunction1) 



    !****************************************************************************
    ! For removing extended light round bright sources (mod-8 patterns, etc
    !****************************************************************************
    call BrightSourceMask(noSources, iStart, sourceList, smoothedImage, backgroundimage)
    

!call saveimage(backgroundsigmaimage, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 421 displaying backgroundsigmalimage.fits")


!write(*, *) "Line 354 omdetectdetection Pixel 145, 354 maskimage=", maskImage(ipixel, jpixel)

end subroutine DetectPointSources
!*********************************************
! Deallocates various arrays used in the
! detection routines
!
!********************************************
subroutine Delete
    if (associated(maskImage)) then
        deallocate(maskImage, emaskimage)
        nullify(maskImage, emaskimage)


        deallocate(rCosTheta, rSinTheta, fCircle, mean, sigma)
        nullify(rCosTheta, rSinTheta, fCircle, mean, sigma)
    end if



    if (associated(psfParams%psfImage)) then
        deallocate(psfParams%psfImage)
        nullify(psfParams%psfImage)
    end if

    if (associated(imagePtr)) then
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)
    end if

    if (associated(backgroundsigmaimagePtr)) then
        nullify(backgroundsigmaimagePtr)
    end if


end subroutine Delete   


!***********************************************************
! 
!
!***********************************************************
subroutine SearchForPointSources(&
           image,           &  ! The image
           smoothedImage,   &  ! The smoothed image
           backgroundImage, &  ! The background image
           backgroundsigmaImage, &  ! The background sigma image

           levelImage,      &  ! The level-image image
           sourceList,      &  ! The linked-list source-list
           noSources,       &  ! The total number of detected point-sources
           heightAboveBck,  &  ! The minimum pixel-height above the background to be checked
           IsValidPointSource, &   ! The function to be used
           HeightFunction)      
    real(single), pointer, dimension(:, :) :: image, smoothedImage, backgroundImage, backgroundsigmaimage
    type(SourceLinkedList), intent(inout):: sourceList
    integer(int16), intent(inout) :: noSources
    real(single), intent(in) :: heightAboveBck
    integer(int32) :: npixels
    interface
       
        Function IsValidPointSource(i0, j0, sourceP) result(d)
        use dal
        use OmSourceParam_mod    ! Defines structures for storing source parameters
            integer(int16), intent(inout) :: i0, j0    
            type(SourceParameters) :: sourceP
            logical :: d 
        end function IsValidPointSource
    end interface

 

 
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface



    type(SourceParameters) :: data
       
    integer(int16), pointer :: levelImage(:, :)     
    integer(int16) :: i0, j0
    integer(int8) :: dx, dx2
    integer(int16) :: count1, nl     

    integer(int16), pointer, dimension(:) :: iIndex => null(), jIndex => null()
    integer(int16) :: i, j, k
    logical(boolean), pointer :: tempMaskImage(:, :) => null()
    integer(int16) :: nx, ny

    nx = nx1 + 1
    ny = ny1 + 1

    imagePtr => image    
    smoothedImagePtr => smoothedImage    

    backgroundImagePtr => backgroundImage    
    backgroundsigmaImagePtr => backgroundsigmaImage    



    allocate(tempMaskImage(0 : nx1, 0 : ny1))




!write(*, *) "line 534 maskimage(", ipixel, ",", jpixel, ")=", maskimage(ipixel, jpixel)


 
    dx = 6 / binx  
    dx2 = 4 / binx
   
    do j = lby, ny1
        do i = lbx, nx1
            if (maskImage(i, j)) then
                tempMaskImage(i, j) = .true.
                cycle
            end if
            if (image(i, j) .lt. heightfunction(i, j, heightAboveBck, backgroundimage)) then
                tempMaskImage(i, j) = .true.
            else         
                tempMaskImage(i, j) = .false.
       
            end if
         end do
    end do

!write(*, *) "line 556 tempmaskimage(", ipixel, ",", jpixel, ")=", tempmaskimage(ipixel, jpixel)



    ! Create the 1-d arrarys arrayIndex, iIndex and jIndex
    ! From the 2-d arrays image and maskImage
    call SetUpIndices(image, tempMaskImage, iIndex, jIndex, nPixels)

!write(*, *) "line 564 tempmaskimage(", ipixel, ",", jpixel, ")=", tempmaskimage(ipixel, jpixel)
!write(*, *) "maskimage(", ipixel, ",", jpixel, ")=", maskimage(ipixel, jpixel)

    deallocate(tempMaskImage)
    nullify(tempMaskImage)

    nl = max(1_int16, noSources)
  
    sourceNumber = maxval(levelImage)
    
    ! Set up the mask so that where any image pixel
    ! where the mask image pixel is false
    ! the image pixel is skipped


!*****************************************************************
! NB FOR TESTING TO SEE WHY A PARTICULAR SOURCE IS REJECTED
! SET MASK IMAGE TO PIXEL COORDINATES OF MAXIMUM SOURCE PIXEL
!****************************************************************

    count1 = 10

    do k = 1, nPixels

        ! Detect sources in order of decreasing maximum pixel value    
        ! Get the location (i, j) of the maximum pixel- excluding
        ! those pixels where maskImage = true

        i0 = iIndex(k)
        j0 = jIndex(k)

        if (maskImage(i0, j0)) cycle        



!        if (i0 .eq. iPixel .and. j0 .eq. jPixel) then
!            checkPixel = .true.
!            write(*, *) "Line 595 checking pixel ", ipixel, ", ", jpixel
!        else
!            checkPixel = .false.
!        end if  
    
        ! See if this pixel is centred on a source      
        if (IsValidPointSource(i0, j0, data)) then

                                                                                               
            sourceNumber = sourceNumber + 1 ! The new source number
            data%qualityFlag = 0
          

            !*****************************************************
            ! 
            !****************************************************
            if (IsMod8Source(data, image)) then
                write(errstr, '(a, i6, a, f10.2, a, f10.2, a)') "Source ", sourceNumber, " (", data%xpos, ", ", data%ypos, &
                  ") is probably affected by mod-8 noise"
                call message(AppMsg, Verbose)
            
       
            end if 


            if (mod(sourceNumber, count1) == 0) then

                write(errstr,'(a, i5, a, f8.1)') &
                     "No of likely point-sources detected so far=", sourceNumber, &
                     " pixel level=", image(i0, j0)
                call message(AppMsg, Verbose)
                if (sourceNumber > 100) count1 = 100
            end if 
            data%number = sourceNumber
            data%imin = max(lbx, i0 - dx)
            data%imax = min(nx1, i0 + dx)
            data%jmin = max(lby, j0 - dx)
            data%jmax = min(ny1, j0 + dx)

            call RefineSourceMap3(data, image, backgroundImage, backgroundsigmaimage, levelImage, 2.0, &
                    detectparams%exposure, maskImage) ! was 2.0


            if (image(i0, j0) .gt. 5000) then
           
                call ComputePointSourceParameters(sourceNumber, nx1, ny1, &
                         image, backgroundImage, i0, j0, data, dx2, computemonteerrors=.true.)
            else

                call ComputeExtendedSrcParameters2(int(data%number, int16), nx1, ny1, image, backgroundImage, &
                    levelImage, data, .true.)  ! True for Monte-carlo errors
 
            end if           

!if (btest(data%qualityFlag, 4))  call Centroid(data, imagePtr, binx)

 
            !**********************************************************************
            ! Classify the source - point (0) or extended (1)
            !********************************************************************* 
            call ClassifySource(data, image, backgroundImage)
            if (data%extensionFlag .eq. 1_int16) then
                data%noSourcePixels = count(levelImage(data%imin : data%imax, data%jmin : data%jmax) &
                             .eq. data%number)

            else

                data%noSourcePixels = 0_int16 
     
  

            end if

!write(*, *) "Line 638 source ", data%number, " extensionflag=", data%extensionFlag
!write(*, *) "semia=", data%semia, "+/-", data%semiaerr,  " semib=", data%semib, "+/-", data%semiberr
!write(*, *) "psf semia=", psfparams%psfsemia, " semib=", psfparams%psfsemib
!data%extensionFlag = 0 



           ! Add it to the source list                
            call AddSourceToList(sourceList, data)
    

        end if

    end do    


    noSources = maxval(levelImage)
    write(errstr,'(a, i6)') "Number of likely point-sources detected=", noSources
    call message(AppMsg, Verbose)

    deallocate(iIndex, jIndex)  
    nullify(iIndex, jIndex)  

    
end subroutine SearchForPointSources


!***********************************************************
! Function IsPointSource
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source
 

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k, kbound, kmask
    real(single) :: xc, yc
    
!if (checkPixel) then
!write(*, *) "Line 1510 ispointsource Checking  pixel ", i0, ", ", j0
!end if
 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5

    do l = 1, 5

        xCentre = xc + xOffset(l)
         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) then
                        kbound = kbound + 1
                        cycle

                    end if 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) then
                        kbound = kbound + 1                   
                        cycle
                    end if 
                    if (imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        


                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (i .gt. 1) then
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) then
                            kmask = kmask + 1
                            cycle
                        end if   
                    end if  

                    k = k + 1 
                    fCircle(k, 1) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do

!                if (kbound .gt. 0 .and. kmask .gt. 0) then  ! Source near an edge

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line 683 k=", k, " rejecting source, n=", detectParams%nSize

!end if

!                    cycle mLoop 
           
!                else if (k .lt. detectParams%kMin) then 
                 if (k .lt. detectParams%kMin) then 


!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line 496 k=", k, " rejecting source, n=", detectParams%nSize

!end if

                    cycle mLoop 
           
                end if   
 

                call computeStatistics3(k, fCircle(1 : detectParams%nSize , 1), mean(i), sigma(i))

                if (i == detectParams%nCircles) cycle

                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              
!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "line 1581 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                    
                    cycle mLoop          
                end if

            end do

            !**********************************************************
            ! Found a possible point-source
            !*********************************************************
            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5
 
            IsPointSource = .true.

!if (checkPixel) then
!write(*, *) "Source ", i0, ", ", j0, " has been found!"
!write(*, *) "Image value=", imagePtr(i0, j0), " background =", backgroundImagePtr(i0, j0)
!write(*, *) "mean semi a=", detectParams%meanSemiA

!            write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!            write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!            write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if

            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource = .false.
end Function IsPointSource

!***********************************************************
! Function IsPointSourceMod8
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSourceMod8(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source
   

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
 
!   IsPointSourceMod8 = .false.
!return

!if (checkPixel) then

!write(*, *) "Line 1510 ispointsourcemod8 Checking  pixel ", i0, ", ", j0
!end if
 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 5

        xCentre = xc + xOffset(l)
         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 
                    if (imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (i .gt. 1) then
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) cycle   
                    end if  

                    k = k + 1 
                    fCircle(k, 1) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do
                if (k .lt. detectParams%kMin - 2) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre     
!write(*, *) "Line 496 k=", k, " rejecting source, n=", detectParams%nSize
!end if

                    cycle mLoop 
           
                end if   
 
                ! Compute the mean pixel value and variance around the circle 

                call computeStatistics3(k, fCircle(1 : detectParams%nSize , 1), mean(i), sigma(i))
              

                if (i == detectParams%nCircles) cycle

                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then

!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
       
!write(*, *) "line 1581 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    

!CALL  CheckSource(xCentre, yCentre)
!end if  
                    
                    cycle mLoop                
                end if
            end do

            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5
           
            IsPointSourceMod8 = .true.

!if (checkPixel) then
!write(*, *) "Source ", i0, ", ", j0, " has been found!"
!write(*, *) "Image value=", imagePtr(i0, j0), " background =", backgroundImagePtr(i0, j0)
!write(*, *) "mean semi a=", detectParams%meanSemiA

!            write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!            write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!            write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if


            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSourceMod8 = .false.
end Function IsPointSourceMod8



!***********************************************************
! Function IsPointSource2
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource2(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
 
    real(single) :: fCentre  ! The centre pixel counts



    if (.not. BackgroundGradientOkay(i0, j0)) then
        IsPointSource2 = .false.
        return
    end if


  

!if (checkPixel) then
!write(*, *) "Line 1697 ispointsource2  Checking  pixel ", i0, ", ", j0
!end if
 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5




    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !*************************************************************************
            ! Counts at the centre
            !************************************************************************
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre
            fcentre = fcentre - 2.0 * sqrt(fcentre)


            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 
                    if ( imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (i .gt. 1) then
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) cycle   
                    end if  

                    k = k + 1 
                    fCircle(k, 1) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do
                
                if ( any(fcircle(1 : k, 1) .ge. fcentre)) then
                    cycle mloop
                end if


!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line 1742  rejecting source count(anycircle > ge fcentre=",  count(fcircle(1 : k, 1) .ge. fcentre)

!end if

!                call MinimumList(k, fCircle(1 : detectParams%nSize , 1), k - 2)

!                if ( count(fcircle(1 : k, 1) .ge. fcentre) .gt. 0) then

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line 1742  rejecting source count(anycircle > ge fcentre=",  count(fcircle(1 : k, 1) .ge. fcentre)

!end if

!                    cycle mloop 

!                end if

                if (k .lt. detectParams%kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line 496 k=", k, " rejecting source"

!end if

                    cycle mLoop 
           
                end if   
 

                ! Compute the mean pixel value and variance around the circle 
                call computeStatistics3(k, fCircle(1 : detectParams%nSize , 1), mean(i), sigma(i))



                if (i == detectParams%nCircles) cycle

 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource2 xCentre=", xCentre, " yCentre=", yCentre
       
!write(*, *) "line 1341 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    

!end if  
                 
                    cycle mLoop

                
           
                end if


            end do


            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5


            IsPointSource2 = .true.

!if (checkPixel) then
!write(*, *) "Ispointsource2 Source ", i0, ", ", j0, " has been found!"


!            write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!            write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!            write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if





            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource2 = .false.
end Function IsPointSource2


!***********************************************************
! Function IsPointSource3
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource3(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
    real(single) :: fCentre  ! The centre pixel counts
    real(single) :: fMin, fMax, minPixRatio
    logical :: hasNeighbour

!if (checkPixel) write(*, *) "ispointsource3 line 1211 checking pixel ", ipixel, " , ", jpixel
    if (.not. BackgroundGradientOkay(i0, j0)) then ! Was 0.1
        IsPointSource3 = .false.
        return
    end if

!if (checkPixel) write(*, *) "ispointsource3 line 1216 checking pixel ", ipixel, " , ", jpixel



    minPixRatio = 0.6 ! WAS 0.7


 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)
            hasNeighbour = .false.
            !*************************************************************************
            ! Counts at the centre
            !************************************************************************
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre

            fcentre = fcentre - 2.0 * sqrt(fcentre)

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 

                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    if (i .gt. 1) then                          
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) then
                            hasNeighbour = .true.
                            
                            cycle

                        end if   
                    end if

                    k = k + 1 
                    fCircle(k, 1) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do
                if (k .lt. 10) cycle mLoop

                if (hasNeighbour) then
                    minPixRatio = 0.6   ! WAS 0.6
                    ! Compute the mean pixel value and variance around the circle 

                    call heapsort(fCircle(1 : detectParams%nSize , 1), k)

                    call MinimumList(k, fCircle(1 : detectParams%nSize , 1), k - 4) ! was 4


                end if

!if (checkPixel) then
!write(*, *) "Line 1672 ispointsource3 hasneighbour=", hasneighbour, " minpixratio=", minpixratio, " k=", k
      
!end if




                ! Compute the mean and variance

                call computeStatistics3(k, fCircle(1 : detectParams%nSize , 1), mean(i), sigma(i))


                fMax = maxval(fcircle(1 : k, 1))                 
                if ( fmax .ge. fcentre) then
                      !  if (checkPixel) then
                      !      write(*, *) "ispointsource3 line 1694 rejecting source fmax>fcentre"
!                        end if
                    cycle mloop
                end if


                fmin = minval(fcircle(1 : k, 1))                 
             
                if ( fMin / fMax .lt. minpixratio) then
                       ! if (checkPixel) then
                       !     write(*, *) "ispointsource3 line 1692 rejecting source"
                       !     write(*, *) "pixel ratio=",  (fmin / fmax)
                       !     write(*, *) "fcentre=", fcentre, " mean1=", mean(1), " std=", sqrt(sigma(1))
                       ! end if

                    cycle mloop
                end if


                if (i == detectParams%nCircles) cycle
 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource3 line 1712 xCentre=", xCentre, " yCentre=", yCentre
       
!write(*, *) "line 1714 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    

!end if  
                 
                    cycle mLoop

                
           
                end if


            end do

      
            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5

    
            IsPointSource3 = .true.

!if (checkPixel) then
!write(*, *) "line 1740 ispointsource3 Source ", ipixel, ", ", jpixel, " has been found!"
!write(*, *) "i0=", i0, " j0=", j0

!            write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!            write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!            write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if


!write(*, *) "ispointsource3 line 2300"



            return
        end do mLoop
    end do




    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource3 = .false.
end Function IsPointSource3


!***********************************************************
! Function IsPointSource3
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource33(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source


    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
    real(single) :: minPixRatio
   



    if (.not. BackgroundGradientOkay(i0, j0)) then
        IsPointSource33 = .false.
        return
    end if




!   checkPixel = .false.

   minPixRatio = 0.7

!if (checkPixel) write(*, *) "Line 1988 ispointsource33  Checking  pixel ", i0, ", ", j0

 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)
          
       

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 

                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if ( imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    if (i .gt. 1) then                          
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) then
                          
                            
                            cycle

                        end if   
                    end if

                    k = k + 1 
                    fCircle(k, 1) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do

                if (k .lt. detectParams%kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line ispointsource33 line 1870 k=", k, " rejecting source"

!end if

                    cycle mLoop 
           
                end if   
 

                ! Compute the mean and variance
                call computeStatistics3(k, fCircle(1 : detectParams%nSize , 1), mean(i), sigma(i))

  
                if (i == detectParams%nCircles) cycle

 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource33 line 2108 xCentre=", xCentre, " yCentre=", yCentre
       
!write(*, *) "line 2107 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    

!end if  
                 
                    cycle mLoop           
                end if
            end do

      
            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5

   
            IsPointSource33 = .true.

!if (checkPixel) then
!write(*, *) "line 2156 ispointsource3 Source ", ipixel, ", ", jpixel, " has been found!"
!write(*, *) "i0=", i0, " j0=", j0

!            write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!            write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!            write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if





            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource33 = .false.
end Function IsPointSource33
!*******************************************************
!
!*****************************************************
logical function BackgroundGradientOkay1(i0, j0)
    integer(int16), intent(in) :: i0, j0
    
   
    integer :: il, iu, jl, ju
 
    integer(int16) :: dx = 2
    real :: fmin, fmax, fmax1, mean, fraction, sourceCounts, sigNoise
    if (imagePtr(i0, j0) .ge. backgroundImagePtr(i0, j0) + 50.0) then
        BackgroundGradientOkay1 = .true.
        return
    end if

    fmax = imagePtr(i0, j0)
    dx = 14 / binx

    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    fmax1 = maxval( imagePtr(il : iu, jl : ju))
    if (fmax1 .gt. fmax * 1000.0) then

!if (checkPixel) &
!write(*, *) "Backgroundgradientokay1 line 1708 rejecting source at ", ipixel, ", ",  jpixel
      

        BackgroundGradientOkay1 = .false.
        return
    end if




    dx = 4 / binx

    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)






    fmax = maxval(backgroundImagePtr(il : iu, jl : ju))
    if (fmax .lt. 0.1) then
        BackgroundGradientOkay1 = .true.
        return
    end if
 
    fmin = minval(backgroundImagePtr(il : iu, jl : ju), backgroundImagePtr(il : iu, jl : ju) .ge. 0.0)

    mean = (fmax + fmin) / 2.0
    sourceCounts = imagePtr(i0, j0) - mean
    sigNoise = sourceCounts / sqrt(imagePtr(i0, j0))

    fraction =  (fmax - fmin) / fmax
   

    select case (nint(signoise))
        case (:5)  ! All values below 5
            if ( fraction .gt. 0.25) then
                BackgroundGradientOkay1 = .false.
!if (checkPixel) &
!write(*, *) "Backgroundgradientokay1 line 1890 rejecting source at ", ipixel, ", ",  jpixel
                return
            end if
        case (6 : 10) ! Values between 6 and 7
            if ( fraction .gt. 0.25) then
                BackgroundGradientOkay1 = .false.
!if (checkPixel) &
!                write(*, *) "Backgroundgradientokay1 line 1890 rejecting source at ", ipixel, ", ",  jpixel

                return
            end if
        
        case default
            BackgroundGradientOkay1 = .true.
            return
    end select
   
end function BackgroundGradientOkay1   
!*******************************************************
!
!*****************************************************
logical function BackgroundGradientOkay2(i0, j0, maxFrac)
    integer(int16), intent(in) :: i0, j0
    real(single), intent(in) :: maxFrac
   
    integer(int16) :: il, iu, jl, ju
 
    integer(int16), parameter :: dx = 2
    real :: fMin, fMax, mean, fraction, sourceCounts, sigNoise
! BackgroundGradientOkay2 = .true.
!return


    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    fmax = maxval(backgroundImagePtr(il : iu, jl : ju))
    if (fmax .lt. 0.1) then
        BackgroundGradientOkay2 = .true.
        return
    end if
 
    fmin = minval(backgroundImagePtr(il : iu, jl : ju), backgroundImagePtr(il : iu, jl : ju) .ge. 0.0)

    mean = (fmax + fmin) / 2.0
    sourceCounts = imagePtr(i0, j0) - mean
    sigNoise = sourceCounts / sqrt(imagePtr(i0, j0))
    if (sigNoise .ge. 10.0) then

        BackgroundGradientOkay2 = .true.
        return

    end if


    fraction =  (fmax - fmin) / fmax
    if (fraction .gt. maxFrac) then

        BackgroundGradientOkay2 = .false.
    else
        BackgroundGradientOkay2 = .true.
    

    end if


   
end function BackgroundGradientOkay2   

!*******************************************************
! i0, j0, coordinates of maximum source pixel counts
!*****************************************************
logical function BackgroundGradientOkay3(i0, j0, pixel)
    integer(int16), intent(in) :: i0, j0
    type(pixelStruct) :: pixel(:)
    real(single) :: maxRadius
    real(single), parameter :: maxFrac = 0.4   
    integer(int16) :: n, i, j, k, il, iu, jl, ju
    integer(int16) :: dx = 2
    real(single) :: fSourceMax
    real(single) :: radius, fMin, fMax, mean, fraction

! BackgroundGradientOkay3 = .true.
!return

    do k = 1, 2  
        dx = 6 * k / binx ! was 6
        il = max(lbx, minval(pixel%x) - dx)
        iu = min(nx1, maxval(pixel%x) + dx)
        jl = max(lby, minval(pixel%y) - dx)
        ju = min(ny1, maxval(pixel%y) + dx)

        fSourceMax = imagePtr(i0, j0) * real(5 * k) ! was 1.5
    

        do j = jl, ju
            do i = il, iu                             
                if (elevelImageptr(i, j) .ne. number) then
                    if (imagePtr(i, j) .gt. fSourceMax) then
                        BackgroundGradientOkay3 = .false.

!if (abs(i0 - ipixel) .lt. 4 .and. abs(j0 - jPixel) .lt. 4) then
!write(*, *) "Backgroundgradientokay1 line 1857 rejecting source at ", ipixel, ", ",  jpixel
!write(*, *) "fsourcemax=", fsourcemax, " image=", imagePtr(i, j), " i0=", &
!i0, " j0=", j0, " i=", i, " j=", j
!end if

                        return
                    end if
                end if
            end do
        end do
    end do




    n = size(pixel)
   



    maxRadius = real(16 / binx)

    dx = ceiling(maxradius) + 1
    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)

   
    fmax = 0.0
    fmin = 1e6
    do j = jl, ju
        do i = il, iu
            radius = sqrt(real(i - i0) ** 2 + (j - j0) ** 2)
            if (radius .le. maxRadius) then
                 
                fmax = max(fmax, backgroundImagePtr(i, j))
                fmin = min(fmin, backgroundImagePtr(i, j))
                 
                if (elevelImageptr(i, j) .eq. number .and. &
                       imagePtr(i, j) .gt. fSourceMax) then
                    BackgroundGradientOkay3 = .false.
!if (abs(i0 - ipixel) .lt. 4 .and. abs(j0 - jPixel) .lt. 4) &
!write(*, *) "Backgroundgradientokay1 line 1897 rejecting source at ", ipixel, ", ",  jpixel


                    return
                end if


            end if
        end do
    end do

    if (fmax .lt. 0.1) then
        BackgroundGradientOkay3 = .true.
        return
    end if

    fmin = max(0.2, fmin)  

    mean = (fmax + fmin) / 2.0



    fraction =  (fmax - fmin) / fmax
    if (fraction .gt. maxFrac) then

!if ( abs(i0 - ipixel) .lt. 4 .and. abs(j0 - jpixel) .lt. 4) then
!write(*, *) "Line 2056 backgroundgradientokay fraction=", fraction, " maxfrac=", maxfrac
!write(*, *) "fmin =", fmin, " fma=", fmax
!end if


        BackgroundGradientOkay3 = .false.
    else
        BackgroundGradientOkay3 = .true.
    

    end if


   
end function BackgroundGradientOkay3   

!*******************************************************
! WidthOkay
! Checks to see that the width of the source
! in the middle is greater than that at the ends
!*****************************************************
logical function WidthOkay(pixel)
    type(pixelStruct) :: pixel(:)

    integer(int16) :: xc, yc, xmin, xmax, ymin, ymax, dx, dy, dx1, dy1
!    integer :: i
!    logical :: checkPixel

    xmin = minval(pixel%x)
    xmax = maxval(pixel%x)
!    checkPixel = .false.
!widthOkay = .true.
!return


!if (any(pixel%x .eq. ipixel .and. pixel%y .eq. jpixel)) then 
!checkPixel = .true.
!write(*, *) "Width okay checking pixel ", ipixel, " , ", jpixel
!do i = 1, size(pixel)
!write(*, *) i, " x=", pixel(i)%x, " y=", pixel(i)%y
!end do

!write(*, *) "xmin=", xmin, " xmax=", xmax
!        ymin = minval(pixel%y)
!        ymax = maxval(pixel%y) 
    
!write(*, *) "ymin=", ymin, " ymax=", ymax


!end if

    if ( (xmax - xmin) .gt. 8) then

!if (checkPixel) write(*, *) "Line 1973"
        xc = (xmin + xmax) / 2              ! Centre x


        ymin = minval(pixel%y, pixel%x .eq. xc)
        ymax = maxval(pixel%y, pixel%x .eq. xc)
        dy = ymax - ymin                    ! Vertical width at centre
        
    
        ymin = minval(pixel%y, pixel%x .lt. xmin + 1 .or. pixel%x .gt. xmax - 1)  
      
        ymax = maxval(pixel%y, pixel%x .lt. xmin + 1 .or. pixel%x .gt. xmax - 1)  
      
   

  
        dy1 = ymax - ymin                   ! Vertical width at left-hand edge

        if (dy1 .gt. dy * 3) then ! was 2
!if (checkPixel) write(*, *) "Line 1892"

            widthOkay = .false.
!            call setMaskPixels(pixel)
            return
        end if
    

    end if

    ymin = minval(pixel%y)
    ymax = maxval(pixel%y)
!    if (checkPixel) write(*, *) "Line 1904 ymin=", ymin, " ymax=", ymax

    if ( (ymax - ymin) .gt. 8) then
        yc = (ymin + ymax) / 2              ! Centre y
        xmin = minval(pixel%x, pixel%y .eq. yc)
        xmax = maxval(pixel%x, pixel%y .eq. yc)
        dx = xmax - xmin                    ! Horizontal width at centre


        xmin = minval(pixel%x, pixel%y .lt. ymin + 1 .or. pixel%y .gt. ymax - 1)  
        xmax = maxval(pixel%x, pixel%y .lt. ymin + 1 .or. pixel%y .gt. ymax - 1)  
      
   
        dx1 = xmax - xmin                   ! Horizontal width at left-hand edge


        if (dx1 .gt. dx * 3) then ! was 2
            widthOkay = .false.
!            call setMaskPixels(pixel)
            widthOkay = .false.
            return
        end if   
        

    end if

    widthOkay = .true.

end function WidthOkay
!**********************************************
!
!**********************************************
subroutine SetMaskPixels(pixel)
    type(pixelStruct) :: pixel(:)

    integer(int16) :: i

    if (associated(eMaskImage)) then
    do i = 1, size(pixel)

        eMaskImage(pixel(i)%x, pixel(i)%y) = .true.
    end do


    end if

    if (associated(MaskImage)) then
    do i = 1, size(pixel)

        MaskImage(pixel(i)%x, pixel(i)%y) = .true.
    end do


    end if


end subroutine SetMaskPixels

!***********************************************************
! Function IsPointSource22
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource22(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
    real(single) :: fCentre

!if (checkPixel) then
!write(*, *) "Line ispointsource22 line 1946  Checking  pixel ", i0, ", ", j0
!end if


    if (.not. BackgroundGradientOkay(i0, j0)) then
        IsPointSource22 = .false.
        return
    end if

 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5




    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !*************************************************************************
            ! Counts at the centre
            !************************************************************************
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre
          
          

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 

                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
  

                    k = k + 1 
                    fCircle(k, 1) = Interpolate(smoothedImagePtr, xPos, yPos)
                 end do

            
                if (k .lt. detectParams%kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
      
!write(*, *) "Line 496 k=", k, " rejecting source"

!end if

                    cycle mLoop 
           
                end if   
 

                ! Compute the mean pixel value and variance around the circle 


                call computeStatistics3(k, fCircle(1 : detectParams%nSize , 1), mean(i), sigma(i))
!                if (i .lt. 3) then
!                    fBackgroundCentre = Interpolate(backgroundImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre
!                    if (mean(i) .lt. ( fBackgroundCentre + max(1.0, 0.5 * sqrt(fBackgroundCentre)))) cycle mloop
!                end if

                if (i == detectParams%nCircles) cycle


 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource22 xCentre=", xCentre, " yCentre=", yCentre
       
!write(*, *) "line 1341 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    


!end if  
                 
                    cycle mLoop

                
           
                end if


            end do

         
   
           if (fcentre - mean(1) .lt. detectparams%nstd(1) * sqrt(sigma(1))) then

!if (checkPixel) then
!write(*, *) "IsPointsource22 line 2070 xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "Rejecting source - fcentre=", fcentre, " mean(1)=", mean(1), " sigma(1)=", sqrt(sigma(1))
!end if



               cycle mLoop
           
            end if

            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5
            IsPointSource22 = .true.



!if (checkPixel) then
!write(*, *) "Source ispointsource22 ", i0, ", ", j0, " has been found!"
!            write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!            write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!            write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if





            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource22 = .false.
end Function IsPointSource22

!***********************************************************
! Function IsPointSource4
! For finding possible sources with just 1 or 2 bright pixels
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource4(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    integer(int16) :: i, il, iu, jl, ju
    real(single) :: median, sigma, fb, fs, ft, sn

    real(single) :: f(36)
    integer(int8) :: dx = 2
    integer :: count = 0

    dx = 4 / binX

    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)    
   
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)    


    count = 0
    if (ju .gt. j0) then  ! Top row     
        do i = il, iu
            count = count + 1
            f(count) = smoothedImagePtr(i, ju)
        end do
    end if

    if (jl .lt. j0) then  ! Bottom row     
        do i = il, iu
            count = count + 1
            f(count) = smoothedImagePtr(i, jl)
        end do
    end if


    if (il .lt. i0) then  ! Left side column     
        do i = jl + 1, ju - 1
            count = count + 1
            f(count) = smoothedImagePtr(il, i)
        end do
    end if


   
    if (iu .gt. i0) then  ! Right side column     
        do i = jl + 1, ju - 1
            count = count + 1
            f(count) = smoothedImagePtr(iu, i)
        end do
    end if



    !**************************************************
    ! Compute the median
    !**************************************************

    if (count .gt. 5) then
        call computeStatistics3(count, f(1 : 36), median, sigma)


    else
        IsPointSource4 = .false.
        return  
    end if


    dx = dx - 1

    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)    
   
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)    

    
    ! The total counts within this box
  
    ft = sum(smoothedImagePtr(il : iu, jl : ju))

   
    ! The total background counts within this box
  
    fb = real(count) * median


    fs = ft - fb
    if (fs .le. 0.0) then
        IsPointSource4 = .false.
        return
    end if 



    ! The signal-to-noise ratio



    sn = fs / sqrt(ft)

!if (checkPixel) then
!write(*, *) "Line 1680 background=", fb, " total=", ft, " source=", fs, " sn=", sn
!end if


    if (sn .gt. 2.0) then
        
        source%xPos = float(i0 + 1)
        source%yPos = float(j0 + 1)
       

!if (checkPixel) write(*, *) "Line 1493"

  
    else
        IsPointSource4 = .false.
    end if
end Function IsPointSource4



!***********************************************************
! Function IsPointSource4
! For finding possible sources with just 1 or 2 bright pixels
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource5(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre
    integer(int16) :: l, m
    
    real(single) :: xc, yc
   
  
    real(single) :: area1, area2, radius1, radius2
    real(single) :: counts1, counts2, fs, fb
    real(single) :: bgArea, constant, sn
 

    radius1 = real(2 / binX)
    radius2 = radius1 + real(2 / binx)

    area1 = PI * radius1 * radius1
    area2 = PI * radius2 * radius2
    bgArea = area2 - area1

    constant = area1 / bgArea
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 3
        xCentre = xc + xOffset(l)
        do m = 1, 3 
            yCentre = yc + yOffset(l)
            
            call GetCounts(smoothedimagePtr, xCentre, yCentre, radius1, counts1) 

!write(*, *) "Line 1772 counts1=", counts1

            ! Get the background using a larger aperture
            call GetCounts(smoothedimagePtr, xCentre, yCentre, radius2, counts2)

!write(*, *) "Line 1775 counts2=", counts2

            fb = (counts2 - counts1) * constant 
            fs = counts1 - fb
            sn = fs / sqrt(counts1)

!if (checkPixel) then
!write(*, *) "Line 1777 background=", fb, " total=", counts1, " source=", fs, " sn=", sn
!call fatal("stopped")
!end if


            if (sn .ge. 3.0) then
        
                source%xPos = xCentre + 0.5
                source%yPos = yCentre + 0.5
       

!                call ComputePointSourceParameters(sourceNumber, nx1, ny1, &
!                         smoothedimagePtr, backgroundImagePtr, &
!                         i0, j0, source)


                
                    IsPointSource5 = .true.
                return  
            
            end if

        end do
    end do
    IsPointSource5 = .false.

!  write(*, *) "IsPointSource5 line 1812"
end Function IsPointSource5

!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction1(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: image


    HeightFunction1 = max(minHeight, image(i, j) + nSigma * backgroundsigmaimagePtr(i, j)) ! Friday was 7.0

end Function  HeightFunction1

!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function ExtendedHeightFunction(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: image

    ExtendedHeightFunction = max(5.0, image(i, j) + nSigma * backgroundsigmaimagePtr(i, j))

!    ExtendedHeightFunction = max(4.0, image(i, j) + nSigma * backgroundsigmaimagePtr(i, j))

end Function  ExtendedHeightFunction


!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction11(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer :: image(:, :)

    HeightFunction11 = max(minHeight, image(i, j) + nSigma * sqrt(image(i, j))) ! Friday was 7.0


end Function  HeightFunction11


!*****************************************************
! Returns the minimum pixel value above the backround
! to be used in the source-detection
!
!
!****************************************************
real(single) Function HeightFunction2(i, j, height, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: height
    real(single), pointer :: image(:, :)

    HeightFunction2 = image(i, j) + height  ! Was 4.0

end Function  HeightFunction2


!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction3(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer :: image(:, :)


    HeightFunction3 = max(2.0, image(i, j) + nSigma * backgroundsigmaimagePtr(i, j)) ! Friday was 7.0


end Function  HeightFunction3


!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction4(i, j, nSigma, backgroundimage, backgroundsigmaimage)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage, backgroundsigmaImage


    HeightFunction4 = max(2.0, backgroundimage(i, j) + nSigma * backgroundsigmaimage(i, j)) ! Friday was 7.0

  
end Function  HeightFunction4

!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction5(i, j, nSigma, backgroundimage)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage


    HeightFunction5 = max(2.0, backgroundimage(i, j) + nSigma * sqrt(backgroundimage(i, j))) ! Friday was 7.0

 

end Function  HeightFunction5





!********************************************************************
! subroutine SetUpIndices
! 
!
!********************************************************************
subroutine SetUpIndices(image, tempmaskImage, iIndex, jIndex, k)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: tempmaskImage(:, :)
    integer(int16), pointer, dimension(:) :: iIndex, jIndex
    integer(int32), intent(out) :: k
    integer(int32), pointer, dimension(:) :: arrayIndex => null(), tempIndex => null()

    integer(int16) :: nx, ny, i, j
    integer(int32) ::  i1, j1

  
    real(single), pointer :: pixelValue(:) => null()

    nx = ubound(tempmaskImage, 1)
    ny = ubound(tempmaskImage, 2)


    k = count(.not. maskImage) 
    arrayIndex => allocateInt32(1, k, "arrayIndex")
    iIndex => allocateInt16(1, k, "iIndex")
    jIndex => allocateInt16(1, k, "jIndex")
    pixelValue => allocateSingle(1, k, "pixelValue")

  

!if (MaskImage(ipixel, jpixel)) then

!write(*, *) "omdetectdetection setupindices line 2466 tempmaskimage ipixel, jpixel is true"
!end if


    if (binx .eq. 2) then
        detectParams%minNoPixels = 1
    else
        detectParams%minNoPixels = 2
    end if
    !**********************************************************
    !
    !**********************************************************
    minHeight = 5.0
    k = 0 



! detectParams%minNoPixels = 0

    do j = lby, ny
        do i = lbx, nx
            if (.not. tempmaskImage(i, j)) then
!if (i .eq. ipixel .and. j .eq. jpixel) write(*, *) "setupindices line 2493"          

                !*********************************************************
                ! Check to see if have at least mincount neighbours
                !*********************************************************

!                call RecursiveSignalToNoise(i, j, number, image, backgroundImage, levelImage, count, sCounts, bCounts)

!if (i .eq. ipixel .and. j .eq. jpixel) then
!write(*, *) "setupindices line 2502"          

!write(*, *) "Number of neighbours=",  numberOfNeighbours(maskImage, i, j)

!end if


                if ( numberOfNeighbours(maskImage, i, j) .lt. detectParams%minNoPixels) cycle
         
                k = k + 1
                pixelValue(k) = image(i, j) 
                iIndex(k) = i
                jIndex(k) = j

            end if
        end do
    end do
    minHeight = 7.0

    ! Index sort the pixel values
    call IndexSort(pixelValue(1 : k), arrayIndex)


    deallocate(pixelValue)
    nullify(pixelValue)


    ! Reorder the arrays into order of decreasing pixel value
    tempIndex => allocateInt32(1, k, "tempIndex")
    tempIndex(1 : k) = iIndex(1 : k)
   
    do i1 = 1, k
        j1 = arrayIndex(i1)
         
        iIndex(k + 1 - i1) = tempIndex(j1)
    end do

    tempIndex(1 : k) = jIndex(1 : k)

    do i1 = 1, k
        j1 = arrayIndex(i1)    
        jIndex(k + 1 - i1) = tempIndex(j1)
    end do


    deallocate(arrayIndex, tempIndex)
    nullify(arrayIndex, tempIndex)


end subroutine SetUpIndices



!******************************************
! Initialises various things that are
! used in source detection routine
!*****************************************
subroutine InitialiseDetection(nx, ny,         &
                               radiusConstant, &
                               nCircle,        &
                               c1,             &
                               c2,             &
                               psfSemiaConstant, &
                               nPoints,        &
                               startAngle,     &
                               finishAngle)
    integer(int16), intent(in) :: nx, ny
    real(single), intent(in) :: radiusConstant
    integer, intent(in) :: nCircle
    real(single), intent(in) :: c1, c2
    real(single), intent(in), optional :: psfSemiaConstant
    integer, optional :: nPoints
    real(single), intent(in), optional :: startAngle, finishAngle
    integer(int8) :: i, j
    real(single) :: radius, angle, deltaAngle, theta1, theta2
    real(single) :: rCon


    detectParams%minSigNoise = realParameter("minsignificance")

    if (present(startAngle)) then
        theta1 = startAngle
        theta2 = finishAngle
    else
        theta1 = 0.0
        theta2 = 360.0
    end if
    
    if (binx .eq. 1) then
        detectParams%c2 = 1.5
  
    else
        detectParams%c2 = 1.0
    end if
   


    nx1 = nx - 1
    ny1 = ny - 1
    
    xLower = 0.1
    yLower = 0.1
    xUpper = real(nx) - 0.1 ! was 1.1
    yUpper = real(ny) - 0.1 ! was 1.1



    if (present(nPoints)) then
        detectParams%nSize = nPoints
    else 
        detectParams%nSize = 16
    end if 

!********************************************
! NB added on FRIDAY subtracted 1 for difficulut source
    detectParams%kMin = detectParams%nSize / 2 - 1

    detectParams%nCircles = nCircle 
    detectParams%nstd(1) = c1
    detectParams%nstd(2) = c2


    if (.not. associated(maskImage)) then
        allocate(maskImage(0 : nx1, 0 : ny1))
        maskImage = .false.
    end if    
  
    if (.not. associated(emaskImage)) then
        allocate(emaskImage(0 : nx1, 0 : ny1))
        emaskImage = .false.
    end if    
  


    if (associated(rCosTheta)) deallocate(rCosTheta)
    if (associated(rsinTheta)) deallocate(rsinTheta)
    if (associated(fCircle)) deallocate(fCircle)
    if (associated(mean)) deallocate(mean)
    if (associated(sigma)) deallocate(sigma)

    rCosTheta => Allocate2DSingle(1, detectParams%nSize, 1, detectParams%nCircles, "rCosTheta")
    rSinTheta => Allocate2DSingle(1, detectParams%nSize, 1, detectParams%nCircles, "rCosTheta")
    fCircle => Allocate2dSingle(1, detectParams%nSize, 1, detectParams%nCircles, "fCircle")
    mean => AllocateSingle(1, detectParams%nCircles, "mean")
    sigma => AllocateSingle(1, detectParams%nCircles, "sigma")


    deltaAngle = (theta2 - theta1) / real(detectParams%nSize)

    if (binx .eq. 1) then
        rcon = 1.5
    else 
        rcon = 1.0    
    end if

    do i = 1, detectParams%nCircles
   
        radius = radiusConstant * real(i) * rcon
   

        do j = 1, detectParams%nSize
            angle = (theta1 + real(j - 1) * deltaAngle) * degreesToRadians 
            rCosTheta(j, i) = radius * cos(angle)
            rSinTheta(j, i) = radius * sin(angle)
        end do
    end do


 

    detectParams%semiaSum = 0.0_double
    detectParams%n = 0
 
    detectParams%semiAConstant = psfSemiaConstant


    detectParams%semiAConstant = 1.25 ! WAS 1.25


end subroutine InitialiseDetection

!***********************************************************
!
!**********************************************************
subroutine CheckSource(xCentre, yCentre)
    real(single), intent(in) :: xCentre, yCentre
    real(single) :: xPos, yPos
    real(single) ::  radius, fCircle(detectParams%nSize)

    integer :: i, j, k

    real(single) :: dTheta, angle, mean(detectParams%nSize), sigma(detectParams%nSize), diff, error

    dTheta = 360.0 / real(detectParams%nSize)
write(*, *) "**************************************************"
write(*, *) "xcentre=", xcentre, " ycentre=", ycentre
    do i = 1, 8
   
         
        radius = real(i) * 0.5
        k = 0
        do j = 1, detectParams%nSize
            angle = real(j - 1) * dTheta * degreesToRadians 
         
            xPos = xCentre + radius * cos(angle)
            if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
            yPos = yCentre + radius * sin(angle)
                
            if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle   

            k = k + 1 
            fCircle(k) = Interpolate(smoothedimagePtr, xPos, yPos)
            
        end do
        if (k .lt. 8) cycle 
        call computeStatistics1(k, fCircle, mean(i), sigma(i))
        if (i .eq. 1) then
            write(*, 100) "Radius=", radius, " mean=", mean(i), " std=", sqrt(sigma(i))
        else
            diff = mean(i - 1) - mean(i)
            error = sqrt(sigma(i - 1) + sigma(i))
            write(*, 100) "Radius=", radius, " mean=", mean(i), " std=", sqrt(sigma(i)), " diff=", diff, " error=", error
        end if

    end do   
100 format(a, f6.2, 8(a, f10.2))
end subroutine CheckSource




!*********************************************************
! The following functions are for extended source detection
!*********************************************************


!****************************************************************
! Extended-source detection- case where no point-sources
! have been detected
!***************************************************************
subroutine DetectExtendedSources1(nx, ny, image,  & 
                        smoothedImage, backgroundIMage,           &
                        maxImage, elevelImage,            &
                        sourceList, nPSources, noSources, &
                        minNoPixels, maxNoPixels,         &
                        minCount, height, heightFunction, &
                        IsValidSource, extensionFlag1,    &
                        filterLevelImage)

    integer(int16), intent(in) :: nx, ny
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   


    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
    end interface

    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
            logical(bool) :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
 
    integer(int16), pointer :: elevelImage(:, :)
    integer(int16), pointer :: plevelImage(:, :) => null()



    plevelImage => allocate2dint16(lbx, nx1, lby, ny1, "pLevelImage")
    plevelImage = 0

    call DetectExtendedSources2(nx, ny, smoothedImage, & 
                        image, backgroundIMage, &
                        maxImage, plevelImage, elevelImage, &
                        sourceList, nPSources, noSources, &
                        minNoPixels, maxNoPixels, &
                        minCount, height, heightFunction, &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
   
    deallocate(pLevelImage)
    nullify(plevelImage)

end subroutine DetectExtendedSources1

!********************************************************
! Subroutine DetectExtendedSources
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources2(nx, ny, image,    & 
                        smoothedImage, backgroundIMage,             &
                        maxImage, levelImage1, levelImage2, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, height, heightFunction,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
    integer(int16), intent(in) :: nx, ny
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   
    logical, pointer :: usePixel(:, :) => null()
       
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical(bool) :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int16), pointer :: levelImage2(:, :)

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
 




    extensionFlag = extensionFlag1   
       

!    write(errstr,'(a, a)') "Looking for ", title
!    call message(AppMsg, Verbose)
    
    comment = "No of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1
    lbx = 0
    lby = 0
   
 
    ratio1 = 12 / binx   ! was 10 For test
    ratio2 = 300 / binX  ! was 500
          
    ratio2 = int(PI * 144.0) / binX  ! was 500
  

    dx = 12 / binX
    dx2 = 2 * dx 


    eLevelImagePtr => levelImage2  ! stores extended sources
    pLevelImagePtr => levelImage1  ! stores point sources
    


    imagePtr => image
    smoothedimagePtr => smoothedimage
    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source


    do j = lby, ny1
        do i = lbx, nx1


            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  HeightFunction(i, j, height, backgroundImage)) &
                                  elevelImagePtr(i, j) = number  
             
        end do
    end do 

    if (filterLevelImage) then
        call DilateImage(elevelImagePtr, number)
    end if
  
  

    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = -nosources    

    allocate(usePixel(0 : nx1, 0 : ny1))
    usePixel = .true. ! WAS TRUE
    do j = lby, ny1
 

        do i = lbx, nx1
            
            if (eLevelImagePtr(i, j) .ne. number) cycle
            
            usePixel(i, j) =  PixelOkay(i, j, minCount, number) 
            if (usePixel(i, j)) then

                count1 = count1 + 1

            end if
        end do
    end do


  
    where(.not. usePixel) elevelImagePtr = 0
    deallocate(usePixel)
    nullify(usePixel)

                      
    if (count1 .lt. minNoPixels) then
         
        return
        
    end if


    ! Set the levelImage pixels in the extended regions to the source
    ! number

                        
    call SetPixelNumbers(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 


    noSources = abs(minval(elevelImagePtr))
    nullify(pLevelImagePtr, elevelImagePtr)    

    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine DetectExtendedSources2

!********************************************************
! Subroutine DetectExtendedSources3
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources3(nx, ny, il, iu, jl, ju, image,    & 
                        smoothedImage, backgroundIMage,             &
                        maxImage, levelImage1, levelImage2, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, height, heightFunction,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
    integer(int16), intent(in) :: nx, ny
  
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   
    logical, pointer :: usePixel(:, :) => null()
       
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical(bool) :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int16), pointer :: levelImage2(:, :)

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
   

    extensionFlag = extensionFlag1   
       

!    write(errstr,'(a, a)') "Looking for ", title
!    call message(AppMsg, Verbose)
    
    comment = "No of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1
    lbx = 0
    lby = 0
   
    ratio1 = 12 / binx   ! For test
         
    ratio2 = int(PI * 144.0) / binX  ! was 500
  

    dx = 12 / binX
    dx2 = 2 * dx 


    eLevelImagePtr => levelImage2  ! stores extended sources
    pLevelImagePtr => levelImage1  ! stores point sources
    


    imagePtr => image
    smoothedimagePtr => smoothedimage
    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source




    do j = jl, ju
        do i = il, iu

            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  HeightFunction(i, j, height, backgroundImage)) &
                                  elevelImagePtr(i, j) = number  
             
        end do
    end do 



    if (filterLevelImage) then
        call DilateImage(elevelImagePtr(il : iu, jl : ju), number)
    end if
  
      
 

    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = minval(elevelimagePtr)    

    allocate(usePixel(il : iu, jl : ju))
    usePixel(il : iu, jl : ju) = .false.



    do j = jl, ju
 

        do i = il, iu
            
            if (eLevelImagePtr(i, j) .ne. number) cycle

            usePixel(i, j) =  PixelOkay(i, j, minCount, number)
            if (usePixel(i, j)) then

                count1 = count1 + 1
            else
                elevelImagePtr(i, j) = npsources         
        
            end if
          
        end do
    end do
 

  
    where(.not. usePixel(il : iu, jl : ju)                    &
          .and. elevelImagePtr(il : iu, jl : ju) .eq. number) &
          elevelImagePtr(il : iu, jl : ju) = 0

    deallocate(usePixel)
    nullify(usePixel)


    if (count1 .lt. minNoPixels) then

        noSources = abs(minval(elevelImagePtr))
 
        return
         
    end if


 

    ! Set the levelImage pixels in the extended regions to the source
    ! number

    call SetPixelNumbers(il, iu, jl, ju, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 

    noSources = abs(minval(elevelImagePtr))
    nullify(pLevelImagePtr, elevelImagePtr)    
    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine DetectExtendedSources3

!********************************************************
! Subroutine DetectExtendedSources4
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources4(image,    & 
                        smoothedImage, backgroundIMage,             &
                        levelImage1,&
                        sourceList, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        number1,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
     real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

     
    type(SourceLinkedList) :: sourceList
   
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int16), intent(in) :: number1 
   

    integer(int16) :: nx, ny
 
    
    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical(bool) :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int8) :: dx, dx2      
    integer(int32) :: count1


    nx = size(image, 1)
    ny = size(image, 2)


    extensionFlag = extensionFlag1   
       
    
    comment = "No of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1

   
    ratio1 = 20 / binX  ! NB was 20
    ratio1 = 10 / binx   ! For test
    ratio2 = 300 / binX  ! was 500
          
    ratio2 = int(PI * 144.0) / binX  ! was 500
  

    dx = 12 / binX
    dx2 = 2 * dx 

    eLevelImagePtr => levelImage1  ! stores extended sources



    if (.not. associated(pLevelImagePtr)) then
        pLevelImagePtr => allocate2dInt16(lbx, nx1, lby, ny1, "pLevelImagePtr")

    end if 
    



    imagePtr => image
    smoothedimagePtr => smoothedimage
    backgroundImagePtr => backgroundImage


      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
   


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source

    number = number1

    if (filterLevelImage) then
        call DilateImage(elevelImagePtr, number)
    end if
       
    sourceNumber = minval(elevelimagePtr)    

    count1 = count(elevelImagePtr .eq. number)
   
 
    if (count1 .lt. minNoPixels) then
         
        return
         
    end if


    ! Set the levelImage pixels in the extended regions to the source
    ! number

  

    call SetPixelNumbers1(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 



    noSources = abs(minval(elevelImagePtr))
   
    nullify(elevelImagePtr, imagePtr, smoothedImagePtr, backgroundImagePtr)

    deallocate(pLevelImagePtr)
    nullify(pLevelImagePtr)

end subroutine DetectExtendedSources4


!********************************************************
! Subroutine DetectExtendedSources
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources5(image,                    & 
                        smoothedImage, backgroundIMage,     &
                        maxImage, levelImage1, levelImage2, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, height, heightFunction,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
   
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   
    logical, pointer :: usePixel(:, :) => null()
       
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical(bool) :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int16), pointer :: levelImage2(:, :)

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
 




    extensionFlag = extensionFlag1   
       

!    write(errstr,'(a, a)') "Looking for ", title
!    call message(AppMsg, Verbose)
    
    comment = "No of likely " // trim(title) // " detected so far="


    nx1 = ubound(image, 1)
    ny1 = ubound(image, 2)
    lbx = lbound(image, 1)
    lby = lbound(image, 2)
   
 
    ratio1 = 12 / binx   ! was 10 For test
    ratio2 = 300 / binX  ! was 500
          
    ratio2 = int(PI * 144.0) / binX  ! was 500
  

    dx = 12 / binX
    dx2 = 2 * dx 


    eLevelImagePtr => levelImage2  ! stores extended sources
    pLevelImagePtr => levelImage1  ! stores point sources
    


    imagePtr => image
    smoothedimagePtr => smoothedimage

    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source


    do j = lby, ny1
        do i = lbx, nx1           
!            if (emaskImage(i, j)) cycle

            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  HeightFunction(i, j, height, backgroundImage)) &
                                  elevelImagePtr(i, j) = number  
             
        end do
    end do 

    if (filterLevelImage) then
        call DilateImage(elevelImagePtr, number)
    end if
  
  

    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = -nosources    

    allocate(usePixel(0 : nx1, 0 : ny1))
    usePixel = .true. ! WAS TRUE
    do j = lby, ny1
 

        do i = lbx, nx1
            
            if (eLevelImagePtr(i, j) .ne. number) cycle
            
            usePixel(i, j) =  PixelOkay(i, j, minCount, number) 
            if (usePixel(i, j)) then

                count1 = count1 + 1

            end if
        end do
    end do


  
    where(.not. usePixel) elevelImagePtr = 0
    deallocate(usePixel)
    nullify(usePixel)

                      

    if (count1 .lt. minNoPixels) then
         
        return
        
    end if


 

   
    backgroundImagePtr => backgroundImage
    ! Set the levelImage pixels in the extended regions to the source
    ! number


    imagePtr => image
    smoothedimagePtr => smoothedimage


    call SetPixelNumbers(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 


    noSources = abs(minval(elevelImagePtr))
    nullify(pLevelImagePtr, elevelImagePtr)    

    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine DetectExtendedSources5

!********************************************************
! Subroutine GrismDetectExtendedSources
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine GrismDetectExtendedSources(nx, ny, image,    & 
                        smoothedImage, backgroundIMage,             &
                        maxImage, elevelImage, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, isValidSource, extensionFlag1)
    integer(int16), intent(in) :: nx, ny
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
   
    logical, pointer :: usePixel(:, :) => null()
       

    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical(bool) :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    integer(int16), pointer :: elevelImage(:, :)
    

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
 




    extensionFlag = extensionFlag1   
       

    write(errstr,'(a, a)') "Looking for ", title
    call message(AppMsg, Verbose)
    
    comment = "No of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1
    lbx = 0
    lby = 0
   
 
    ratio1 = 12 / binx   ! was 10 For test
    ratio2 = 300 / binX  ! was 500
          
    ratio2 = int(PI * 144.0) / binX  ! was 500
  

    dx = 12 / binX
    dx2 = 2 * dx 


    eLevelImagePtr => elevelImage  ! stores extended sources
   
    


    imagePtr => image
    smoothedimagePtr => smoothedimage

    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source


    do j = lby, ny1
        do i = lbx, nx1


            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  backgroundImage(i, j)) elevelImagePtr(i, j) = number  
             
        end do
    end do 

  
 
    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = -nosources    

    allocate(usePixel(0 : nx1, 0 : ny1))
    usePixel = .true. ! WAS TRUE
    do j = lby, ny1
 

        do i = lbx, nx1
            
            if (eLevelImagePtr(i, j) .ne. number) cycle
            
            usePixel(i, j) =  PixelOkay(i, j, minCount, number) 
            if (usePixel(i, j)) then

                count1 = count1 + 1

            end if
        end do
    end do


  
    where(.not. usePixel) elevelImagePtr = 0
    deallocate(usePixel)
    nullify(usePixel)

                      

    if (count1 .lt. minNoPixels) then
         
        return
        
    end if


 

   
    backgroundImagePtr => backgroundImage
    ! Set the levelImage pixels in the extended regions to the source
    ! number


    imagePtr => image
    smoothedimagePtr => smoothedimage


    call SetPixelNumbers(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 


    noSources = abs(minval(elevelImagePtr))
    nullify(elevelImagePtr)    

    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine GrismDetectExtendedSources




!******************************************************************
! Set the levelImage pixels in the extended regions to the source
! number
!******************************************************************
subroutine SetPixelNumbers(istart, ifinish, jstart, jfinish, count1, minNoPixels, &
              maxNoPixels, sourceList, IsValidSource) 
     
    integer(int16), intent(in) :: istart, ifinish, jstart, jfinish
    integer(int32), intent(inout) :: count1
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    type(SourceLinkedList) :: sourceList

    interface
      
            function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
    
            logical(bool) :: status

        end function 

    end interface
    type(SourceParameters) :: srcParams
    integer(int16) :: i1, j1, noSources   
    integer(int32) :: i, j, il, il1, il2
    Type(PixelStruct), pointer :: pixel(:) => null()

    srcParams%extensionFlag = extensionFlag

    allocate(pixel(1 : count1))



    count1 = 0
 

    ! Fill the array of pixel strctures with information on the pixels
    do j1 = jstart, jfinish
        do i1 = istart, ifinish   
            if (elevelImagePtr(i1, j1) .eq. number) then
                count1 = count1 + 1
                pixel(count1)%x = i1
                pixel(count1)%y = j1
                pixel(count1)%counts = smoothedimagePtr(i1, j1)
                pixel(count1)%backgroundCounts = backgroundImagePtr(i1, j1)
                pixel(count1)%number = 0  ! Source number                         
                pixel(count1)%noNeighbours = NumberOfNeighbours(i1, j1, number)

            end if
        end do
    end do


    sourceNumber = minval(elevelImagePtr)
    noSources = -1
    il1 = 1
    il = 1 ! First source pixel

iLoop : do i = 1, count1 - 1

        do j = i + 1, count1
                    
            if (Neighbours(pixel(i), pixel(j))) then
               

                ! The first new extended source                                                                                                   
                noSources = 0                             
                sourceNumber = sourceNumber - 1                                      
                                
                pixel(i)%number = sourceNumber ! The first source - first pixel
                pixel(j)%number = sourceNumber ! First source - second pixel
                call MovePixel(i, j, pixel)    ! Added today- Pixel j moved to immediately after pixel i
                il = i                         ! The first pixel for this source       
                il1 = i + 1 
                exit iLoop     
                                   
            end if
        end do

    end do iLoop     


    il2 = il1
    ! Repeat the above process for remaining pixels
    do i = il1, count1 - 1

        do j = i + 1, count1


            if (pixel(i)%number .lt. 0 .and. pixel(j)%number .lt. 0) cycle
            if (Neighbours(pixel(i), pixel(j))) then
        
                if (pixel(i)%number .lt. 0) then
                    ! Pixel j is connected to pixel i
                    pixel(j)%number = pixel(i)%number

                    il1 = il1 + 1                         
                else


                    ! Both pixels i and j belong to an unnumbered source                    
                    !
                    ! Examine list of source pixels
                    call SetExtendedSource(sourceList, &
                        noSources, il, il1, minNoPixels, maxNoPixels, pixel, & 
                        srcParams, IsValidSource)

                    pixel(i)%number = sourceNumber
                    pixel(j)%number = sourceNumber
                    il = i       ! First pixel of new source
                    il1 = i + 1  ! Second pixel of new source
                    il2 = i + 1
   
                end if    

                ! Reposition pixel in pixel array
                ! to immediately after position i
                call MovePixel(i, j, pixel)
       
                   
            end if
        end do

    end do     



    ! Examine any remaining pixels
    if (il2 .lt. il1 .or. il .eq. 1) then


        call SetExtendedSource(sourceList, &
           noSources, il, il1, minNoPixels, maxNoPixels, pixel, &
           srcParams, IsValidSource)
                  
    end if

  
    deallocate(pixel)
    nullify(pixel)

    if (noSources == -1) noSources = 0    
!    write(errstr,'(a, i6)') "Total number found=", noSources
!    call message(AppMsg, Verbose)
                                 
                  
end subroutine SetPixelNumbers

!******************************************************************
! Set the levelImage pixels in the extended regions to the source
! number
!******************************************************************
subroutine SetPixelNumbers1(istart, ifinish, jstart, jfinish, count1, minNoPixels, &
              maxNoPixels, sourceList, IsValidSource) 
     
    integer(int16), intent(in) :: istart, ifinish, jstart, jfinish
    integer(int32), intent(inout) :: count1
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    type(SourceLinkedList) :: sourceList

    interface
      
            function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical(bool) :: status

        end function 

    end interface
    type(SourceParameters) :: srcParams
       
    integer(int32) :: i, j, il, il1, il2
    integer(int16) :: noSources
    Type(PixelStruct), pointer :: pixel(:) => null()


    allocate(pixel(1 : count1))



    count1 = 0


    ! Fill the array of pixel strctures with information on the pixels
    do j = jstart, jfinish
        do i = istart, ifinish   
            if (elevelImagePtr(i, j) .eq. number) then
                count1 = count1 + 1
                pixel(count1)%x = i
                pixel(count1)%y = j
               

            end if
        end do
    end do



    sourceNumber = minval(elevelImagePtr)
    noSources = -1
    il1 = 1
    il = 1 ! First source pixel

iLoop : do i = 1, count1 - 1

        do j = i + 1, count1
                    
            if (Neighbours(pixel(i), pixel(j))) then
               

                ! The first new extended source                                                                                                   
                noSources = 0                             
                sourceNumber = sourceNumber - 1                                      
                                      
                pixel(i)%number = sourceNumber ! The first source - first pixel
                pixel(j)%number = sourceNumber ! First source - second pixel
                call MovePixel(i, j, pixel)    ! Added today- Pixel j moved to immediately after pixel i
                il = i                         ! The first pixel for this source       
                il1 = i + 1 
                exit iLoop     
                                   
            end if
        end do

    end do iLoop     



    il2 = il1
    ! Repeat the above process for remaining pixels
    do i = il1, count1 - 1

        do j = i + 1, count1


            if (pixel(i)%number .lt. 0 .and. pixel(j)%number .lt. 0) cycle
            if (Neighbours(pixel(i), pixel(j))) then
        
                if (pixel(i)%number .lt. 0) then
                    ! Pixel j is connected to pixel i
                    pixel(j)%number = pixel(i)%number

                    il1 = il1 + 1                         
                else
                    ! Both pixels i and j belong to an unnumbered source                    
                    !
                    ! Examine list of source pixels
                    call SetExtendedSource1(sourceList, &
                        noSources, il, il1, minNoPixels, maxNoPixels, pixel, & 
                        srcParams, IsValidSource)
           
                    pixel(i)%number = sourceNumber
                    pixel(j)%number = sourceNumber
                    il = i       ! First pixel of new source
                    il1 = i + 1  ! Second pixel of new source
                    il2 = i + 1

   
                end if    

                ! Reposition pixel in pixel array
                ! to immediately after position i
                call MovePixel(i, j, pixel)
       
                   
            end if
        end do

    end do     


    ! Examine any remaining pixels
    if (il2 .lt. il1 .or. il .eq. 1) then


        call SetExtendedSource1(sourceList, &
           noSources, il, il1, minNoPixels, maxNoPixels, pixel, &
           srcParams, IsValidSource)
                  
    end if

  
    deallocate(pixel)
    nullify(pixel)

    if (noSources == -1) noSources = 0    
    write(errstr,'(a, i6)') "Total number found=", noSources
    call message(AppMsg, Verbose)
                                 
                  
end subroutine SetPixelNumbers1


!*****************************************
! Subroutine SetExtendedSource
!*****************************************
subroutine SetExtendedSource(sourceList,               &
                             noSources, il, iu,        &
                             minNoPixels, maxNoPixels, &
                             pixel, srcParams, IsValidSource)
    type(SourceLinkedList), intent(inout) :: sourceList

    integer(int16), intent(inout) :: noSources    
    integer(int32), intent(in) :: il, iu
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
        
    Type(PixelStruct), pointer :: pixel(:)
 
    type(SourceParameters), intent(inout) :: srcParams


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use Parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
    
            logical(bool) :: status

        end function 

    end interface
   
  
    if (AcceptExtendedSource(il, iu,  &
                 minNoPixels,         &
                 maxNoPixels,         &
                 srcParams)) then

!        pixel(il : iu)%number = sourceNumber
   
        ! Check if valid source  
        if ( IsValidSource(srcParams, pixel(il : iu)) ) then  ! Valid source
          
            
            ! Set the pixels on the level-image map
            call SetLevelImagePixels(il, iu, pixel)


                                          
            ! Add it to the source list
            srcParams%number = sourceNumber      
            srcParams%extensionFlag = extensionFlag


        
            srcParams%axialRatio = srcParams%semiA / srcParams%semiB
            srcParams%alpha = srcParams%theta

            ! Add the source to the linked-list
            call AddSourceToList(sourceList, srcParams)            

            noSources = noSources + 1          

!            if(len(trim(title)) .gt. 3) then
!                write(errstr,'(a, i5)') trim(comment), noSources
!                call message(AppMsg, Verbose)
!            end if
            sourceNumber = sourceNumber - 1  ! The next extended-source number

         end if   


    end if             
end subroutine SetExtendedSource

!*****************************************
! Subroutine SetExtendedSource
!*****************************************
subroutine SetExtendedSource1(sourceList,               &
                             noSources, il, iu,        &
                             minNoPixels, maxNoPixels, &
                             pixel, srcParams, IsValidSource)
    type(SourceLinkedList), intent(inout) :: sourceList

    integer(int16), intent(inout) :: noSources    
    integer(int32), intent(in) :: il, iu
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
        
    Type(PixelStruct), pointer :: pixel(:)
 
    type(SourceParameters), intent(inout) :: srcParams


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use Parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
    
            logical(bool) :: status

        end function 

    end interface
   

                         
    ! Cgeck to see if the number of pixels is within the allowed range 
    if (AcceptExtendedSource(il, iu,  &
                 minNoPixels,         &
                 maxNoPixels,         &
                 srcParams)) then
!        pixel(il : iu)%number = sourceNumber
   
      
        ! Check if valid source  
        if ( IsValidSource(srcParams, pixel(il : iu)) ) then  ! Valid source
            
 
            ! Set the pixels on the level-image map
            call SetLevelImagePixels(il, iu, pixel)
          
                                          
            ! Add it to the source list
            srcParams%number = sourceNumber      
            srcParams%extensionFlag = extensionFlag


            ! Add the source to the linked-list
            call AddSourceToList(sourceList, srcParams)            

            noSources = noSources + 1        ! The new number of sources        
           
            sourceNumber = sourceNumber - 1  ! The next extended-source number

         end if   


    end if             
end subroutine SetExtendedSource1

!*****************************************************
! Function AcceptExtendedSource
!****************************************************
logical function AcceptExtendedSource(il, iu, &
                 minNoPixels,         &
                 maxNoPixels,         &
                 srcParams)
    integer(int32), intent(in) :: il, iu
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
  
    type(SourceParameters), intent(inout) :: srcParams
   
   
    srcParams%noSourcePixels = iu - il + 1 ! The number of pixels associated with the source
 
    if (srcParams%noSourcePixels .lt. minNoPixels .or. &
             srcParams%noSourcePixels .gt. maxNoPixels) then

      
        AcceptExtendedSource = .false.
         
   
    else 
        AcceptExtendedSource = .true.
        
    end if

end function AcceptExtendedSource


!****************************************************
! Subroutine SetLevelImagePixels
!***************************************************
subroutine SetLevelImagePixels(il, iu, pixel)

    integer(int32), intent(in) :: il, iu
    Type(PixelStruct), pointer :: pixel(:)
    integer(int32) :: i
    

    do i = il, iu        
        
        elevelImagePtr(pixel(i)%x, pixel(i)%y) = pixel(i)%number    
        
        
    end do
end subroutine setLevelImagePixels

!******************************************************
! Returns the number of stars within an extended region
!
!*****************************************************
integer(int16) Function NumberOfStars1(pixel)
    type(PixelStruct) :: pixel(:)


    integer(int16) :: i
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
    integer :: x, y
   
    numberOfStars1 = 0
 
  
    ! Find the number of star pixels amongst the extended source pixels

    starNumber = -1
    starNumberList = 0

    do i = 1, size(pixel)
        x= pixel(i)%x
        y= pixel(i)%y

        if (plevelImagePtr(x, y) .gt. 0) then
            if (plevelImagePtr(x, y) .ne. starNumber) then
                starNumber = plevelImagePtr(x, y)
                if (any(starNumberList .eq. starNumber)) cycle  
                numberOfStars1 = numberOfStars1 + 1
                starNumberList(numberOfStars1) = starNumber
                if (numberOfStars1 .eq. 20) exit
            end if     
        end if
    end do 
           
 

end function NumberOfStars1

!******************************************************
!
! Returns the number of stars within an extended region
! and the location and maximum pixel image of that star
!
!*****************************************************
integer(int16) Function NumberOfStars2(pixel, maxStarPixel)
    type(PixelStruct) :: pixel(:)
    type(PixelStruct), intent(out) :: maxStarPixel
    

    integer(int16) :: i
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
    integer :: x, y
    numberOfStars2 = 0
 
  
    ! Find the number of star pixels amongst the extended source pixels

    starNumber = -1
    starNumberList = 0
    

    do i = 1, size(pixel)
        x= pixel(i)%x
        y= pixel(i)%y

        if (plevelImagePtr(x, y) .gt. 0) then
            if (imagePtr(x, y) .gt. maxStarPixel%counts) maxStarPixel = pixel(i)
            if (plevelImagePtr(x, y) .ne. starNumber) then
                starNumber = plevelImagePtr(x, y)
                if (any(starNumberList .eq. starNumber)) cycle  
                numberOfStars2 = numberOfStars2 + 1
                starNumberList(numberOfStars2) = starNumber
                if (numberOfStars2 .eq. 20) exit
            end if     
        end if
    end do 
           
 
end function NumberOfStars2
!******************************************************
!
! Returns the number of stars within an extended region
! and the location and maximum pixel image of that star
!
!*****************************************************
integer(int16) Function NumberOfStars3(pixel, starCounts)
    type(PixelStruct) :: pixel(:)
    real(double), intent(out) :: starCounts
    

    integer(int16) :: i
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
    integer :: x, y
    numberOfStars3 = 0
 
  
    ! Find the number of star pixels amongst the extended source pixels

    starCounts = 0.0_double
    starNumberList = 0
    

    do i = 1, size(pixel)
        x= pixel(i)%x
        y= pixel(i)%y

        if (plevelImagePtr(x, y) .gt. 0) then
            starCounts = starCounts + dble(imagePtr(x, y))
            if (plevelImagePtr(x, y) .ne. starNumber) then
                starNumber = plevelImagePtr(x, y)
                if (any(starNumberList .eq. starNumber)) cycle  
                numberOfStars3 = numberOfStars3 + 1
                starNumberList(numberOfStars3) = starNumber
                if (numberOfStars3 .eq. 20) exit
            end if     
        end if
    end do 
           
 
end function NumberOfStars3







!******************************************************
!
!
!*****************************************************
integer(int16) Function NumberOfHoles(pixel)
    type(PixelStruct) :: pixel(:)


    integer(int16) :: i, j
  
    numberOfHoles = 0
 

    do j = minval(pixel%y), maxval(pixel%y)
        

        do i = minval(pixel%x, pixel%y .eq. j), maxval(pixel%x, pixel%y .eq. j)

            if (eLevelImagePtr(i, j) .eq. 0) NumberOfHoles = NumberOfHoles + 1
          

        end do
    end do


end function NumberOfHoles




!***************************************************
! Function PixelOkay
! 
!**************************************************
logical function PixelOkay(i, j, minCount, number)

    integer(int16), intent(in) :: i, j 

    integer(int8), intent(in) :: minCount
    integer(int16), intent(in) :: number
    integer(int8)  :: count

    ! Local variables
    
    integer(int16) :: k, l, l1, l2
 

    count = 0
    l1 =  max(lby, j - 1_int16)
    l2 =  min(ny1, j + 1_int16)

    do k = max(lbx, i - 1_int16), min(nx1, i + 1_int16)
        do l = l1, l2
            if (k == i .and. l == j) cycle

            if (elevelImagePtr(k, l) .eq. number) then 
                count = count + 1
                
                
                if (count .ge. minCount) then
                    PixelOkay = .true.
                    return
                end if
          
            end if

        end do
    end do      

             
    PixelOkay = .false.

end function PixelOkay
!***************************************************
! Function NumberOfNeighbours1
! 
!**************************************************
function NumberOfNeighbours1(i, j, number) result(count)
    integer(int16), intent(in) :: i, j 
    integer(int16), intent(in) :: number
    integer(int8)  :: count
    ! Local variables
   
 
    count = 0

    if (i .gt. 0) then
        if (elevelImagePtr(i - 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (i .lt. nx1) then
        if (elevelImagePtr(i + 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .gt. 0) then
        if (elevelImagePtr(i, j - 1) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .lt. ny1) then
        if (elevelImagePtr(i, j + 1) .eq. number) then 
            count = count + 1
             
        end if
    end if


end function NumberOfNeighbours1
!***************************************************
! Function NumberOfNeighbours2
! 
!**************************************************
function NumberOfNeighbours2(levelimagePtr, i, j, number) result(count)
    integer(int16), pointer :: levelimagePtr(:, :)
    integer(int16), intent(in) :: i, j 
    integer(int16), intent(in) :: number 
  
    integer(int16) :: nx, ny

  
    integer(int8)  :: count
    ! Local variables
   
 
    count = 0
    nx = ubound(levelimagePtr, 1)
    ny = ubound(levelimagePtr, 2)

    if (i .gt. 0) then
        if (levelImagePtr(i - 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (i .lt. nx) then
        if (levelImagePtr(i + 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .gt. 0) then
        if (levelImagePtr(i, j - 1) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .lt. ny) then
        if (levelImagePtr(i, j + 1) .eq. number) then 
            count = count + 1
             
        end if
    end if


end function NumberOfNeighbours2

!***************************************************
! Function NumberOfNeighbours2
! 
!**************************************************
function NumberOfNeighbours3(maskImagePtr, i, j) result(count)
    logical(boolean), pointer :: maskimagePtr(:, :)
    integer(int16), intent(in) :: i, j 
  
  
    integer(int16) :: nx, ny
    integer(int16) :: il, iu, jl, ju, i1, j1
    
    integer(int8)  :: count
    ! Local variables
   
 
    count = 0
    nx = ubound(maskimagePtr, 1)
    ny = ubound(maskimagePtr, 2)

    il = max(0_int16, i - 1_int16)
    iu = min(nx, i + 1_int16)

    jl = max(0_int16, j - 1_int16)
    ju = min(ny, j + 1_int16)
 
    count = 0
    do j1 = jl, jl
        do i1 = il, iu

            if (i1 .eq. i .and. j1  .eq. j) cycle

            if (maskImagePtr(i1, j1)) cycle
             
            if ( imagePtr(i, j) .ge. HeightFunction1(i1, j1, 2.0, backgroundimagePtr)) then ! was 2.0
                  count = count + 1
            end if

        end do
    end do


end function NumberOfNeighbours3




!***********************************************
! Subroutine MovePixel - Inserts pixel(j), between
! pixels i and pixels i + 1
!**********************************************
subroutine MovePixel(i, j, pixel)
    integer(int32), intent(in) :: i, j
    Type(PixelStruct), pointer :: pixel(:)
    Type(PixelStruct) :: pixelJ
    integer(int32) :: k

    if (j .eq. i + 1) return  
    
    pixelJ = pixel(j)

    do k = j, i + 2, -1
        pixel(k) = pixel(k - 1)
    end do

    pixel(i + 1) = pixelJ 
      
end subroutine MovePixel
!***********************************************************
! Function Neighbours
!***********************************************************
logical function Neighbours(pixel1, pixel2)
    Type(PixelStruct), intent(in) :: pixel1, pixel2

    if (abs(pixel1%x - pixel2%x) .le. 1) then
        if(abs(pixel1%y - pixel2%y) .le. 1) then
            Neighbours = .true.
        else
            Neighbours = .false.

        end if
    else
        Neighbours = .false.
    end if
end function Neighbours


!*****************************************************
! IsExtendedSourceNoChecks1
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical(bool) Function IsExtendedSourceNoChecks1(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    integer(int32) :: noHoles

    noHoles = NumberOfHoles(pixel)
!write(*, *) "isextendedsourcenochecks no pixels=", size(pixel)
!write(*, *) "pixel(1)=", pixel(1)


    if ( (noHoles * 7) .lt. size(pixel)) then
!        write(*, *) "no holes=", noHoles, " rejecting source"
        IsExtendedSourceNoChecks1 = .false.
        return
    end if
 
    IsExtendedSourceNoChecks1 = .true.
    return 
  
    !******************************************************
    ! Check that centre is near 992 - x0
    !******************************************************
    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)

    
    IsExtendedSourceNoChecks1 = .true.

end function IsExtendedSourceNoChecks1

!*****************************************************
! IsExtendedSourceNoChecks2
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical(bool) Function IsExtendedSourceNoChecks2(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    integer :: noStars
   
   
    type(PixelStruct) :: maxStarPixel



    !**********************************************
    !
    !*********************************************
    if(maxval(pixel%counts) .lt. 1000.0) then
       IsExtendedSourceNoChecks2 = .false.
       return
     

    end if


    noStars = NumberOfStars(pixel, maxStarPixel)

    if (noStars .eq. 0 .or. noStars .gt. 2) then
        IsExtendedSourceNoChecks2 = .false.
        return
     

    end if


    !********************************************************
    ! Ignore star if peak counts < 20 sigma
    !*******************************************************
    if ( (maxStarPixel%counts - maxStarPixel%backgroundCounts) / &
             sqrt(maxStarPixel%counts) .lt. 20.0) then
        IsExtendedSourceNoChecks2 = .false.
        return
    end if
    

    IsExtendedSourceNoChecks2 = .true. ! One star
  
    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)
    if ( abs(source%imax - source%imin - source%jmax + source%jmin) .gt. 10) then
        IsExtendedSourceNoChecks2 = .false.
    else
        IsExtendedSourceNoChecks2 = .true.
    end if

   
end function IsExtendedSourceNoChecks2


!*****************************************************
! IsSmokeRing
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical(bool) Function IsSmokeRing(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
  
    integer(int16) :: i, j, il, iu, jl, ju, i0, j0, di, n1, n2
    real(single) :: radius, dx, dy
    

    !*******************************************
    ! Compute the source parameters
    !*******************************************
    call ComputeExtendedSourceParameters(sourceNumber, &
                                         pixel,        &
                                         source)
   

    if (source%semia / source%semib .gt. 1.2) then
        isSmokeRing = .false.
        return
    end if

    !*******************************************
    ! Check that there is a hole at the centre
    !*******************************************

    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)

    di = 4 / binx
    il = max(i0 - di, source%imin)
    iu = min(i0 + di, source%imax)

    jl = max(j0 - di, source%jmin)
    ju = min(j0 + di, source%jmax)





    radius = real(di)
    n1 = 0
    n2 = 0
    do j = jl, ju
        dy = source%ypos - real(j + 1)
        do i = il, iu
            dx = source%xpos - real(i + 1)
            if (sqrt(dx * dx + dy * dy) .gt. radius) cycle
            if (elevelimageptr(i, j) .ne. 0) then
                n1 = n1 + 1
            end if
            n2 = n2 + 1
           
        end do
    end do 
    
    if (real(n1) / real(n2) .gt. 0.75) then
        IsSmokeRing = .false.
    else
        IsSmokeRing = .true.
    end if 
   


end function IsSmokeRing


!*****************************************************
! IsValidRubbishSource
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical(bool) Function IsValidRubbishSource(srcParams, pixel)
     
  

    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:) 
    
  
    srcParams%iMin = minval(pixel%x)
    srcParams%iMax = maxval(pixel%x)
    srcParams%jMin = minval(pixel%y)
    srcParams%jMax = maxval(pixel%y)


    if (areaOkay(srcParams, Pixel, 3_int8)) then 
        IsValidRubbishSource  = .false.
    else
        IsValidRubbishSource  = .true.
 

    end if

    
end function IsValidRubbishSource

!*****************************************************
! IsValidExtendedSource
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical(bool) Function IsValidExtendedSource(srcParams, pixel)
     
  

    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:) 
    
    integer(int16)  :: noHoles, noNeighbours, noStars
    integer(int16) :: width, height 

    integer(int16), parameter :: ratio = 7
  
    integer(int16) :: i0, j0, n, il, iu, jl, ju

!    logical :: checkPixel
    real(double) :: starCounts


    if (maxval(pixel%noNeighbours) .le. 6 / binx) then
        IsValidextendedsource  = .false.
        return
    end if




!checkPixel = .false.
!if (any(pixel%x .eq. ipixel .and. pixel%y .eq. jpixel)) checkPixel = .true.

!if (checkPixel) then
!write(*, *) "Line 4715 srcparams%nossource pixels =", srcParams%noSourcepixels
!end if
    !*******************************************
    ! Compute the source parameters
    !*******************************************
    call ComputeExtendedSourceParameters(sourceNumber, &
                                         pixel,        &
                                         srcParams)


!if (checkPixel) then
!write(*, *) "Line 4716 imin=", srcParams%imin, " imax=", srcParams%imax
!write(*, *) "Line 4717 jmin=", srcParams%jmin, " jmax=", srcParams%jmax
!write(*, *) "number=", srcParams%number, " no of pixels=", size(pixel)
!write(*, *) "srcparams%nossource pixels =", srcParams%noSourcepixels
!end if


    if (.not. areaOkay(srcParams, Pixel, 5_int8)) then 
        IsValidextendedsource  = .false.
         return
     end if

    


    i0 = max(lbx, int(srcParams%xpos - 0.5, int16))
    j0 = max(lby, int(srcParams%ypos - 0.5, int16))
    il = max(lbx, i0 - 5_int16)
    iu = min(nx1, i0 + 5_int16)
    jl = max(lby, j0 - 5_int16)
    ju = min(ny1, j0 + 5_int16)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
        IsValidExtendedSource = .false.

        return
         
    end if
    



    n = size(pixel)
    if (n .lt. 2 * count(pLevelImagePtr(srcParams%imin : srcParams%imax, srcParams%jmin : srcParams%jMax) .gt. 0)) then 
        IsValidExtendedSource = .false.

        return
         
    end if
    


!    if (count(pixel%x .eq. i0 .and. pixel%y .eq. j0) .eq. 0) then

!        IsValidExtendedSource = .false.

!        return
         
!    end if
    


!if (checkPixel) then
!write(*, *) "4777 srcparams%noSourcePixels =", srcParams%noSourcepixels

!write(*, *) "Line 4777 source number = ", srcParams%number
!write(*, *) "x=", srcParams%xpos, " y=", srcParams%yPos
!write(*, *) "Pixels-"
!do i = 1, size(pixel)
!write(*, *) i, pixel(i)%x, " ",  pixel(i)%y
!end do
!end if
    ! Check here to see if might be just mod-8 pattern
   
    width = srcParams%iMax -  srcParams%iMin
    height = srcParams%jMax - srcParams%jMin


    if (width .gt. ratio * height .or. height .gt. ratio * width) then 
        
        IsValidExtendedSource = .false.

        return
         
    end if


    
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel, starCounts)

!if (checkpixel) then
!write(*, *) "Line 4804 source number = ", srcParams%number
!write(*, *) "x=", srcParams%xpos, " y=", srcParams%yPos
!write(*, *) "No of pixels =", srcParams%noSourcepixels
!write(*, *) "total counts=", sum(pixel%counts), " starcounts=", starcounts

!end if


!    if (sum(pixel%counts) .lt. starCounts * 1.5) then


!        IsValidExtendedSource  = .false.
!        return
!    end if


    noNeighbours = sum(pixel%noNeighbours)

    if (noNeighbours  .lt. srcParams%noSourcePixels * 2) then 

        IsValidExtendedSource  = .false.
        return
    end if



!if (checkpixel) then
!write(*, *) "Line 4831 source number = ", srcParams%number
!write(*, *) "x=", srcParams%xpos, " y=", srcParams%yPos
!write(*, *) "No of pixels =", srcParams%noSourcePixels
!write(*, *) "noHoles=", noHoles, " ratio1=", ratio1


!end if

    if (maxval(imagePtr(srcParams%imin : srcParams%imax, srcParams%jmin : srcParams%jmax)) .gt. 2000.0) then
      


        if (srcParams%noSourcePixels .lt. noHoles * ratio1) then    ! was 200             
!if (checkpixel) then

!write(*, *) "Rejecting source at line 4846 nosourcepixels=", srcParams%noSourcePixels

!end if




            IsValidExtendedSource = .false.
            return
        end if 
    else 



        if (real(srcParams%noSourcePixels) .lt. real(noHoles * 3)) then    ! was 3             

!if (checkpixel) then

!write(*, *) "Rejecting source at line 4864 nosourcepixels=", srcParams%noSourcePixels

!end if



            IsValidExtendedSource = .false.
            return
        end if  
    end if
    

           

!*******************************************
! MONDAY BLANKED OUT
!******************************************
!if (checkPixel) then
!write(*, *) "Line 4864 noHoles=", noHoles, " nosourcepixels=", srcParams%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2, " nostars=", nostars
!end if
    if (srcParams%noSourcePixels .lt. noStars * ratio2) then              ! was 50
!write(*, *) "Line 3818 rejecting source, noStars=", noStars, " ratio2=", ratio2
        IsValidExtendedSource = .false.           
        return
    end if



    if (srcParams%semiA / srcParams%semiB .gt. 10) then
        IsValidExtendedSource = .false.
  
    else
                       
        IsValidExtendedSource = .true.
!if (checkPixel) then
!write(*, *) "new source no=", srcParams%number, " x=", srcParams%xpos, " y=", srcParams%yPos
!end if
    end if


!write(*, *) "Line 1494 noHoles=", noHoles, " nosourcepixels=", srcParams%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

  
end Function IsValidExtendedSource

!*****************************************************
! IsValidExtendedSource
! Checks to see if the parameters of an extended split
! source lie within certain limits
!*****************************************************
logical(bool) Function IsValidSplitSource(srcParams, pixel)
     
  

    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:) 
    
    integer(int16)  :: noHoles, noNeighbours, noStars
    integer(int16) :: width, height 

    integer(int16), parameter :: ratio = 7
    integer(int16), parameter :: dx = 5
    integer(int16) :: i0, j0, il, iu, jl, ju
    integer i, n, k
    


  

    if (sum(pixel%counts, pixel%counts .ge. 0.0) .lt. minimumCounts) then
        IsValidSplitSource = .false.

        return
         
    end if

    n = size(pixel)



    k = 0

    do i = 1, n
        k = k + count(plevelImagePtr( max(lbx, pixel(i)%x - 1_int16) : min(nx1, pixel(i)%x + 1_int16), &     
                                      max(lby, pixel(i)%y - 1_int16) : min(ny1, pixel(i)%y + 1_int16)) &
                      .gt. 0)
            
    end do


    if (k .gt. n / 4) then

        IsValidSplitSource = .false.

        return
         
    end if
    
 

    !*******************************************
    ! Compute the source parameters
    !*******************************************
    call ComputeExtendedSourceParameters(sourceNumber, &
                                         pixel,        &
                                         srcParams)
   

    if (.not. areaOkay(srcParams, Pixel, 5_int8)) then 
        IsValidSplitSource  = .false.
        return
    end if


    i0 = max(lbx, int(srcParams%xpos - 0.5, int16))
    j0 = max(lby, int(srcParams%ypos - 0.5, int16))


    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
        IsValidSplitSource = .false.

        return
         
    end if
  
  


  
   
    if (maxval(imagePtr(il : iu, jl : ju)) .gt. maxval(pixel%counts)) then
        IsValidSplitSource = .false.

        return
         
    end if
    
  
    if (count(pixel%x .eq. i0 .and. pixel%y .eq. j0) .eq. 0) then

        IsValidSplitSource = .false.

        return
         
    end if
    

!write(*, *) "Line 5031 isvalidsplitsource source number = ", srcParams%number
!write(*, *) "x=", srcParams%xpos, " y=", srcParams%yPos
!write(*, *) "imin=", srcParams%imin, " imax=", srcParams%imax
!write(*, *) "jmin=", srcParams%jmin, " jmax=", srcParams%jmax


!write(*, *) "Pixels-"
!do i = 1, size(pixel)
!write(*, *) i, pixel(i)%x, " ",  pixel(i)%y
!end do

    ! Check here to see if might be just mod-8 pattern
   
    width = srcParams%iMax -  srcParams%iMin
    height = srcParams%jMax - srcParams%jMin


    if (width .gt. ratio * height .or. height .gt. ratio * width) then 
        
        IsValidSplitSource = .false.

        return
         
    end if
    
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel)

    noNeighbours = sum(pixel%noNeighbours)
    if (noNeighbours  .lt. srcParams%noSourcePixels * 2) then 

        IsValidSplitSource  = .false.
        return
    end if




    if (srcParams%noSourcePixels .lt. noHoles * ratio1) then    ! was 200             

!write(*, *) "Rejecting source"
!write(*, *) "Line 3800 noHoles=", noHoles, " nosourcepixels=", srcParams%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

        IsValidSplitSource = .false.
        return
    end if  

           


!write(*, *) "Line 1494 noHoles=", noHoles, " nosourcepixels=", srcParams%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

    if (srcParams%noSourcePixels .lt. noStars * ratio2) then              ! was 50
!write(*, *) "Line 3818 rejecting source, noStars=", noStars, " ratio2=", ratio2
        IsValidSplitSource = .false.           
        return
    end if



    if (srcParams%semiA / srcParams%semiB .gt. 10) then
        IsValidSplitSource = .false.
  
    else
                       
        IsValidSplitSource = .true.



!write(*, *) "new source no=", srcParams%number, " x=", srcParams%xpos, " y=", srcParams%yPos

    end if

!write(*, *) "Line 1494 noHoles=", noHoles, " nosourcepixels=", srcParams%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

  
end Function IsValidSplitSource

!*****************************************************
! IsValidExtendedSource
! Checks to see if the parameters of an extended split
! source lie within certain limits
!*****************************************************
logical(bool) Function IsValidSplitGrismSource(srcParams, pixel)
     
  

    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:) 
    
 

    integer(int16), parameter :: ratio = 7
    integer(int16), parameter :: dx = 2
    integer(int16) :: i0, j0, il, iu, jl, ju
    integer n
    


  

    if (sum(pixel%counts, pixel%counts .ge. 0.0) .lt. minimumCounts) then
        IsValidSplitGrismSource = .false.

        return
         
    end if


    if (.not. IsValidFirstOrderSource(srcParams, pixel)) then
        IsValidSplitGrismSource = .false.
        return
    end if


    n = size(pixel)



    if (.not. areaOkay(srcParams, Pixel, 5_int8)) then 
        IsValidSplitGrismSource  = .false.
        return
    end if


    i0 = max(lbx, int(srcParams%xpos - 0.5, int16))
    j0 = max(lby, int(srcParams%ypos - 0.5, int16))


    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
        IsValidSplitGrismSource = .false.

        return
         
    end if
  
  
   
  
    if (count(pixel%x .eq. i0 .and. pixel%y .eq. j0) .eq. 0) then

        IsValidSplitGrismSource = .false.

        

    else         
        IsValidSplitGrismSource = .true.

    end if
    

  
end Function IsValidSplitGrismSource

!***********************************************************************
!
!**********************************************************************
logical(bool) Function areaOkay(source, pixel, constant)

    type(SourceParameters), intent(in) :: source
    type(PixelStruct) :: pixel(:) 
    integer(int8), intent(in) :: constant
    integer :: area     

    area = (source%imax - source%imin + 1) * (source%jmax - source%jmin + 1) 
 
    if (( size(pixel) * constant) .lt. area) then ! Was 5 
        areaOkay = .false.

    else

        areaOkay = .true.
    end if

    
end function areaOkay
!*****************************************************
! IsValidSmallExtendedSource
! For small extended sources
!*****************************************************
logical(bool) Function IsValidSmallExtendedSource(source, pixel)
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    
    real(single), parameter :: delta1 = 1.1, delta2 = 5  ! was 1.2
    integer(int16) :: noHoles, i0, j0, il, iu, jl, ju, noStars
    real(single) :: ratio  
    type(PixelStruct) :: maxStarPixel

!    logical :: check
  


!check = .false.
! if (any(pixel%x .eq. ipixel) .and. any(pixel%y .eq. jpixel)) then
!write(*, *) "is valid smallextendedsoure Checking possible smallextended source at ", &
!ipixel, ", ", jpixel, " number of pixels=", size(pixel)
!check = .true.

!end if 

    if (maxval(pixel%noNeighbours) .lt. 4) then
        IsValidsmallextendedsource  = .false.
        return
    end if
!if (check) write(*, *) "Line 5537"

    if (.not. widthOkay(pixel)) then
        IsValidsmallextendedsource  = .false.

!if (check) then

!write(*, *) "isvalidsmallextebdedsource - Line 5284 rejecting source pixel width not okay"

!end if




        return
    end if

  
    if (size(pixel) .lt. 80 / binx) then
        if (.not.PixelBackgroundOkay(pixel)) then

!if (check) then

!write(*, *) "isvalidsmallextebdedsource - Line 5548 rejecting source pixel background not okay"

!end if


            IsValidsmallextendedsource  = .false.
            return
    end if
    end if
    ! Compute the source parameters
    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)

 
  



   

    if (.not. areaOkay(source, Pixel, 5_int8)) then 
        IsValidsmallextendedsource  = .false.
        return
    end if
!if (check) write(*, *) "Line 5567"

  


    i0 = max(lbx, int(source%xpos - 0.5, int16))
    j0 = max(lby, int(source%ypos - 0.5, int16))


    il = max(lbx, i0 - 5_int16)
    iu = min(nx1, i0 + 5_int16)
    jl = max(lby, j0 - 5_int16)
    ju = min(ny1, j0 + 5_int16)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
        IsValidSmallExtendedSource = .false.

        return
         
    end if
!    if (check) write(*, *) "Line 5588"

   
    ! Check here to see if might be just mod-8 pattern
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel, maxStarPixel)

!if (check) then

!write(*, *) "isvalidsmallextendedsourceLine 5220 source no ", source%number, " xpos=", source%xpos, " y=", source%ypos
!write(*, *) "no pixels=", source%noSourcePixels, " no holes=", noHoles, " nostars=", nostars, " ratio1=", ratio1
!write(*, *) "a=", source%semiA, " b=", source%semiB
!end if

    if (noStars .gt. 0) then

        IsValidSmallExtendedSource = .false.           
        return
    end if
!if (check) write(*, *) "Line 5577, noholes=", noholes, " size pixel=", size(pixel) 
      
    if ((noHoles * 2) .gt. size(pixel)) then
        IsValidSmallExtendedSource = .false.           
        return
    end if 
!if (check) write(*, *) "Line 5583" 

    if (source%semiA .lt. 1.0) then
        IsValidSmallExtendedSource = .false.           
        return
    end if
 
!if (check) write(*, *) "Line 5589" 
    ratio = source%semiA / source%semiB 
    
    if (ratio .gt. delta2) then
        IsValidSmallExtendedSource = .false.           

        return
    end if 
        

    !*************************************************
    ! Reach here then new extended source
    !************************************************
    IsValidSmallExtendedSource = .true.           
    
!if (check) then

!write(*, *) "New source - IsValidSmallextendedsource- source no=", source%number
!write(*, *) "num pixerls=", source%noSourcePixels
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) " noholes=", noholes

!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 5353 displaying levelimage.fits")


!end if 
end Function IsValidSmallExtendedSource

!*****************************************************
! IsValidESource3
! For small point sources
!*****************************************************
logical(bool) Function IsValidESource3(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
  
    real(single), parameter :: delta = 5.0   ! was 1.5
    real(single) :: ratio
    integer(int16) :: j, n, noHoles, noStars, i(1), i0, j0, noNeighbours
   
    integer(int8) :: minSize = 10
  
!    logical :: check

    n = size(pixel)




    if (n .gt. 10) then
        if (maxval(pixel%noNeighbours) .lt. 4) then
            IsValidESource3  = .false.
            return
        end if
    end if


!    if ( count( (pixel%counts - pixel%backgroundCounts) .ge. minheight) .lt. 2) then
!        IsValidESource3 = .false.
!        return
!    end if





    
!    if (any(pixel%x .eq. ipixel) .and. any(pixel%y .eq. jpixel)) then
!        check = .true.
!    else
!        check = .false.
!    end if




    if (.not. widthOkay(pixel)) then
        IsValidEsource3  = .false.
        return
    end if



    i = maxloc(pixel%counts)

    i0 = pixel(i(1))%x  ! X coordinate of maximum
    j0 = pixel(i(1))%y  ! Y coordinare of maximum





    if (sum(pixel%counts - pixel%backgroundcounts) / sqrt(sum(pixel%counts)) .lt. detectParams%minSigNoise) then
 
!       if (check) write(*, *) "esource3 rejecting at line 5312"

        IsValidESource3 = .false.
     
    end if



!if (check) write(*, *) "esource3 xchecking pixel ", ipixel, " , ", jpixel, " in isvalidesource3 at line 4760" 

  

    if (n .lt. 100 / binx) then
        if (.not. BackgroundGradientOkay(i0, j0, pixel)) then
            IsValidESource3 = .false.
!if (check) write(*, *) "esource3 rejecting at line 4764"
            return
            end if
    end if

!if (check) write(*, *) "isvalidesource3 line 4774"


    do j = 1, n
        if ( any(elevelImagePtr( max(lbx, pixel(j)%x - 1_int16) : min(nx1, pixel(j)%x + 1_int16), &     
                                 max(lby, pixel(j)%y - 1_int16) : min(ny1, pixel(j)%y + 1_int16)) &
             .eq. 1)) then
!if (check) then
!   write(*, *) "isvalidesource3 line 5126 rejecting "

!write(*, *) "offending pixel=", pixel(j)%x, ", ", pixel(j)%y
!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 5353 displaying levelimage.fits")

!end if
      
            IsValidESource3  = .false.
            return
        end if
     
        if ( count (elevelImagePtr( max(lbx, pixel(j)%x - 1_int16) : min(nx1, pixel(j)%x + 1_int16), &     
                                 max(lby, pixel(j)%y - 1_int16) : min(ny1, pixel(j)%y + 1_int16)) &
             .lt. 0) .gt. 1) then

!if (check) then


!write(*, *) "Number of pixel .lt. 0=", count(elevelImagePtr( max(lbx, pixel(j)%x - 1_int16) : min(nx1, pixel(j)%x + 1_int16), &     
!                                 max(lby, pixel(j)%y - 1_int16) : min(ny1, pixel(j)%y + 1_int16)) .lt.0 )

!write(*, *) "isvalidesource3 line 5136 rejecting at pixel ", pixel(j)%x, " , ", pixel(j)%y
!write(*, *) "nx1=", nx1, " ny1=", ny1, " lbx=", lbx, " lby=", lby

!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 5353 displaying levelimage.fits")
!end if
      

            IsValidESource3  = .false.



            return
        end if
     




        if ( any(plevelImagePtr( max(lbx, pixel(j)%x - 1_int16) : min(nx1, pixel(j)%x + 1_int16), &     
                                 max(lby, pixel(j)%y - 1_int16) : min(ny1, pixel(j)%y + 1_int16)) &
             .ne. 0)) then
!      if (check) then
!write(*, *) "isvalidesource3 line 5371 rejecting "
!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 5373 displaying levelimage.fits")
!end if

            IsValidESource3  = .false.
            return
        end if
    end do

  
!if (check) write(*, *) "isvalidesource3 line 4815"

    !*********************************************************************
    ! For testing
    !********************************************************************
    if (n .lt. 18 / binx) then  
    ! Check background
        if (.not.PixelBackgroundOkay(pixel)) then

!if (check) then

!write(*, *) "ESOURCE3 - Line 5709 i0=", i0, " j0=", j0, " rejecting source pixel background not okay"

!end if


            IsValidESource3  = .false.
            return
        end if

    end if

 
    noNeighbours = sum(pixel%noNeighbours)

    if (n .gt. 10) then
       
        if (noNeighbours .lt. int(1.5 * real(n))) then

!if (check) then

!write(*, *) "ESOURCE3 - Line 5338 rejecting source n=", n, " no neighbours=", noNeighbours
!end if


            IsValidESource3  = .false.
            return
        end if
    end if

 
    ! Check here to see if might be just mod-8 pattern
  
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel)

                
!if (check) then
!write(*, *) "Line 5363, no pixels=", n, " noholes=", noHoles
!end if
 

    if (noStars .gt. 0 .or. n .lt. noHoles) then              ! was 3


!if (check) then
!write(*, *) "Line 5372 rejecting source"
!end if
 

        IsValidESource3 = .false.           
        return
    end if

    if (n .lt. minSize) then


        source%xPos = real(i0 + 1) 
        source%yPos = real(j0 + 1)
        
        source%extensionFlag = 0
!write(*, *) "Line 2135, no pixels=", size(pixel) 
       call ComputePointSourceParameters(int(source%number, int16), nx1, ny1, &
                         imagePtr, backgroundImagePtr, &
                         i0, j0, source)
!  write(*, *) "Line 2578 no source pixels=", source%noSourcePixels

    else
        source%extensionFlag = 1

!write(*, *) "Line 2581, no pixels=", size(pixel) 
    
        call ComputeExtendedSourceParameters(int(source%number, int16), &  ! The source number
                                             pixel,        &  ! The pixels for this source
                                             source) ! The computed moments
!   write(*, *) "Line 2587 no source pixels=", source%noSourcePixels




    end if

    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)

!if (check) then
!write(*, *) "esource3 line 5717 x=", source%xpos, " y=", source%ypos
!write(*, *) "imin=", source%imin, " imax=", source%imax, " jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "no pixels=", size(pixel)
!end if 

    if (.not. areaOkay(source, Pixel, 3_int8)) then 
        IsValidEsource3  = .false.
        return
    end if

    

!if (check) then
! write(*, *) "Line 5421 Source ", source%number, " x=", source%xpos, " y=", source%ypos, " a=", source%semia, " b=", source%semib
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "Number of pixels=", n
!write(*, *) "lbound pixer=", lbound(pixel), " ubound=", ubound(pixel)

!end if

    if (n .gt. 3) then
        if (source%imin .eq. source%imax .or. source%jmin .eq. source%jmax) then

!if (check) then
!write(*, *) "Line 5441 rejectin source"
!end if
 

            IsValidESource3 = .false.
            return 
        end if
    end if

!if (check) then

! write(*, *) "Line 5475 Source ", source%number, " x=", source%xpos, " y=", source%ypos, " a=", source%semia, " b=", source%semib
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "Number of pixels=", n
!write(*, *) "lbound pixer=", lbound(pixel), " ubound=", ubound(pixel)
!write(*, *) "no neighbours=", noNeighbours

!end if



!    write(*, *) "Source ", source%number, " x=", source%xpos, " y=", source%ypos, " a=", source%semia, " b=", source%semib

    ratio = source%semia / source%semiB

    if (ratio .gt. delta) then

!if (check) then
!write(*, *) "Line 5507 rejecti8n source"
!end if
        IsValidESource3 = .false.
        return
    end if           

    if (ratio .gt. 1.5) then
        source%extensionFlag = 1
     
    end if     

    IsValidESource3 = .true.           


!if (check) then
!write(*, *) "Line 5531source ", ipixel, " , ", jpixel, " has been found - no=", source%number
!write(*, *) "extension flag=", extensionflag, " no of pixels=", size(pixel)
!write(*, *) "sourcenumber=", sourcenumber




!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 5790 displaying levelimage.fits")

!end if


 
end Function IsValidESource3
!**************************************************
! Check that no surrounding pixel is greater than the
! source pixel
!
!**************************************************
logical Function PixelBackgroundOkay(pixel)
    type(PixelStruct), intent(in) :: pixel(:)

    integer :: il, iu, jl, ju, i, j, k
    integer(int16) ::  dx = 2
    integer :: n
    real(single) :: maxSignalToNoise
    logical :: check            


!    PixelBackgroundOkay = .false.
!    return
   
    n = size(pixel)
!    write(*, *) "Line 5711 n=", n
    if (n .lt. max(4, 6 / binx)) then
        maxSignalToNoise = maxval((pixel%counts - pixel%backgroundCounts) / sqrt(pixel%counts))
        if (maxSignalToNoise .gt. 4.0) then
            check = .false.
        else
            check = .true.
        end if
    else
        check = .false.
    end if
 
    dx = 2 / binx  ! was 4 / binx

!write(*, *) "Line 5724"

    do i = 1, n
        il = max(lbx, pixel(i)%x - dx)
        iu = min(pixel(i)%x + dx, nx1)
        jl = max(lby, pixel(i)%y - dx)
        ju = min(pixel(i)%y + dx, ny1)
       
        do k = jl, ju
            do j = il, iu
                if (j .eq. pixel(i)%x .and. k .eq. pixel(i)%y) cycle
                if (elevelImagePtr(j, k) .lt. 0) then
                    PixelBackgroundOkay = .false.
!write(*, *) "Line 5734"
!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 5740 displaying levelimage.fits")

                    return
                end if
                if (plevelImagePtr(j, k) .gt. 0) then
                    PixelBackgroundOkay = .false.
!write(*, *) "Line 5739"

                    return
                end if
            
                if (.not. check) cycle
                if (elevelImagePtr(j, k) .eq. 0 .or. elevelImagePtr(j, k) .eq. 1) then ! Was == 1
                 
                   
                    if (pixel(i)%counts .lt. imagePtr(j, k) + max(1.0, 0.25 * sqrt(imagePtr(j, k)))) then ! was 1.0
                   
 
                    
                        PixelBackgroundOkay = .false.
!write(*, *) "Line 5754 pixel%counts=", pixel(i)%counts, " image=", imagePtr(j, k), " n=", n

                        return
                    end if
                end if
            end do
        end do
    end do 
    PixelBackgroundOkay = .true.

   

end Function PixelBackgroundOkay
   
!**************************************************
!
!
!**************************************************
logical Function PixelBackgroundOkay1(pixel)
    type(PixelStruct), intent(in) :: pixel

    integer :: il, iu, jl, ju, i, j
    integer :: n
    real(double) :: counts
    integer(int16), parameter :: dx = 2

    il = max(lbx, pixel%x - dx)
    iu = min(pixel%x + dx, nx1)
    jl = max(lby, pixel%y - dx)
    ju = min(pixel%y + dx, ny1)
   
    
    counts = 0.0_double
    n = 0
    do j = jl, ju
        do i = il, iu
            if (i .eq. pixel%x .and. j .eq. pixel%y) cycle
            if (elevelImagePtr(i, j) .ne. 0) cycle
            
            n = n + 1
            counts = counts + dble(smoothedimagePtr(i, j))
            
        end do
    end do 
    if (n .eq. 0) then
        PixelBackgroundOkay1 = .true.
        return
    end if

    counts = counts / dble(n)

    if (pixel%counts .ge. (max(5.0_double, counts + 2.0 * sqrt(counts)))) then
        PixelBackgroundOkay1 = .true.
    else
        PixelBackgroundOkay1 = .false.
    end if

   

end Function PixelBackgroundOkay1
   
!**************************************************
!
!
!**************************************************
logical Function SignalToNoiseOkay(pixel, nSigma)
    type(PixelStruct) :: pixel(:)
    real(single), intent(in) :: nSigma

    integer :: il1, iu1, jl1, ju1, il, iu, jl, ju, i, j
    integer :: n, n1
   

    integer(int16) :: dx = 3
    integer, parameter :: maxCounts = 10000
    
    real(single) :: median, sigma, sigNoise, minCount
    real(single) :: sourceCounts, backgroundCounts, totalCounts, pcount(maxCounts)
    
    dx = 12 / binx
    
    il = max(lbx, minval(pixel%x) - dx)
    iu = min(maxval(pixel%x) + dx, nx1)
    jl = max(lby, minval(pixel%y) - dx)
    ju = min(maxval(pixel%y) + dx, ny1)
   
    dx = 6 / binx

    il1 = max(lbx, minval(pixel%x) - dx)
    iu1 = min(maxval(pixel%x) + dx, nx1)
    jl1 = max(lby, minval(pixel%y) - dx)
    ju1 = min(maxval(pixel%y) + dx, ny1)
  

!write(*, *) "Signal to noise check line 4204"    
    pcount = 0.0
    n = 0

 iLoop :   do j = jl, ju
        do i = il, iu
            if (i .ge. il1 .and. i .le. iu1 .and. &
                j .ge. jl1 .and. j .le. ju1) cycle 
            if (elevelImagePtr(i, j) .ne. 0) cycle
                   
            n = n + 1
     
            pcount(n) = smoothedimagePtr(i, j)
            if (n .eq. maxCounts) exit iLoop
        end do
    end do iLoop




    if (n .lt. 3) then

        SignalToNoiseOkay = .false.
        return
    end if


     
!write(*, *) "n=", n, " minval(pcount)=", minval(pcount(1 : n)), " max=", maxval(pcount(1 : n))


    call computeStatistics3(n, pcount(1 : n), median, sigma)
    sigma = sqrt(sigma)



!write(*, *) "median from background mode=", median, " sigma=", sigma



    
    minCount = (maxval(pixel%counts) - median) * 0.5 ! was 0.5

    n = size(pixel)

    n1 = count(pixel%counts .lt. max(5.0, (median + nSigma * sqrt(median))))

!write(*, *) "Line 4304 n=", n, " n1=", n1


if (n .lt. 4 .and. n1 .gt. 0) then

        SignalToNoiseOkay = .false.
        return



else  if (n1 .gt. n / 3 ) then

!write(*, *) "Line 4308 signaltonoise is false"
        SignalToNoiseOkay = .false.
        return
    end if
        

    if (sigma .gt. max(1.0, 2.0 * sqrt(median))) then
        SignalToNoiseOkay = .false.
        return
    end if



!    if ( count((pcount(1 : n) - median) .ge. minCount ) .ge. n / 3) then


!write(*, *) "Line 4327"

 if ( any((pcount(1 : n) - median) .ge. minCount )) then


!       write(*, *) "number of counts - median ge. mincount=", count((pcount(1 : n) - median) .ge. minCount)
!        SignalToNoiseOkay = .false.
!        return
    end if

    
!write(*, *) "Line 4344"

    totalCounts = sum(pixel%counts)
    backgroundCounts = real(n) * median

    sourceCounts = totalCounts - backgroundCounts

    if (sourceCounts .le. 0.0) then
!write(*, *) "Line 3915 source counts=", sourceCounts
        SignalToNoiseOkay = .false.
        return
    end if

    sigNoise = sourceCounts / sqrt(totalCounts)

    if (sigNoise .ge. 2.0) then
        SignalToNoiseOkay = .true.
    else
!write(*, *) "Line 3925 sigNoise=", sigNoise, " sourcecounts=", sourceCounts, &
!            " total counts=", totalCounts, " backgroundcounts=", backgroundCounts

        SignalToNoiseOkay = .false.


    end if

end Function SignalToNoiseOkay
   
!*****************************************************
! IsValidPointSource2
! For small extended sources
!*****************************************************
logical(bool) Function IsValidPointSource2(srcParams, pixel)
     

    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:)
    
    real(single), parameter :: delta1 = 1.5, delta2 = 5.0  ! was 1.2
    real(single) :: ratio
    integer :: noHoles, noStars

    ! Check here to see if might be just mod-8 pattern


 
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel)
               
    if (noStars .gt. 0) then              ! was 50
        IsValidPointSource2 = .false.           
        return
    end if

    if (srcParams%semiA .lt. 1.0) then
        IsValidPointSource2 = .false.           
        return
    end if

    if (srcParams%semiB .lt. 1.0) then
        IsValidPointSource2 = .false.           
        return
    end if


    if (srcParams%semiA .gt. 2.0) then
        IsValidPointSource2 = .false.           
        return
    end if


    ratio = srcParams%semiA / srcParams%semiB 
    
    if (ratio .ge. delta1) then
 
        IsValidPointSource2 = .true.           
        return 
    end if
    srcParams%extensionFlag = 0

    
    IsValidPointSource2 = .true.           
  

end Function IsValidPointSource2

!*****************************************************
! NoChecks
! For fast-chain
!*****************************************************
logical(bool) Function NoChecks(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
   
 
    integer(int16) :: i(1), i0, j0 

     i = maxloc(pixel%counts)
     i0 = min(nx1, pixel(i(1))%x)
     j0 = min(ny1, pixel(i(1))%y)

     source%xPos = real(i0 + 1) 
     source%yPos = real(j0 + 1)
        
     source%extensionFlag = 0

     call ComputePointSourceParameters(int(source%number, int16), nx1, ny1, &
                         smoothedimagePtr, backgroundImagePtr, &
                         i0, j0, source)
    
    
    NoChecks = .true.

end Function NoChecks

!*****************************************************
! IsValidGrismSource
! Checks to see if the parameters of a point
! source lie within certain limits
!*****************************************************
logical(bool) Function IsValidGrismSource(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
 
     
    ! Compute the source parameters
    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)
   
     
  
    ! Check here to see if might be just mod-8 pattern
    if (source%semiA .lt. 3.0) then
        
        IsValidGrismSource = .false.
    else
        IsValidGrismSource = .true.
    endif

end Function IsValidGrismSource


!*****************************************************
! IsValidExtendedGrismSource1
! For zeroth-order spectrum
!*****************************************************

logical(bool) Function IsValidZeroOrderSource(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
    integer(int16) :: i, il, iu, jl, ju, i0(1), j0, minY, maxY
    integer(int16) :: dx = 10, dx1      
   

    real(single), parameter :: delta = 10.0


    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source) ! The computed moments
   
 


    if (source%semia < 2.0) then
        IsValidZeroOrderSource = .false.
        return
    end if    
  

    if (source%semiA / source%semiB > delta) then
        IsValidZeroOrderSource = .false.
        return
    end if

  

    if (abs(abs(source%theta) - 90.0) > 30.0) then
        IsValidZeroOrderSource = .false.
        return
    end if


    if (size(pixel) .gt. 400 / binx) then
        IsValidZeroOrderSource = .true.
        return
    end if




   
    ! Check no source immediately below

    dx = 40 / binx
    dx1 = 4 / binx

    j0 = minval(pixel%y)
    i0 = maxloc(pixel%counts, mask=pixel%y .eq. j0)
    i0(1) = pixel(i0(1))%x    
    il = max(0_int16, i0(1) - dx1)
    iu = min(nx1, i0(1) + dx1)
    ju = j0
    jl = max(0_int16, j0 - dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
       IsValidZeroOrderSource = .false.
       return
    end if

    ! Check no source immediately above


    j0 = maxval(pixel%y)
    i0 = maxloc(pixel%counts, pixel%y .eq. j0)
    i0(1) = pixel(i0(1))%x    

    il = max(0_int16, i0(1) - dx1)
    iu = min(nx1, i0(1) + dx1)
    jl = j0
    ju = min(ny1, j0 + dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
       IsValidZeroOrderSource = .false.
       return
   
    end if
!  IsValidZeroOrderSource = .true.
!return

    minY = minval(pixel%y)
   
    maxY = maxval(pixel%y)

    jl = minY + 2
    ju = maxY - 2

    if (jl .gt. ju) then
       IsValidZeroOrderSource = .true.
       return
    end if
       
    do i = jl, ju

        il = minval(pixel%x, pixel%y .eq. i) - 1
        if (il .gt. 0) then
            if (imagePtr(il, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
        end if
        il = il - 1
        if (il .gt. 0) then
            if (imagePtr(il, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
        end if


        iu = maxval(pixel%x, pixel%y .eq. i) + 1
        if (iu .le. nx1) then
            if (imagePtr(iu, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
    
        end if
        iu = iu + 1
        if (iu .le. nx1) then
            if (imagePtr(iu, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
   

        end if

    end do

    IsValidZeroOrderSource = .true.


end Function IsValidZeroOrderSource

!*****************************************************
! IsValidExtendedGrismSource2 -
! for first-order spectra
!*****************************************************
logical(bool) Function IsValidFirstOrderSource(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
    integer(int16) :: minxc, maxxc
    integer(int16) :: i0, j0
    integer(int16) :: dx = 5


    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)
 
    if (abs(abs(source%theta) - 90.0) > 2.0) then
        IsValidFirstOrderSource = .false.

        return
    end if

 
    j0 = nint(source%ypos - 0.5)
    if (j0 .lt. 0) then
        j0 = 0
    else 
        if (j0 .gt. ny1) j0 = ny1
    end if
    
    i0 = nint(source%xpos - 0.5)
    if (i0 .lt. 0) then
        i0 = 0
    else 
        if (i0 .gt. nx1) i0 = nx1
    end if
  
    dx = 6 / binx
    minxc = minval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    minxc = min(minxc, i0 - dx)
    maxxc = maxval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    maxxc = max(maxxc, i0 + dx)


    where(pixel%x .lt. minxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where

    where(pixel%x .gt. maxxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where

    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)


    i0 = nint(source%xpos - 0.5)
    if (i0 .lt. 0) then
        i0 = 0
    else 
        if (i0 .gt. nx1) i0 = nx1
    end if
  

    j0 = nint(source%ypos - 0.5)
    if (j0 .lt. 0) then
        j0 = 0
    else 
        if (j0 .gt. ny1) j0 = ny1
    end if


    minxc = minval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    minxc = min(minxc, i0 - dx)
    maxxc = maxval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    maxxc = max(maxxc, i0 + dx)

    where(pixel%x .lt. minxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where


    where(pixel%x .gt. maxxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where


    if (abs(abs(source%theta) - 90.0) > 2.0) then
        IsValidFirstOrderSource = .false.

        return
    end if


    if (source%semia < 10.0 .or. source%semiA / source%semiB < 10) then
        IsValidFirstOrderSource = .false.

    else
       
       IsValidFirstOrderSource = .true.
    end if    



  
end Function IsValidFirstOrderSource

!************************************************************
!
!
!************************************************************
subroutine GetPSfParameters

    real(single) :: xCentre, yCentre, xPos, yPos
 
    integer :: i, j, k
    real(single), pointer :: backgroundImage(:, :) => null()
    real(single), pointer :: tempImage(:, :) => null()


    type(SourceParameters) :: source
   

    fCircle = 0.0
    xCentre = size(psfParams%psfImage, 1) / 2.0
    yCentre = size(psfParams%psfImage, 2) / 2.0

    do i = 1, detectParams%nCircles
        k = 0
        do j = 1, detectParams%nSize
            xPos = xCentre + rCosTheta(j, i)
            if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
            yPos = yCentre + rSinTheta(j, i)
                
            if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle  
            k = k + 1
            fCircle(k, i) = Interpolate(psfParams%psfImage, xPos, yPos)
        end do    
        call computeStatistics3(k, fCircle(1 : detectParams%nSize , i), mean(i), sigma(i))

        if (i .gt. 1) then
            psfParams%psfRatio(i - 1) = mean(i - 1) / mean(i)
            
        end if
    end do

 !   fwhm = ComputeFWHM(psfParams%psfimage, xcentre, ycentre)

 

    psfParams%psfRatio(1) = mean(1) / mean(2)
  
    psfParams%psfRatio(2) = mean(1) / mean(3)
 
    psfParams%psfRatio = psfParams%psfRatio * 0.5



    !******************************************************************************
    !
    !******************************************************************************


    backgroundimage => allocate2dsingle(0, ubound(psfparams%psfimage, 1), 0, ubound(psfparams%psfimage, 2), "backgroundimage")
    backgroundimage = 10.0


    tempimage => allocate2dsingle(0, ubound(psfparams%psfimage, 1), 0, ubound(psfparams%psfimage, 2), "backgroundimage")
    tempimage = backgroundimage + psfparams%psfimage * 100.0
    



    source%xpos = size(psfparams%psfImage, 1) / 2.0 + 0.5
  
    source%ypos = size(psfparams%psfImage, 2) / 2.0 + 0.5
  


    deallocate(tempimage, backgroundImage)
    nullify(tempimage, backgroundImage)


end subroutine GetPSFParameters


!*****************************************************************************
! Compute the FWHM of the source at the given position 
!
!*****************************************************************************
real(single) function ComputeFWHM(image, xpos, ypos)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: xPos, yPos
    real(single) :: radius
    real(single) :: f1, f2
    real(double) :: rMin, rMax
    integer :: i    

    FWHMImagePtr => image

    fcentre = Interpolate(FWHMimagePtr, xPos, yPos)
    psfXcentre = xpos
    psfycentre = ypos

    fcentre = fcentre / 2.0
  
    f1 = GridInterpolate(dble(xPos + 0.1), dble(yPos), psfGridFit)
    f2 = GridInterpolate(dble(xPos + 2.0), dble(yPos), psfGridFit)

    rmin = 0.25
    f1 = Interpolate(FWHMimagePtr, real(xpos + rmin), real(yPos))


    do i = 1, 50
        rmax = rmin + real(i) * 0.25
        f2 = Interpolate(FWHMimagePtr, real(xPos + rmax), real(yPos))
        if ((f1 - fcentre) / (f2 - fcentre) .lt. 0.0) exit
    end do 

    rmax = rmax + 0.25

    rmin = 0.5 / real(binx)
    rmax = real(4 / binx)
    radius = ZeroIn(FWHMFunction, rmin, rmax, 0.0_double, 0.0_double)
   



!    write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre + radius), real(psfyCentre))
!  write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre - radius), real(psfyCentre))
!  write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre), real(psfyCentre + radius))
!write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre), real(psfyCentre - radius))

 
    ComputeFWHM = 2.0 * radius

    write(errstr,'(a, f6.2, a, f12.8)') "Computed FWHM=",  (2.0 * radius), " Function value=", FWHMFunction(dble(radius))
    call message(AppMsg, Verbose)
 

    nullify(FWHMImagePtr)

end function ComputeFWHM

!*****************************************************************************
! Function FWHMFunction
!
!*****************************************************************************
real(double) Function FWHMFunction(radius)
    real(double), intent(in) :: radius
    real(double) :: xPos, yPos, theta, dtheta
    real(single) :: xlOwer, xUpper, yLower, yUpper
    integer :: i, k
    real(double) :: mean

    dtheta = TWO_PI / 36.0

    k = 0

    xLower = 1.0
    yLower = 1.0
    xUpper = size(FWHMimagePtr, 1) - 1.0
    yUpper = size(FWHMimagePtr, 2) - 1.0

    mean = 0.0_double
    do i = 1, 36
        theta = dtheta * dble(i)
        xPos = psfxCentre + radius * cos(theta)
        if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
        yPos = psfyCentre + radius * sin(theta)
                
        if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle  
        k = k + 1
        mean = mean + dble(Interpolate(FWHMimagePtr, real(xPos), real(yPos)))

        
    end do

    if (k .gt. 10) then

        mean = mean / dble(k)

        FWHMFunction = fCentre - mean
    else
        FWHMFunction = -1.0_double

    end if

!write(*, *) "k=", k, " radius=", radius, " mean=", mean, " FWHMFunction=", FWHMFunction  

end Function FWHMFunction

!***********************************************
!
!**********************************************
subroutine MinimumList(n, data, nsize)
    integer, intent(inout) :: n
    real(single) :: data(1: n)

    integer, intent(in) :: nsize
    real(single) :: tempdata(1: n)

    real(double) :: minimum
    integer :: i, j, k, kmin
    real(double) :: fmin
   
    fmin = huge(1.0)

    kmin = 1
    do i = 1, n
        k = i
        minimum = 0.0_double
        do j = 1, nsize
            minimum = minimum + dble(data(k))
            k = k + 1
            if ( k .gt. n) k = 1
        end do
   
        if (minimum .lt. fmin) then
            fmin = minimum
            kmin = i
        end if
 
    end do


    do j = 1, nsize

        tempdata(j) = data(kmin)
        kmin = kmin + 1
        if ( kmin .gt. n) kmin = 1

 
    end do
    data = tempdata    
    n = nsize

end subroutine MinimumList    

!****************************************************************************
! For removing extended light round bright sources (mod-8 patterns, etc
!
!***************************************************************************  
subroutine BrightSourceMask(npSources, iStart, pSourceList, image, backgroundimage)
    integer(int16), intent(in) :: npSources, iStart
    type(SourceLinkedList) :: pSourceList  ! The linked-list point source-list
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    integer(int16) :: il, iu, jl , ju, di
      
    type(sourceParameters), pointer :: pSourceData(:) => null()
    integer(int16) :: i, j, k 

     
    real(single) :: height, radius, maxRadius1, maxradius2, dx, dy  

!real(single), pointer :: tempLevelImagePtr(:, :)
    if (npSources .eq. 0) return



    height = 1.5

    ! Load the point-source data
    psourceData => SourceDataArray(psourceList, npSources)


    di = 24 / binx ! was 30   
    maxRadius1 = real(di)
    
    di = 40 / binx   

!di = 30 / binx

    maxRadius2 = real(di)


    do i = iStart, npSources

      
        if (.not. btest(pSourceData(i)%qualityFlag, 4)) cycle



        !*****************************************************************
        ! Check each bright source
        !**************************************************************** 
        il = max(lbx, pSourceData(i)%imin - di)
        iu = min(nx1, pSourceData(i)%imax + di)
        jl = max(lby, pSourceData(i)%jmin - di)
        ju = min(ny1, pSourceData(i)%jmax + di)
     
        do k = jl, ju
            dy = pSourceData(i)%ypos - real(k + 1)
       

            dx = pSourceData(i)%xpos - real(j + 1)
            do j = il, iu

                if( image(j, k) .lt. HeightFunction5(j, k, height, backgroundImage)) cycle
          
               
                dx = pSourceData(i)%xpos - real(j + 1)
      
                radius = sqrt(dx * dx + dy * dy)
                if (radius .gt. maxRadius2) cycle
           
       
                emaskimage(j, k) = .true.
           
                if (radius .le. maxradius1) maskImage(j, k) = .true. ! Don't use these pixels in source detection
       
            end do
        end do
      
    end do
  
    deallocate(pSourceData)
    nullify(pSourceData)


!allocate(templevelimageptr(0 : ubound(backgroundimage, 1), 0 : ubound(backgroundimage, 2)))
!templevelimageptr= 0
!where(emaskimage) templevelimagePtr = 1
!call saveimage(backgroundimage, nx, ny, "backgroundimage.fits", "testimage")
!call saveimage(templevelimagePtr, nx, ny, "templevelimage.fits", "testimage")
!call fatal("stopped omdetection line 8499 displaying templevelimage.fits")


end subroutine BrightSourceMask


!******************************************************************
! Looks for a mod-8 pattern around a source
!******************************************************************
logical Function IsMod8Source(source, image, i1out, i2out, j1out, j2out)
    type(sourceParameters), intent(inout) :: source
    real(single), pointer :: image(:, :) 
    integer(int16), intent(out), optional :: i1out, i2out, j1out, j2out
    integer(int16) :: i1, i2, j1, j2, k, nx, ny, i0, j0, il, iu, jl, ju
    integer :: i11, i21, j11, j21
    integer :: iPlus1, iPlus2
    integer :: jPlus1, jPlus2
    integer :: iMinus1, iMinus2
    integer :: jMinus1, jMinus2

    integer(int16) :: dx, dx1, n, n1, n2
    integer(int16) :: k1, k2
    real(single) :: fmax
    integer(int8) :: kmax = 4



    nx = ubound(image, 1)
    ny = ubound(image, 2)

    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)

    dx = 40 / binx
    dx1 = 16 / binx
    il = max(lbx, i0 - dx)
    iu = min(nx, i0 + dx)


    !****************************************************************
    ! Check both sides of the source
    !****************************************************************
    jl = max(lby, j0 - dx1)
    ju = min(ny, j0 + dx1)

!    isMod8Source = .false. 
!return

    fmax = maxval(image(il : iu, jl : ju))
    if (fmax .lt. 1000000.0 / detectParams%exposure) then
        isMod8Source = .false.
       
        return


  elseif(fmax .gt. 7.5 * detectParams%exposure) then ! WAS 7.5

        isMod8Source = .true. 
        source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag
  
        return

    end if


    
    n = nint(0.6 * real(ju - jl + 1))

    kmax = 8 / binx 

    k1 = 0
    i11 = i0
    j11 = jl



    !****************************************************************
    ! Checking left-hand side of the source for a mod-8 pattern
    !***************************************************************
 iLoop1  : do i1 = i0 - 10 / binx, il, -1
       
        iMinus1 = i1 - 1
        if (iminus1 .lt. 0) exit               
        n1 = count(image(i1 : i1, jl : ju) .gt. image(iMinus1 : iMinus1, jl : ju) + &
                  sqrt(image(iMinus1 : iMinus1, jl : ju)))

        do k = 1, kmax
            iminus1 = i1 - k
            iminus2 = iminus1 - 1
            if (iminus2 .lt. 0) exit iLoop1
            n2 = count(image(iMinus1 : iMinus1, jl : ju) .lt. image(iMinus2 : iMinus2, jl : ju) - &
                     sqrt(image(iMinus1 : iMinus1, jl : ju) ))
            if (n1 .gt. n .and. n2 .gt. n) then
                k1 = k1 + 1
                i11 = i1
                exit iLoop1
            end if
        end do
    end do iLoop1

    i21 = i0
  
    !****************************************************************
    ! Checking right-hand side of the source for a mod-8 pattern
    !***************************************************************

iLoop2 : do i2 = i0 + 10 / binx, iu   
        iPlus1 = i2 + 1
        if (iplus1 .gt. nx) exit       
         
        n1 = count(image(i2 : i2, jl : ju) .gt. image(iPlus1 : iPlus1, jl : ju) +  &
                 sqrt( image(iPlus1 : iPlus1, jl : ju)))
        do k = 1, kmax
            iplus1 = i2 + k
            iplus2 = iplus1 + 1
            if (iplus2 .gt. nx) exit iLoop2
     
            n2 = count(image(iPlus1 : iPlus1, jl : ju) .lt. image(iPlus2 : iPlus2, jl : ju)  - &
                     sqrt(image(iPlus1 : iPlus1, jl : ju)))
       
            if (n1 .gt. n .and. n2 .gt. n) then
                k1 = k1 + 1
                i21 = i2
                exit iLoop2
            end if
        end do        
    end do iLoop2
      

    if (k1 .eq. 2) then

!write(*, *) "Line 6192 mod 8 source x=", source%xpos, " y=", source%ypos
!write(*, *) "Line 6193 have set bit 4"



        source%qualityFlag = ibset(source%qualityFlag, 4)  ! Set bit 4 of quality flag
        IsMod8Source = .true.
        return
    end if



    !****************************************************************
    ! Check below and above the source
    !****************************************************************
    il = max(lbx, i0 - dx1)
    iu = min(nx, i0 + dx1)


    jl = max(lby, j0 - dx)
    ju = min(ny, j0 + dx)
    n = nint(0.6 * real(iu - il + 1))
    k2 = 0
    j21 = j0

jLoop1 : do j1 = j0 + 10 / binx, ju
        !****************************************************************
        ! Checking above the source for a mod-8 pattern
        !***************************************************************
    
        jPlus1 = j1 + 1
        if (jplus1 .gt. ny) exit
      
         
        n1 = count(image(il : iu, j1 : j1) .gt. image(il : iu, jPlus1 : jPlus1) + sqrt(image(il : iu, jPlus1 : jPlus1)))
        do k = 1, kmax
            jplus1 = j1 + k
            jplus2 = jplus1 + 1
            if (jplus2 .gt. ny) exit jLoop1
            n2 = count(image(il : iu, jPlus1 : jPlus1) .lt. image(il : iu, jPlus2 : jPlus2) - &
                  sqrt(image(il : iu, jPlus1 : jPlus1)))
           

            if (n1 .gt. n .and. n2 .gt. n) then
                k2 = k2 + 1
                j21 = j1
                exit jLoop1
            end if
        end do
            
    end do jLoop1


    j21 = j0
    
    !****************************************************************
    ! Checking below the source for a mod-8 pattern
    !***************************************************************
    
jLoop2 : do j2 = j0 - 10 / binx, jl, -1
        jMinus1 = j2 - 1
        if (jMinus1 .lt. 0) exit     
       
         
        n1 = count(image(il : iu, j2 : j2) .gt. image(il : iu, jMinus1 : jMinus1) + sqrt(image(il : iu, jMinus1 : jMinus1))) 
        do k = 1, kmax
            jminus1 = j2 - k
            jminus2 = jminus1 - 1
            if (jminus2 .lt. 0) exit jLoop2
      
            n2 = count(image(il : iu, jMinus1 : jMinus1) .lt. image(il : iu, jMinus2 : jMinus2) - &
                             sqrt(image(il : iu, jMinus1 : jMinus1)))
         


            if (n1 .gt. n .and. n2 .gt. n) then
                k2 = k2 + 1
                j21 = j2
                exit jLoop2
            end if
            
        end do
    end do jLoop2


    if (k2 .eq. 2) then

        source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag
        IsMod8Source = .true.
        if (present(i1out)) i1out = i11
        if (present(i2out)) i2out = i21
        if (present(j1out)) j1out = j11
        if (present(j2out)) j2out = j21

    else
        IsMod8Source = .false.

    end if

end Function IsMod8Source

end module OmdetectDetection_mod


