!*******************************************
! A module to compute the position and shape of
! the detected sources by moments
!******************************************
module Parameters_mod
   
    use dal
    use OmAllocation_mod
    use errorHandling
    use omconstants_mod
    use omsourceparam_mod 

    use omnumerics_mod
    use omutility_mod
!    use ellipse_mod
    

!****************************************
!
!****************************************
type PMoments
    real(double) :: xPos = 0.0_double, xPosErr = 0.0_double
    real(double) :: yPos = 0.0_double, yPosErr = 0.0_double
    real(double) :: semiA = 0.0_double, semiAErr = 0.0_double
    real(double) :: semiB = 0.0_double, semiBErr = 0.0_double
    real(double) :: theta = 0.0_double, thetaErr = 0.0_double
    real(single) :: peakSigNoise = 0.0
end type PMoments

!******************************************
!
!******************************************
type MomentsData
    real(double) :: x = 0.0_double, y = 0.0_double
    real(double) :: counts = 0.0_double
    real(double) :: backgroundCounts = 0.0_double
    real(double) :: xPosD = 0.0_double, yPosD = 0.0_double
    real(double) :: x2d = 0.0_double, y2d = 0.0_double, xyd = 0.0_double
    real(double) :: sigmaSquared = 0.0_double
end type MomentsData


!****************************************
! A structure to store pixel information
!****************************************
Type PixelStruct
    integer(int16) :: x = 0
    integer(int16) :: y = 0
    integer(int16) :: number = 0
    real(single) :: counts = 0.0
    real(single) :: backgroundCounts = 0.0
    integer(int16) :: noNeighbours = 0
end Type PixelStruct



interface  ComputePointSourceParameters
    module procedure  ComputePointSrcParameters1,  &
                      ComputePointSrcParameters2
end interface  ComputePointSourceParameters

interface ComputeExtendedSourceParameters
    module procedure ComputeExtendedSrcParameters1,  &
                     ComputeExtendedSrcParameters2,  &
                     ComputeExtendedSrcParameters3

end interface  ComputeExtendedSourceParameters


contains
!*************************************************************
! Subroutine ComputePointSourceParameters
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputePointSrcParameters1(sourceNumber, nx1, ny1, image, &
                i0, j0, srcParams, dx2, computeMonteErrors)


    integer(int16), intent(in) :: sourceNumber
    integer(int16), intent(in) :: nx1, ny1
    real(single), pointer   :: image(:, :)
    integer(int16), intent(in) :: i0, j0
    integer(int8), intent(in), optional :: dx2
    logical, optional, intent(in) :: computeMonteErrors
    type(SourceParameters), intent(inout) :: srcParams
    integer (int16) :: i, j, i1, j1, n
    integer(int16) :: il, jl, iu, ju 
    integer(int16) :: dx
    type(MomentsData), pointer :: data(:) => null()
    real(double) :: imageVal
    real(double) :: minImage
    type(PMoments) :: computedMoments
    logical :: status

    if (present(dx2)) then
        dx = dx2 
    else
        dx = 4 / binX    ! was 6
    end if


    il = i0 - dx
    if (il .lt. 0) then
        il = 0
        iu = 2 * dx + 1
    else
        iu = i0 + dx
        if (iu .gt. nx1) then
            iu = nx1
            il = nx1 - 2 * dx - 1
        end if
    end if
   

    jl = j0 - dx

    if (jl .lt. 0) then
        jl = 0
        ju = 2 * dx + 1
    else
        ju = j0 + dx
        if (ju .gt. ny1) then
            ju = ny1
            jl = ny1 - 2 * dx - 1
        end if 
    endif
   


    

    srcParams%extensionFlag = 0
    srcParams%number = sourceNumber
    srcParams%iMin = il
    srcParams%iMax = iu
    srcParams%jMin = jl
    srcParams%jMax = ju




    n = (iu - il + 1) * (ju - jl + 1)
    allocate(data(1 : n)) 

    ! Find the minimum pixel value for crude background-subtraction
    minImage = 100000.0
    do j = jl, ju
    
        do i = il, iu
            imageVal = dble(image(i, j)) ! - minImage
            if(imageVal .ne. pixelIgnoreFlag) then
                minImage = min(imageVal, minImage)  
                  
            end if
        end do
    end do

    minImage = max(0.0_double, minImage - 2.0_double) ! Crude background counts / pixel


    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************
    n = 0
    do j = jl, ju
        j1 = j - jl + 1
        do i = il, iu
            imageVal = dble(image(i, j))
            if(imageVal .ne. pixelIgnoreFlag) then
           
                n = n + 1    
                data(n )%sigmaSquared = imageVal + minImage

                                  
                imageVal = max(0.0_double, dble(image(i, j)) - minImage) ! Rough background subtraction
                
                i1 = i - il + 1 
    
                data(n)%x = dble(i1)
                data(n)%y = dble(j1)
                data(n)%x2D = imageVal * dble(i1 * i1)
                data(n)%y2d = imageVal * dble(j1 * j1)
                data(n)%xyD = imageVal * dble(i1 * j1)
                data(n)%counts = imageVal
                data(n)%backgroundcounts = minImage

                data(n)%xposD = imageVal * dble(i1)   
                data(n)%yposD = imageVal * dble(j - jl + 1)   
                  
            end if
        end do
    end do


    call ComputeMoments(data, il, jl, computedMoments, status)





    srcParams%poserr = sqrt(computedMoments%xposErr * computedMoments%xposErr + &
                             computedMoments%yposErr * computedMoments%yposErr)


    if (n .lt. 6) then
 
        call SetShapeParameters(srcParams)
        deallocate(data)
        nullify(data)
        return

    end if




    srcParams%semiaerr = computedMoments%semiaerr
    srcParams%semiberr = computedMoments%semiberr





    if (present(computeMonteErrors) .and. status) then  ! Recompute errors from Monte-Carlo simulation
    
        if (computemonteerrors .and. n .gt. 8) then 
            call computeErrors(data, il, jl, computedMoments)
        end if
    end if
    if (status) then

        call updateSourceParams(srcParams, computedMoments)
             
      
    end if
   
    deallocate(data)
    nullify(data)




end subroutine ComputePointSrcParameters1


!*************************************************************
! Subroutine ComputePointSourceParameters2
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputePointSrcParameters2(sourceNumber, nx1, ny1, image, backgroundImage, &
                i0, j0, srcParams, dx2, computeMonteErrors)

    



    integer(int16), intent(in) :: sourceNumber
    integer(int16), intent(in) :: nx1, ny1
    real(single), pointer   :: image(:, :)
    real(single), pointer   :: backgroundImage(:, :)

    integer(int16), intent(in) :: i0, j0
    integer(int8), intent(in), optional :: dx2
    logical, intent(in), optional :: computeMonteErrors
    type(SourceParameters), intent(inout) :: srcParams
    type(PMoments) :: computedMoments

    integer (int16) :: i, j, i1, j1
    integer(int32) :: n
    integer(int16) :: il, jl, iu, ju 
    integer(int16) :: dx
    type(MomentsData), pointer :: data(:) => null()
    real(double) :: imageVal
    logical :: status   




    if (present(dx2)) then
        dx = dx2 
    else
        dx = 2 / binX    ! was 4
    end if


    il = i0 - dx
    if (il .lt. 0) then
        il = 0
        iu = 2 * dx + 1
    else
        iu = i0 + dx
        if (iu .gt. nx1) then
            iu = nx1
            il = nx1 - 2 * dx - 1
        end if
    end if
   

    jl = j0 - dx
    if (jl .lt. 0) then
        jl = 0
        ju = 2 * dx + 1
    else
        ju = j0 + dx
        if (ju .gt. ny1) then
            ju = ny1
            jl = ny1 - 2 * dx - 1
        end if 
    endif
   


    srcParams%extensionFlag = 0
 
    srcParams%number = sourceNumber
    srcParams%iMin = il
    srcParams%iMax = iu
    srcParams%jMin = jl
    srcParams%jMax = ju


    n = (iu - il + 1) * (ju - jl + 1)


    allocate(data(1 : n)) 



    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************
    n = 0



    do j = jl, ju
        j1 = j - jl + 1
        do i = il, iu
             if(imageVal .ne. pixelIgnoreFlag) then
                n = n + 1
       
                imageVal = max(0.0_double, dble(image(i, j) - backgroundImage(i, j))) ! Background subtraction
                data(n )%sigmaSquared = imageVal 

                               
                i1 = i - il + 1 
  
                data(n)%x = dble(i1)
                data(n)%y = dble(j1)
                data(n)%x2D = imageVal * dble(i1 * i1)
                data(n)%y2d = imageVal * dble(j1 * j1)
                data(n)%xyD = imageVal * dble(i1 * j1)
                data(n)%counts = imageVal
                data(n)%backgroundcounts = backgroundImage(i, j)
                data(n)%xposD = imageVal * dble(i1)   
                data(n)%yposD = imageVal * dble(j - jl + 1)   
             
            end if
        end do
    end do




    !**************************************************
    ! Compute the moments
    !**************************************************

    call ComputeMoments(data, il, jl, computedMoments, status)





    srcParams%poserr = sqrt(computedMoments%xposErr * computedMoments%xposErr + &
                             computedMoments%yposErr * computedMoments%yposErr)


    if (n .lt. 6) then
 
        call SetShapeParameters(srcParams)
        deallocate(data)
        nullify(data)
        return

    end if







    srcParams%semiaerr = computedMoments%semiaerr
    srcParams%semiberr = computedMoments%semiberr





    if (present(computeMonteErrors) .and. status) then
    
        if (computemonteerrors .and. n .gt. 8) then 
            call computeErrors(data, il, jl, computedMoments)
        end if
    end if   


    if (status) then
        call updateSourceParams(srcParams, computedMoments)
 
    

    end if 
    

!if (srcParams%poserr .lt. 0.0) call fatal("bug on paramters_mod at line 395")
    deallocate(data)
    nullify(data)



end subroutine ComputePointSrcParameters2

!*************************************************************
! Subroutine ComputeExtendedSourceParameters
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputeExtendedSrcParameters1(sourceNumber, image, &
                levelImage, srcParams, computeMonteErrors)


    integer(int16), intent(in) :: sourceNumber
      
    real(single), pointer   :: image(:, :)
    integer(int16), pointer :: levelImage(:, :)
   
    type(SourceParameters), intent(inout) :: srcParams
    logical, intent(in), optional :: computeMonteErrors

    type(PMoments) :: computedMoments
    
    integer(int16) :: i, j, i1, j1, il, jl
   
    type(MomentsData), pointer :: data(:) => null()
    real(double) :: imageVal, minImage
   
    integer(int32) :: n
    logical :: status
    srcParams%number = sourceNumber   
    

    n = (srcParams%iMax - srcParams%iMin + 1) * (srcParams%jMax - srcParams%jMin + 1) 




    allocate(data(1 : n)) 
 
    il = srcParams%iMin
    jl = srcParams%jMin

    ! Find the minimum pixel value for crude background subtraction
    minImage = 100000.0
   
    do j = srcParams%jmin, srcParams%jmax
        do i = srcParams%imin, srcParams%imax
            if (levelImage(i, j) .ne. sourceNumber) cycle
     
    
            imageVal = dble(image(i, j)) 
            if(imageVal .ne. pixelIgnoreFlag .and. imageval .ge. 0.0) then
                minImage = min(imageVal, minImage)  
                
            end if
        end do
    end do


    minImage = max(0.0_double, minImage - 2.0_double)





    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************
    n = 0
    
    do j = srcParams%jmin, srcParams%jmax
        j1 = j - jl + 1
        do i = srcParams%imin, srcParams%imax
       
            if (levelImage(i, j) .ne. sourceNumber) cycle
             imageVal = image(i, j)
             if(imageVal .ne. pixelIgnoreFlag) then
                imageVal = dble(image(i, j)) - minImage ! Rough background subtraction
                if (imageVal .le. 0.0) cycle  ! TEST ON WEDNESDAY
 

                n = n + 1    
                data(n)%sigmaSquared = imageVal + minImage
                i1 = i - il + 1 
        
 
                data(n)%x = dble(i1)
                data(n)%y = dble(j1)
                data(n)%x2D = imageVal * dble(i1 * i1)
                data(n)%y2d = imageVal * dble(j1 * j1)
                data(n)%xyD = imageVal * dble(i1 * j1)
                data(n)%counts = imageVal
                data(n)%backgroundcounts = minImage
            
                data(n)%xposD = imageVal * dble(i1)   
                data(n)%yposD = imageVal * dble(j - jl + 1)   
               
            end if
        end do
    end do


    call ComputeMoments(data(1 : n), il, jl, computedMoments, status)



    srcParams%poserr = sqrt(computedMoments%xposErr * computedMoments%xposErr + &
                             computedMoments%yposErr * computedMoments%yposErr)
   
    if (n .lt. 6) then
        
        call SetShapeParameters(srcParams)
        deallocate(data)
        nullify(data)
        return

    end if







  
    if (status) then
        if (present(computemonteerrors)) then  
        if (computemonteerrors .and. n .gt. 10) then 
            call computeErrors(data, il, jl, computedMoments)
        end if
        end if


       srcParams%xPos = computedMoments%xPos
!       srcParams%posErr = sqrt(computedMoments%xPosErr * computedMoments%xPosErr + &
!              computedMoments%yPosErr * computedMoments%yPosErr)

       srcParams%yPos = computedMoments%yPos
  
       srcParams%semia = computedMoments%semia
       srcParams%semiaErr = computedMoments%semiaErr
 


       srcParams%semiB = computedMoments%semiB
       srcParams%semiBErr = computedMoments%semiBErr
 

       srcParams%theta = computedMoments%theta  * radiansToDegrees
       srcParams%thetaErr = computedMoments%thetaErr  * radiansToDegrees
    


    end if

    deallocate(data)
    nullify(data)

!if (srcParams%poserr .lt. 0.0) call fatal("bug on paramters_mod at line 551")


end subroutine ComputeExtendedSrcParameters1





!*************************************************************
! Subroutine ComputeExtendedSourceParameters
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! newImage - A 3-d array containing a 2-d image for each array.
! background - A 2-d background image.
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputeExtendedSrcParameters2(sourceNumber, nx1, ny1, &
                image, backgroundImage,                          &
                levelImage, srcParams, computeMonteErrors)


    integer(int16), intent(in) :: sourceNumber
    integer(int16), intent(in) :: nx1, ny1
      
    real(single), pointer   :: image(:, :), backgroundImage(:, :)
    integer(int16), pointer :: levelImage(:, :)
   
    type(SourceParameters), intent(inout) :: srcParams
    logical, intent(in), optional :: computeMonteErrors
    type(PMoments) :: computedMoments
    
    integer(int16) :: i, j, i1, j1, il, jl, i0, j0
 
    type(MomentsData), pointer :: data(:) => null()
    real(double) :: imageVal
   
    integer(int32) :: n
    logical :: status

    srcParams%number = sourceNumber   



    n = (srcParams%iMax - srcParams%iMin + 1) *  (srcParams%jMax - srcParams%jMin + 1) 


!write(*, *) "Computeextendedsrcparamters2 n=", n

    allocate(data(1 : n)) 
 
    il = srcParams%iMin
    jl = srcParams%jMin


    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************
    n = 0



    
    do j = srcParams%jmin, srcParams%jmax
        j1 = j - jl + 1
        do i = srcParams%imin, srcParams%imax
           
            if (levelImage(i, j) .ne. srcParams%number) cycle
            imageVal = dble(image(i, j))

            if(imageVal .ne. pixelIgnoreFlag) then


                n = n + 1  
                data(n)%sigmaSquared = imageVal + backgroundImage(i, j)
                               
                imageVal = max(0.0_double, dble(image(i, j) - backgroundImage(i, j))) ! Background subtraction
                i1 = i - il + 1 
     
                data(n)%x = dble(i1)
                data(n)%y = dble(j1)
                data(n)%x2D = imageVal * dble(i1 * i1)
                data(n)%y2d = imageVal * dble(j1 * j1)
                data(n)%xyD = imageVal * dble(i1 * j1)
                data(n)%counts = imageVal
                data(n)%backgroundcounts = backgroundImage(i, j)
            
                data(n)%xposD = imageVal * dble(i1)   
                data(n)%yposD = imageVal * dble(j1)   
               
            end if
        end do
    end do


    if (n .lt. 6) then  ! was 10


        deallocate(data)
        nullify(data)    
        i0 = int(srcParams%xPos - 0.5)
        i0 = max(0_int16, i0)
        i0 = min(nx1, i0)
        j0 = int(srcParams%yPos - 0.5)
        j0 = max(0_int16, j0)
        j0 = min(ny1, j0)
    
        srcParams%extensionFlag = 0 

        call ComputePointSrcParameters2(sourceNumber, nx1, ny1, image, backgroundImage, &
                i0, j0, srcParams, int(6 / binx, int8), computeMonteErrors)
        
   

        call SetShapeParameters(srcParams)
        
        return
    end if

   

    call ComputeMoments(data(1 : n), il, jl, computedMoments, status)
 


    srcParams%posErr = sqrt(computedMoments%xPosErr * computedMoments%xPosErr + &
                     computedMoments%yPosErr * computedMoments%yPosErr)
  


  
    if (status) then
        if (present(computemonteerrors)) then  
   
            if (computemonteerrors) then 
                call computeErrors(data, il, jl, computedMoments)
            end if
        end if


        srcParams%xPos = computedMoments%xPos

!        srcParams%posErr = sqrt(computedMoments%xPosErr * computedMoments%xPosErr + &
!              computedMoments%yPosErr * computedMoments%yPosErr)

  

        srcParams%yPos = computedMoments%yPos
  
        srcParams%semia = computedMoments%semia
        srcParams%semiaErr = computedMoments%semiaErr
 


        srcParams%semiB = computedMoments%semiB
        srcParams%semiBErr = computedMoments%semiBErr
 

        srcParams%theta = computedMoments%theta  * radiansToDegrees
        srcParams%thetaErr = computedMoments%thetaErr * radiansToDegrees
      
    end if


    deallocate(data)
    nullify(data)    
   
  

!if (srcParams%poserr .lt. 0.0) call fatal("bug on paramters_mod at line 721")


end subroutine ComputeExtendedSrcParameters2



!*************************************************************
! Subroutine ComputeExtendedSourceParameters
! compute the source parameters- position,
! count  rates, etc
! Parameters:
! A structure pixel list  containing information on the
! source pixels (coordinates, counts and background counts)
! Returns: srcParams- a 1-d array of structures containing
! the parameters for each source. 
!*************************************************************
subroutine ComputeExtendedSrcParameters3(sourceNumber, &
                                         pixel,        &
                                         srcParams,    &
                                         computeMonteErrors)


    integer(int16), intent(in) :: sourceNumber
    type(PixelStruct), intent(in) :: pixel(:)
    type(SourceParameters), intent(inout) :: srcParams
    type(PMoments) :: computedMoments
    logical, intent(in), optional :: computeMonteErrors

    integer(int16) :: i, j, i1, j1, il, jl
 
 
    type(MomentsData), pointer :: data(:) => null()
    real(double) :: counts, backgroundCounts
   
    integer(int32) :: n
    logical :: status
    ! Find the extent of the source in x and y directions
    srcParams%number = sourceNumber   
    srcParams%iMin = minval(pixel%x)
    srcParams%iMax = maxval(pixel%x)
    srcParams%jMin = minval(pixel%y)
    srcParams%jMax = maxval(pixel%y)

   
    n = srcParams%noSourcePixels  ! The number of source pixels

    allocate(data(1 : n)) 
 
    il = srcParams%iMin
    jl = srcParams%jMin



 
    !********************************************************************
    ! Compute the x and y centroid positions (xpos & ypos)
    !********************************************************************
    n = 0
    
    do i = 1, srcParams%noSourcePixels  
       
        counts = pixel(i)%counts
        if(counts .ne. pixelIgnoreFlag) then

            i1 = pixel(i)%x - il + 1
            j = pixel(i)%y 
            j1 = j - jl + 1 

            backgroundCounts = pixel(i)%backgroundCounts
           
            n = n + 1  
            data(n)%sigmaSquared = counts + backgroundCounts
                               
            counts = max(0.0_double, counts - backgroundCounts) ! Background subtraction
        
            data(n)%x = dble(i1)
            data(n)%y = dble(j1)
            data(n)%x2D = counts * dble(i1 * i1)
            data(n)%y2d = counts * dble(j1 * j1)
            data(n)%xyD = counts * dble(i1 * j1)
            data(n)%counts = counts
            data(n)%backgroundcounts =  backgroundCounts
            
            data(n)%xposD = counts * dble(i1)   
            data(n)%yposD = counts * dble(j - jl + 1)   
               
        end if
    end do





    call ComputeMoments(data(1 : n), il, jl, computedMoments, status)

    srcParams%poserr = sqrt(computedMoments%xposErr * computedMoments%xposErr + &
                             computedMoments%yposErr * computedMoments%yposErr)
    


    if (n .lt. 6) then
        
   

        call SetShapeParameters(srcParams)
        deallocate(data)
        nullify(data)
        return

    end if




   
    if (status) then
        if (present(computemonteerrors)) then  
   
        if (computemonteerrors .and. n .gt. 10) then 
            call computeErrors(data, il, jl, computedMoments)
        end if
        end if 
  
  
        srcParams%xPos = computedMoments%xPos
!        srcParams%posErr = sqrt(computedMoments%xPosErr * computedMoments%xPosErr + &
!              computedMoments%yPosErr * computedMoments%yPosErr)

        srcParams%yPos = computedMoments%yPos
  
        srcParams%semia = computedMoments%semia
 
        srcParams%semiaErr = computedMoments%semiaErr
 

        srcParams%semiB = computedMoments%semiB
        srcParams%semibErr = computedMoments%semibErr
 

        srcParams%theta = computedMoments%theta  * radiansToDegrees
        srcParams%thetaErr = computedMoments%thetaErr * radiansToDegrees

   

    end if
    deallocate(data)
    nullify(data)

!if (srcParams%poserr .lt. 0.0) call fatal("bug on paramters_mod at line 861")



end subroutine ComputeExtendedSrcParameters3


!*************************************************************
!
!
!*************************************************************
subroutine computeErrors(data, il, jl, computedMoments)
    type(MomentsData) :: data(:)
    integer(int16), intent(in) :: il, jl
    type(PMoments), intent(inout) :: computedMoments
    type(MomentsData), pointer :: tempData(:) => null()

    real(single) :: number
    type(PMoments), pointer :: tempMoments(:) => null()
    real(double) :: mean, standardDeviation
 
    integer(int32) :: i, j, k, l, index
    integer(int32) :: n
    logical :: status
    n = size(data)
    allocate(tempData(1 : n))

    k = max(100, 10 * n)
    allocate(tempMoments(1 : k))

    l = 0
    do j = 1, k  ! Each simulation
        ! Generate a new data set for each simulation
        do i = 1, n
            call random_number(number)

            index = int(number * (real(n - 1) + 0.0001)) + 1
    
            tempdata(i) = data(index) ! The new data set
        end do
        
        call ComputeMoments(tempData, il, jl, tempMoments(j), status)




        if (status) then
            l = l + 1
            tempMoments(l) = tempMoments(j)
        end if          
    end do



    if (l .lt. 2) then

        return

    end if
    ! Compute the errors
    call ComputeStatistics(l, tempMoments%xPos, mean, standardDeviation)
    computedMoments%xPosErr = sqrt(standardDeviation)



    call ComputeStatistics(l, tempMoments%yPos, mean, standardDeviation)
    computedMoments%yPosErr = sqrt(standardDeviation)
      
    call ComputeStatistics(l, tempMoments%semia,mean, standardDeviation)
    computedMoments%semiAErr = sqrt(standardDeviation)
  
    
    call ComputeStatistics(l, tempMoments%semiB, mean, standardDeviation)
    computedMoments%semiBErr = sqrt(standardDeviation)
  
 
    call ComputeStatistics(l, tempMoments%theta, mean, standardDeviation)
    computedMoments%thetaerr = sqrt(standardDeviation)


 
    deallocate(tempData, tempMoments)
    nullify(tempData, tempMoments)
    
end subroutine computeErrors

!**************************************************************************
!
!
!*************************************************************************
subroutine ComputeMoments(data, il, jl, computedMoments, status)
    
    type(MomentsData) :: data(:)
    integer(int16) :: il, jl
    type(PMoments), intent(out) :: computedMoments
    logical, intent(out) :: status
    integer(int32) :: i, n
    real(double) :: totalImage, xPosD, yPosD
    real(double) :: varX, varY, covXY
    real(double) :: dx, dy
    real(single) :: x2, y2, xy
    real(single) :: val1, val2, diff
    real(single) :: semiA, semiB
 
    n = size(data)

 
    totalImage = sum(data%counts)

    if (totalImage .le. 0.0) then

        status = .false.

        return
    end if

    status = .true.




     
    xposD = sum(data%xposD) / totalImage
    yposD = sum(data%yposD) / totalImage

    computedMoments%xPos = xPosD + dble(il)
    computedMoments%YPos = yPosD + dble(jl)



    !*******************************************************
    ! Compute the postion errors
    !*******************************************************

    varX = 0.0_double ! ERRX2 var(xbar)
    varY = 0.0_double ! ERRY2 var(ybar)
    covXY = 0.0_double

    do i = 1, n
   
        dx = data(i)%x - xPosD
        dy = data(i)%y - yPosD
        varX = varX + data(i)%sigmaSquared * dx * dx
        varY = varY + data(i)%sigmaSquared * dy * dy
        covXY = covXY + data(i)%sigmaSquared * dx * dy


    end do


    x2 = real(sum(data%x2D) / totalImage - xposD * xposD)
    y2 = real(sum(data%y2D) / totalImage - yposD * yposD)
    xy = real(sum(data%xyD) / totalImage - xposD * yposD)
    totalImage = totalImage * totalImage




    varX = varX / totalImage
    varY = varY / totalImage
    covXY = covXY / totalImage


    computedMoments%xPosErr = sqrt(varX)   ! Source pixel x-position error
    computedMoments%yPosErr = sqrt(varY)   ! Source pixel y-position error


  
    ! The error in a
    computedMoments%semiAErr = sqrt((varX + varY) / 2.0_double + &
             sqrt(((varX - varY) / 2.0_double) ** 2 + covXY * covXY))

    computedMoments%semiBErr = computedMoments%semiAErr 




    ! Now convert these into major/minor values and theta
   
  
    val1 = (x2 + y2) / 2.0
    val2 = sqrt(((x2 - y2) / 2.0) ** 2 + xy * xy)

   


    semiA = sqrt(val1 + val2)

        
    if(val1 .gt. val2) then
        semiB = sqrt(val1 - val2)
        computedMoments%semiA = max(semiA, semiB)
        computedMoments%semiB = min(semiA, semiB)

        ! The position angle of the source
        computedMoments%theta = Angle(x2, y2, xy)

        if (computedMoments%theta .lt. 0.0) computedMoments%theta = computedMoments%theta + PI 
    


    else

        diff = abs(val1 - val2)

        if (diff .gt. 1.0) then ! NB original JPDM criterion was 0
            
            ! There is a problem with the source parameterisation
          
            write(errstr,'("Warning: Problem with parameterisation of source")')
  
            call message(AppMsg, Verbose)      ! Verbose was 2

            
        else

        


            ! The position angle of the source
            computedMoments%theta = Angle(x2, y2, xy)

            if (computedMoments%theta .lt. 0.0) computedMoments%theta = computedMoments%theta + PI 
    
        end if
    end if 


        

   
end subroutine ComputeMoments



!*****************************************************************
!
!
!****************************************************************
subroutine updateSourceParams(srcParams, computedMoments)
    type(SourceParameters), intent(inout) :: srcParams
    type(PMoments), intent(in) :: computedMoments
 

    real(single) :: xPos, yPos, dx1, dy1
   
    xPos = computedMoments%xPos
    yPos = computedMoments%yPos


    dx1 = srcParams%xPos - xPos
    dy1 = srcParams%yPos - yPos

   ! Check for shifted position due to close-proximity star
    if (abs(dx1) .le. real(4 / binX) .and. abs(dy1) .le. real(4 / binX)) then  
        srcParams%xPos = xpos
        srcParams%yPos = yPos
            srcParams%semia = computedMoments%semiA 
            srcParams%semiaErr = computedMoments%semiaErr

            srcParams%semib = computedMoments%semiB

            srcParams%semiBErr = computedMoments%semiBErr

            srcParams%theta = computedMoments%theta   * radiansToDegrees
            srcParams%thetaErr = computedMoments%thetaErr * radiansToDegrees 

        else

            srcParams%semiA = computedMoments%semiA
            srcParams%semiaErr = computedMoments%semiaErr

            srcParams%semiB = computedMoments%semiB 
            srcParams%semibErr = computedMoments%semibErr

            srcParams%theta = computedMoments%theta  * radiansToDegrees
            srcParams%thetaErr = computedMoments%thetaErr * radiansToDegrees 
 
        end if
   

!        srcParams%poserr = sqrt(computedMoments%xPosErr * computedMoments%xPosErr + &
!                  computedMoments%yPosErr * computedMoments%yPosErr)            






end subroutine updateSourceParams

!**************************************************
!
!**************************************************
subroutine SetShapeParameters(source)
    type(SourceParameters), intent(inout) :: source
    type(SourceParameters), save :: tempSource
   
    source%extensionFlag = 0
    source%noSourcePixels = 0
    source%semiA = tempSource%semiA
    source%semiAErr = tempSource%semiAErr
    source%semiB = tempSource%semiB
    source%semiBErr = tempSource%semiBErr
    source%theta = tempSource%theta
    source%thetaErr = tempSource%thetaErr

 
end subroutine SetShapeParameters


subroutine FitEllipse(source, levelImage)
    type(SourceParameters), intent(in) :: source
    integer(int16), pointer :: levelImage(:, :)
    integer :: i, j  

    real(double), pointer, dimension(:) :: x => null(), y => null()
    real(double) ::  xc = 0.0_double, yc=0.0_double, a=0.0_double, &
                     b=0.0_double, theta=0.0_double
    integer :: n
    n = source%noSourcePixels
    if (n .lt. 20) return
 
    x => allocateDouble(1, n, "x")
    y => allocateDouble(1, n, "y")

 
    n = 0
    do j = source%jMin, source%jMax
        do i = source%iMin, source%iMax
            if (levelImage(i, j) .eq. source%number) then
                n = n + 1
                x(n) = dble(i + 1)
                y(n) = dble(j + 1)
                exit
            end if    
        end do
    end do

    do j = source%jMin, source%jMax
        do i = source%iMax, source%iMin, -1
            if (levelImage(i, j) .eq. source%number) then
                n = n + 1
                x(n) = dble(i + 1)
                y(n) = dble(j + 1)
                exit
            end if    
        end do
    end do

write(*, *) "Calling ellispefit n=", n
!    call ellipseFit(n, x, y, xc, yc, a, b, theta)
write(*, *) "Ellipse fit, xc=", xc, " yc=", yc, " a=", a, " b=", b, " theta=", theta 
    deallocate(x, y)
    nullify(x, y)

end subroutine FitEllipse


end module Parameters_mod



















