!*********************************************************
! Program for fitting PSF or 2-d gaussian
! functions to image data
!
!**********************************************************
module stardecomp_mod

    use dal
    use cal
    use param  
    use sasdal
    use omconstants_mod
    use omallocation_mod 
    use errorHandling
    use omsmoothimage_mod
    use nonlinls_mod 
    use cubicshepard_mod
    use gridInterpolation_mod
    use omsourceparam_mod
    use splinefit_mod
    use ompsf_mod
    use parameters_mod
    use omderivative_mod
    use integration_mod
    use omfunctions_mod
    use omutility_mod
    use omcoinccorr_mod
    use ompsfintegration_mod
    implicit none 
 

!******************************************************
!
!******************************************************
interface GaussDeconvolve
    module procedure GaussDeconvolve1, GaussDeconvolve2
end interface  GaussDeconvolve

interface GridIntegration
    module procedure GridIntegration1, GridIntegration2
end interface GridIntegration



!********************************************
! Stores the least-squares fitting data
!******************************************* 
type ImageData
    integer(int8) :: noStars = 1
    integer(int32) :: m = 0    ! No of data points
    real(double), pointer ::  x(:) => Null() ! pixel x-coordinates
    real(double), pointer ::  y(:) => Null() ! pixel y-coordinates
    real(double), pointer ::  f(:) => Null() ! pixel counts
    real(double), pointer ::  weight(:) => Null() ! Pixel weights
    real(double), pointer :: fit(:) => null()     ! Fitted value
    real(double) :: xcentre, ycentre, rMax
end type ImageData


type (ImageData), save, private :: lsData     

real(double), private :: sigmaConstant, sigmaXConstant, sigmaYConstant, pa  
real(double), private :: dxSquared
real(double), private :: eps = 0.0_double

!***********************************************************************************
! The following are for psf multi-fitting functions
!**********************************************************************************
integer, private :: jPrivate
integer(int16), private, save ::  kArray(5000) = (/ ( (jPrivate - 1) * 3 + 2, jPrivate = 1, 5000) /)

!***********************************************
! 2-d Gaussian fitted parameters
!***********************************************
type GaussFitParameters
    real(single) :: xCentre = 0.0, xcentreErr = 0.0
    real(single) :: yCentre = 0.0, ycentreErr = 0.0
    real(single) :: background = 0.0, backgrounderr = 0.0
    real(single) :: height = 0.0, heighterr = 0.0
    real(single) :: sigmax = 0.0, sigmaxerr = 0.0
    real(single) :: sigmay = 0.0, sigmayerr = 0.0
    real(single) :: positionAngle = 0.0, positionAngleErr = 0.0
    real(single) :: redchisq = 0.0
end type GaussFitParameters


!***********************************************
! PSF fitted parameters
!***********************************************
type PsfFitParameters
    real(single) :: xCentre = 0.0, xcentreErr = 0.0
    real(single) :: yCentre = 0.0, ycentreErr = 0.0
    real(single) :: background = 0.0, backgrounderr = 0.0
    real(single) :: height = 0.0, heighterr = 0.0
    real(single) :: redchisq = 0.0
end type PsfFitParameters



contains 

!************************************************************
!
!************************************************************
subroutine stardecomp(source, image, binX, binY)
    type(sourceParameters), pointer :: source(:) 
   
    type(sourceParameters), pointer :: tempSource(:) => null()

    real(single), pointer :: image(:, :)
    integer, intent(in) :: binX, binY
    integer(int16) :: nx, ny
    integer :: i, psfNx = 12, psfNy = 12, number = 1
    real(single) :: radius
    real(double) :: psfCounts
    real(single) :: mode, sigma

    radius = real(8 / binX)

    write(errstr, '(a)') "Subroutine stardecomp"
    call message(AppMsg, Verbose)

   
    nx = size(image, 1)
    ny = size(image, 2)


    ! Get the global background mode
    mode = ComputeBackgroundMode2(image, nx, ny, 2.0, sigma) 


    psfCounts = -1.0_double
    call CreatePSF2(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binY, &
           1, 1,  psfCounts, psfSpline = psfsplineFit)


  call CreatePSF4(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binY, &
           1, 1,  psfCounts, psfshepardFit)

  call CreatePSF6(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binY, &
           1, 1,  psfCounts, psfGridFit)

 



    number = 1
    allocate(tempSource(1 : 1))

    do i = 1, size(source)
        
!        if (source(i)%xPos .lt. radius .or. source(i)%xPos .gt. real(nx) - radius - 0.5 &

!          .or. source(i)%yPos .lt. radius .or. source(i)%yPos .gt. real(ny) - radius - 0.5) then
 
            tempSource = source(i : i)

            write(errstr, '(a, i4)') "Recomputing coordinates of source ", i      
            call message(AppMsg, Verbose)
              
!            call stardecon_deconvolve3(binX, binY, number, tempSource, image, psfCounts, mode)

            write(errstr, '(4(a, f8.2))') " Original detected x=", &
                      source(i)%xpos, " new ", tempSource(1)%xPos

                    call message(AppMsg, Verbose)
           
                    write(errstr, '(4(a, f8.2))') " Original detected y=", &
                      source(i)%ypos, " new ", tempSource(1)%yPos

                    call message(AppMsg, Verbose)
           
                    write(errstr, '(4(a, e12.4))') " Original raw count rate", &
                      source(i)%rawCountRate, " new ", tempSource(1)%rawCountRate
      
                    call message(AppMsg, Verbose)
     
                    write(errstr, '(4(a, e12.4))') " Original raw background count rate", &
                      source(i)%rawbackgroundCountRate, " new ", tempSource(1)%rawbackgroundCountRate
      
                    call message(AppMsg, Verbose)
     
!        end if  
    end do

   
!    call DestructSpline(psfSpline)
    
    deallocate(tempSource)


!image = 100.0 * image / maxval(image)

!call TestGussianFit(source, image)
 

 
end subroutine stardecomp

!***************************************************
! This subroutine is fitting the psf to a single object
! with background
! x(1) = background
! x(2) = xcoord
! x(3) = ycoord
! x(4) = psf height
!
!**************************************************
subroutine FitPSFFunction(source, image, binX, dx1, psffitdata)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :)
    
    
    integer, intent(in) :: binX 
    integer, intent(in), optional :: dx1
    type(psfFitParameters), intent(out) :: psfFitData
    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(4), bl(4), bu(4)


    logical(boolean) :: toFit(4)


    real(single) :: mode, fMax 
    integer(int8) :: dx
    integer :: n   
   

    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
 
    write(errstr, '(a, i2)')  "FitPSFFunction Using dx=", dx
    call message(AppMsg, Verbose)
     


    eps = sqrt(EPSILON(0.0_double))
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    ! Get the global background mode
  

!    psfCounts = -1.0_double
!    call CreatePSF2(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binX, &
!           1, 1,  psfCounts, psfSpline = psfsplineFit)

  

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)

    
    toFit = .true.
    mode = minval(image(il : iu, jl : ju))


    fmax = maxval(image(il : iu, jl : ju))


 
    bl(1) = max(0.0, mode - 2.0 * sqrt(mode))
    bu(1) = max(mode + 2,0 * sqrt(mode), mode + 2.0)
   
    x(1) = mode
    bl(2) = max(0.0_double, source%xPos - dble(4 / binx))
    bu(2) = min(dble(nx + 1), source%xPos + dble(4 / binx))
    x(2) = source%xPos
    
    bl(3) = max(0.0_double, source%yPos - dble(4 / binx))
    bu(3) = min(dble(ny + 1), source%yPos + dble(4 / binx))
    x(3) = source%yPos
 
   
   
    bl(4) =  fmax - mode
    bu(4) =  fmax  * 3.0
    x(4) = fmax 
 
    n = 4
    lsData%rMax = dble(dx) + 0.5_double
    call PSFDeconvolve(1, image, &
            il, iu, jl, ju, n, x, bl, bu,  &
            toFit, psfFunction,  status, psffitdata)
    




!    call ChiSquared(n, x, lsData%m, toFit, psfFunction, da, redChisq)



    
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, PSFFunction, da, 10)



    
!    call destructSpline(psfSplineFit)  
  
end subroutine FitPSFFunction


!**************************************************
! 
! This subroutine goes with the above
!
!**************************************************
subroutine PSFDeconvolve(noStars, image, &
           il, iu, jl, ju, n, x, &
           bl, bu, toFit, func, status, psfFitData)
    integer(int32), intent(in) :: noStars
    real(single), pointer :: image(:, :)
   
   
  
    integer(int16), intent(in) :: il, iu, jl, ju
    integer(int32), intent(in) :: n
    real(double), intent(inout) :: x(1 : n)
    real(double), intent(inout) :: bl(1 : n), bu(1 : n)
    logical(boolean)  :: toFit(:)
    type(psfFitParameters), intent(out) :: psfFitData

    integer(int16), intent(inout) :: status
    integer(int32) :: i, j, m
    real(double), pointer :: f(:)=>Null()
    real(double) :: xPos, yPos
    real(double) :: error
    real(double) :: da(1 : n)
 
   
      interface
       
    subroutine func(x, n, f, m, ctrl, c)
    use dal
        real(double), intent(in) :: x(:)
        integer, intent(in) :: n
        real(double), intent(out) :: f(:)
        integer, intent(in) :: m
        integer, intent(inout) :: ctrl
        real(double), pointer :: c(:, :)
    end subroutine func

    end interface

    
    m = (iu - il + 1) * (ju - jl + 1)
    allocate(lsData%x(1 : m)) 
    allocate(lsData%y(1 : m)) 
    allocate(lsData%f(1 : m)) 
    allocate(lsData%weight(1 : m)) 

    m = 0
    lsData%noStars = noStars
 
    lsData%weight = 0.0_double


    do i = il, iu
        xPos = dble(i) + 1.0_double
        
        do j = jl , ju
          
            yPos = dble(j) + 1.0_double
                       
            m = m + 1
            lsData%x(m) = xPos
    
            lsData%y(m) = yPos
           
         
            lsData%f(m) = dble(image(i, j)) 

            error =  lsData%f(m)
         
                      
            if(error > 1.0_double) then
                lsData%weight(m) = 1.0_double / error
            else 
                lsData%weight(m) = 1.0_double 

            end if
     
   
        end do
    end do
    if (m .lt. 10) then

        write(errstr, '(a, i2, a)') "Too few points for fits (", m, ") fit abandoned"
        call message(AppMsg, Verbose)
        return

    end if


    lsData%m = m


    allocate(f(1 : m))



   
    write(errstr, '(a)') "Input trial parameters:"
    call message(AppMsg, Verbose)

    do i = 1, n    

        write(errstr, '(a, i2, a, e12.4, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), " bounds-", bl(i), "-", bu(i)      
        call message(AppMsg, Verbose)

    end do

    do j = 1, 1

        call nonlin(x, n, m, func, bl, bu, toFit, f, status)
   

        write(errstr, '(a, i1, a, e14.8)') "Iteration ", j, " Sum of squares of residuals=", DOT_PRODUCT(f, f)
        call message(AppMsg, Verbose)
    end do



    write(errstr, '(a, i7)') "status=", status   
    call message(AppMsg, Verbose)

    call ChiSquared(n, x, m, toFit, func, da, psfFitData%redChisq)



    deallocate(lsData%x) 
    deallocate(lsData%y) 
    deallocate(lsData%f) 
    deallocate(lsData%weight) 
    nullify(lsData%x) 
    nullify(lsData%y) 
    nullify(lsData%f) 
    nullify(lsData%weight) 




end subroutine PSFDeconvolve



!***********************************************************
!
! This is the psf fitting function for a single object 
! With background
!
!**********************************************************
subroutine PSFFunction(x, n, f, m, ctrl, c)
    integer, intent(in) :: n

    real(double), intent(in) :: x(:)
  
    real(double), intent(out) :: f(:)
    integer, intent(in) :: m
    integer, intent(inout) :: ctrl
    real(double), pointer :: c(:, :)
    integer i
    real(double) :: fit
    real(double) :: x1(size(x)), h, error
    real(double) :: dx, dy

    if (ctrl .eq. -5) then
        x1 = x
        

        ! Partial derivatives with respect to the height
        do i = 1, m            
            c(i, 1) = 1.0_double  ! The derivative with respect to the background                   

            x1(2) = lsData%x(i) - x(2)
            x1(3) = lsData%y(i) - x(3)
            fit = GridInterpolate(x1(2), x1(3), psfGridFit)
 
            f(i) = lsData%f(i) - (x(1) + x(4) * fit) ! The residual

            
            h = 0.1_double
            c(i, 2) = Derivative1(4, x1, 2, h, PSFPixel, derivative3, error) ! The derivative with respect to dx
                           
            h = 0.1_double
            c(i, 3) = Derivative1(4, x1, 3, h, PSFPixel, derivative3, error) ! The derivative with respect to dy

            c(i, 4) = fit ! The derivative with respect to the height
                           


        end do 
!write(*, '(8e14.4)') f(i), (c(i, j), j = 1, 4)

        
        return
    end if

    if (ctrl .eq. 2) then
!ctrl = 0
!return
        x1 = x
        do i = 1, m 
            x1(2) = lsData%x(i) - x(2)
            x1(3) = lsData%y(i) - x(3)
           
            ! Partial derivatives with respect to the background           
            c(i, 1) = -lsData%weight(i)

            ! Partial derivatives with respect to x 
            h = 0.1_double
            c(i, 2) = Derivative(4, x1, 2, h, PSFPixel) * lsData%weight(i)
                           
            h = 0.1_double
            c(i, 3) = Derivative(4, x1, 3, h, PSFPixel) * lsData%weight(i)
                           
            ! Partial derivatives with respect to the height                                          
            c(i, 4) =  -GridInterpolate(x1(2), x1(3), psfgridFit) * lsData%weight(i)                        

        end do

        return
write(*, *) n

    end if



    do i = 1, m 
        dx = lsData%x(i) - x(2)
        dy = lsData%y(i) - x(3)

        fit = x(4) * GridInterpolate(dx, dy, psfGridFit)    
                   
!        weight = lsData%rMax - sqrt(dx * dx + dy * dy)
!        weight = max(0.0_double, weight)  
        f(i) = (lsData%f(i) - (x(1) + fit)) * lsData%weight(i)
     
    end do  
 

        
end subroutine PSFFunction      
!***************************************************************
! 
!**************************************************************
real(double) Function MultiPSFFunction(x, noStars,  xPos, yPos)
    real(double) :: x(:)
    integer(int8), intent(in) :: noStars
    real(double), intent(in) :: xPos, yPos
    integer j, k, k1, k2
    real(double) :: dx, dy

    MultiPSFFunction = x(1)            
     
    do j = 1, noStars   ! Loop through all the stars in the fit
        k = karray(j)            
        k1 = k + 1
        k2 = k + 2          
        dx = xPos - x(k)
        dy = yPos - x(k1)

        MultiPSFFunction = MultiPSFFunction + x(k2) * max(0.0_double, GridInterpolate(dx, dy, psfGridFit))    

    end do

          

end function MultiPSFFunction

!***********************************************************
!
! This function fits one or more stars to a background
! and the psf
! x(1) - The background
! x(2) - x first star
! x(3) - y first star
! x(4) - height first star
! x(5) - x second star
! x(6) - y secind star
! x(7) - height second star
! etc for more stars
!
!**********************************************************
subroutine PSFMultiFitFunction(x, n, f, m, ctrl, c)
    integer, intent(in) :: n

    real(double), intent(in) :: x(:)
  
    real(double), intent(out) :: f(:)
    integer, intent(in) :: m
    integer, intent(inout) :: ctrl
    real(double), pointer :: c(:, :)
    integer i, j, k, k1, k2
    real(double) :: fit
    real(double) :: x1(size(x))
    real(double) :: dx, dy
    real(double) :: fx, fy, xTimesWeight
    real(double) :: minRadius, radius   



    if (ctrl .eq. -5) then ! Want the unweighted residuals and the derivatives

   
        do i = 1, m

            ! The derivatives with respect to the background             
            c(i, 1) = 1.0_double  ! The derivative with respect to the background                   


            !**************************************************************
            ! The derivatives with respect to the remaining stars
            !*************************************************************
            lsData%fit(i) = 0.0_double
            
            do j = 1, lsData%noStars
                         
       
                k = karray(j)            
                k1 = k + 1
                k2 = k + 2           
                x1(2) = lsData%x(i) - x(k)
                x1(3) = lsData%y(i) - x(k1)
           
                fit = max(0.0_double, GridInterpolate(x1(2), x1(3), psfGridFit, fx, fy))

           
                lsData%fit(i) = lsData%fit(i) + x(k2) * fit  ! The fit


                ! The derivatives with respect to x and y
                xTimesWeight = x(k2) * lsData%weight(i)
                c(i, k)  = fx * xTimesWeight
                c(i, k1) = fy * xTimesWeight

                c(i, k2) = fit ! The derivative with respect to the height
            end do  

            f(i) = lsData%f(i) - x(1) - lsData%fit(i) ! The residual
                             

        end do        
        return
        if (n .eq. 2) return
    end if

    if (ctrl .eq. 2) then ! Compute the Jacobian


    
        do i = 1, m

            ! The derivatives with respect to the background             
            c(i, 1) = -lsdata%weight(i)   ! The derivative with respect to the background                   


            !**************************************************************
            ! The derivatives with respect to the remaining stars
            !*************************************************************        
            do j = 1, lsData%noStars
                                
                k = karray(j)            
                k1 = k + 1
                k2 = k + 2
                x1(2) = lsData%x(i) - x(k)
                x1(3) = lsData%y(i) - x(k + 1)
                fit = max(0.0_double, GridInterpolate(x1(2), x1(3), psfGridFit, fx, fy))
                xTimesWeight = x(k2) * lsData%weight(i)
            
                c(i, k) =  fx * xTimesWeight

                c(i, k1) = fy * xTimesWeight

                c(i, k2) = -fit * lsdata%weight(i) ! The derivative with respect to the height
            end do  
           
        end do

        return
    
    end if


    do i = 1, m  ! Loop through the pixels 
        lsData%fit(i) = 0.0_double            
        minRadius = 100.0_double
        do j = 1, lsData%noStars   ! Loop through all the stars in the fit
            k = karray(j)            
            k1 = k + 1
            k2 = k + 2          
            dx = lsData%x(i) - x(k)
            dy = lsData%y(i) - x(k1)

            lsData%fit(i) = lsData%fit(i) + x(k2) * max(0.0_double, GridInterpolate(dx, dy, psfGridFit))    
            radius = sqrt(dx * dx + dy * dy)
            minRadius = min(minRadius, radius)


        end do
        if (minRadius .gt. 8.0) then
            f(i) = 0.0_double
        else  
           
            f(i) = (lsData%f(i) - x(1) - lsData%fit(i)) * lsData%weight(i) !  * (8.0 - radius) 
        end if
 
    end do  
!    fit = maxval(f)
!    f = f / fit

        
end subroutine PSFMultiFitFunction      
!***********************************************************
!
! This function fits one or more stars to a background
! and the psf
! The background is held fixed
!
!**********************************************************
subroutine PSFMultiFitFunction2(x, n, f, m, ctrl, c)
    real(double), intent(in) :: x(:)
    integer, intent(in) :: n
    real(double), intent(out) :: f(:)
    integer, intent(in) :: m
    integer, intent(inout) :: ctrl
    real(double), pointer :: c(:, :)
    integer i, j, k, k1, k2
    real(double) :: fit, sumFit
    real(double) :: x1(size(x))
    real(double) :: dx, dy
    real(double) :: fx, fy



                  

    if (ctrl .eq. -5) then ! Want the unweighted residuals and the derivatives
        x1(1) = x(1)
        do i = 1, m

            ! The derivatives with respect to the background             
            c(i, 1) = 1.0_double  ! The derivative with respect to the background                   


            !**************************************************************
            ! The derivatives with respect to the remaining stars
            !*************************************************************
            sumFit = 0.0_double
            do j = 1, lsData%noStars
                         
       
                k = karray(j)            
                k1 = k + 1
                k2 = k + 2           
                x1(2) = lsData%x(i) - x(k)
                x1(3) = lsData%y(i) - x(k1)

                fit = max(0.0_double, GridInterpolate(x1(2), x1(3), psfGridFit))

                x1(4) = x(k2)
                sumFit = sumFit + x1(4) * fit  ! The fit

                c(i, k) = x1(4) * GridXDerivative(x1(2), x1(3), psfGridFit) * lsData%weight(i)

                c(i, k1) = x1(4) * GridYDerivative(x1(2), x1(3), psfGridFit) * lsData%weight(i)

                c(i, k2) = fit ! The derivative with respect to the height
            end do  

            f(i) = lsData%f(i) - x1(1) - sumFit ! The residual

                         

        end do
        
        return
    end if



    if (ctrl .eq. 2) then ! Compute the Jacobian

        x1(1) = x(1)
        do i = 1, m

           
            !**************************************************************
            ! The derivatives with respect to the remaining stars
            !*************************************************************        
            do j = 1, lsdata%noStars
                                
                k = karray(j)           
                k1 = k + 1
                k2 = k + 2            
                x1(2) = lsData%x(i) - x(k)
                x1(3) = lsData%y(i) - x(k1)

                fit = max(0.0_double, GridInterpolate(x1(2), x1(3), psfGridFit, fx, fy))
      
                x1(4) = x(k2) 
                           
                c(i, k - 1) = x1(4) * fx  * lsData%weight(i)

                c(i, k) = x1(4) * fy * lsData%weight(i)

                c(i, k1) = -fit * lsdata%weight(i) ! The derivative with respect to the height
            end do  
           
        end do

        return
write(*, *) n     
    end if



    do i = 1, m  ! Loop through the pixels 
        fit = 0.0_double            
     
        do j = 1, lsData%noStars   ! Loop through all the stars in the fit
            k = karray(j)           
            k1 = k + 1
            k2 = k + 2          
            dx = lsData%x(i) - x(k)
            dy = lsData%y(i) - x(k1)

            fit = fit + x(k2) * max(0.0_double, GridInterpolate(dx, dy, psfGridFit))    


        end do

          
        f(i) = (lsData%f(i) - x(1) - fit) * lsData%weight(i) 

 
    end do  
 

        
end subroutine PSFMultiFitFunction2      


!************************************************************
!
!
!************************************************************
real(double) Function PSFStars(x, nostars, xpos, ypos)
    real(double) :: x(:)
    integer(int8), intent(in) :: noStars
    real(double), intent(in) :: xpos, ypos
    real(double) :: dx, dy
    integer :: i, k

    psfstars = 0.0_double

    do i = 1, noStars   ! Loop through all the stars in the fit
        k = (i - 1) * 3 + 2            
          
        dx = x(k) - xpos
        dy = x(k + 1) - ypos 
        psfStars = psfStars + x(k + 2) * max(0.0_double, GridInterpolate(dx, dy, psfGridFit))    
       

   
    end do

    psfStars = psfStars + x(1)
end function  PSFStars




!***************************************************
! This subroutine is fitting the psf to a single object
! with no background
! x(1) = xpos
! x(2) = ypos
! x(3) = psf height
!**************************************************
subroutine FitPSFFunction1(source, image, backgroundImage, binX, dx1, psfFitdata)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    
    
    integer, intent(in) :: binX 
    integer, intent(in), optional :: dx1
  
   type(psfFitParameters), intent(out) :: psfFitData

    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(3), bl(3), bu(3)
 
   
    logical(boolean) :: toFit(3)


    real(single) :: mode, sigma, fMax 
    integer(int8) :: dx
    integer :: n   
    real(double) :: constant = 10.0_double
  


    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
 
    write(errstr, '(a, i2)')  "FitPSFFunction1 (No fit to background) Using dx=", dx
    call message(AppMsg, Verbose)
     
!  write(*, *) "test prob chisq=0.0=", Gammaq(dble(100), 0.0_double) * 100.0_double, " %"


    eps = sqrt(EPSILON(0.0_double))

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    ! Get the global background mode
    mode = ComputeBackgroundMode2(image, nx + 1_int16, ny + 1_int16, 2.0, sigma) 

!    psfCounts = -1.0_double
!    call CreatePSF2(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binX, &
!           1, 1,  psfCounts, psfSpline = psfsplineFit)

  

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)

    
    toFit = .true.
    fmax = maxval(image(il : iu, jl : ju) - backgroundImage(il : iu, jl : ju))

    constant = maxval(image(il : iu, jl : ju))
 
    bl(1) = max(0.0_double, source%xPos - dble(4 / binx))
    bu(1) = min(dble(nx + 1), source%xPos + dble(4 / binx))
    x(1) = source%xPos
    
    bl(2) = max(0.0_double, source%yPos - dble(4 / binx))
    bu(2) = min(dble(ny + 1), source%yPos + dble(4 / binx))
    x(2) = source%yPos
    
    bl(3) =  constant / 3.0_double
    bu(3) =  constant * 1.5_double
    x(3) = constant
 
 
    n = 3
    lsData%rMax = dble(dx) + 0.5_double
    call PSFDeconvolve1(1, image, backgroundImage, &
            il, iu, jl, ju, n, x, bl, bu,  toFit, psfFunction2, psfFitData, status)
        
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, PSFFunction2, da, 10)




!    call destructSpline(psfSplineFit)  
  
end subroutine FitPSFFunction1


!***************************************************
! This subroutine is fitting the psf to a single object
! with no background and fixed position
! x(1) = xpos
! x(2) = ypos
! x(3) = psf height
!**************************************************
subroutine FitPSFFunction2(source, image, backgroundImage, binX, dx1, psffitdata)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    
    
    integer, intent(in) :: binX 
    integer, intent(in), optional :: dx1
    type(psfFitParameters), intent(out) :: psfFitData

!    real(single), intent(out) :: redChisq
    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(3), bl(3), bu(3)   
    logical(boolean) :: toFit(3)

    real(single) :: mode, sigma, fMax 
    integer(int8) :: dx
    integer :: n   
    real(double) :: constant = 10.0_double
   


    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
 
    write(errstr, '(a, i2)')  "FitPSFFunction2 (No fit to background and position) Using dx=", dx
    call message(AppMsg, Verbose)
     
!  write(*, *) "test prob chisq=0.0=", Gammaq(dble(100), 0.0_double) * 100.0_double, " %"


    eps = sqrt(EPSILON(0.0_double))

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    ! Get the global background mode
    mode = ComputeBackgroundMode2(image, nx + 1_int16, ny + 1_int16, 2.0, sigma) 

!    psfCounts = -1.0_double
!    call CreatePSF2(1024.5_double, 1024.5_double, 1.0_double, psfNx, psfNy, binX, binX, &
!           1, 1,  psfCounts, psfSpline = psfsplineFit)

  

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)

    
    toFit = .true.
    fmax = maxval(image(il : iu, jl : ju) - backgroundImage(il : iu, jl : ju))

!    if (fMax .gt. constant) then
!        fMax = constant / fMax
!    else
        fMax = 1.0_double
!    end if

    constant = maxval(image(il : iu, jl : ju))
 
    bl(1) = max(0.0_double, source%xPos - dble(4 / binx))
    bu(1) = min(dble(nx + 1), source%xPos + dble(4 / binx))
    x(1) = source%xPos
    
    bl(2) = max(0.0_double, source%yPos - dble(4 / binx))
    bu(2) = min(dble(ny + 1), source%yPos + dble(4 / binx))
    x(2) = source%yPos
    
    bl(3) =  constant / 3.0_double
    bu(3) =  constant * 1.5_double
    x(3) = constant
 
 
    n = 3
    toFit = .true.
    toFit(1) = .false.
    toFit(2) = .false.
    lsData%rMax = dble(dx) + 0.5_double
    call PSFDeconvolve1(1, image, backgroundImage, &
            il, iu, jl, ju, n, x, bl, bu,  toFit, psfFunction2, psfFitData, status)
        
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, PSFFunction2, da, 10)



  
!    call destructSpline(psfSplineFit)  
  
end subroutine FitPSFFunction2



!**************************************************
! 
! This subroutine goes with the above
!
!**************************************************
subroutine PSFDeconvolve1(noStars, image, backgroundImage, &
           il, iu, jl, ju, n, x, &
           bl, bu, toFit, func, psfFitData, status)
    integer(int32), intent(in) :: noStars
    real(single), pointer :: image(:, :), backgroundImage(:, :)
   
   
  
    integer(int16), intent(in) :: il, iu, jl, ju
    integer(int32), intent(in) :: n
    real(double), intent(inout) :: x(1 : n)
    real(double), intent(inout) :: bl(1 : n), bu(1 : n)
    logical(boolean)  :: toFit(:)
    type(psfFitParameters), intent(out) :: psfFitData

    integer(int16), intent(inout) :: status
    integer(int32) :: i, j, m
    real(double), pointer :: f(:)=>Null()
    real(double) :: xPos, yPos
    real(double) :: error
    real(double) :: da(1 : n)
   
      interface
       
    subroutine func(x, n, f, m, ctrl, c)
    use dal
        real(double), intent(in) :: x(:)
        integer, intent(in) :: n
        real(double), intent(out) :: f(:)
        integer, intent(in) :: m
        integer, intent(inout) :: ctrl
        real(double), pointer :: c(:, :)
    end subroutine func

    end interface

    
    m = (iu - il + 1) * (ju - jl + 1)
   
    allocate(lsData%x(1 : m)) 
    allocate(lsData%y(1 : m)) 
    allocate(lsData%f(1 : m)) 
    allocate(lsData%weight(1 : m)) 

    m = 0
    lsData%noStars = noStars
 
    lsData%weight = 0.0_double
   

    do i = il, iu
        xPos = dble(i) + 1.0_double
        
        do j = jl , ju
          
            yPos = dble(j) + 1.0_double
                       
            m = m + 1
            lsData%x(m) = xPos
    
            lsData%y(m) = yPos
           
         
            lsData%f(m) = dble(image(i, j) - backgroundImage(i, j))

            error =  dble(image(i, j) + backgroundImage(i, j))
         
           
       
            
            if(error > 0.0_double) then
                lsData%weight(m) = 1.0_double / error
                lsData%weight(m) = min(1.0_double, lsData%weight(m))
            else
                m = m - 1
            end if
     
   
        end do
    end do
    if (m .lt. 10) then

        write(errstr, '(a, i2, a)') "Too few points for fits (", m, ") fit abandoned"
        call message(AppMsg, Verbose)
        return


    end if

    lsData%m = m

    ! Normalize
   
  
    allocate(f(1 : m))


    write(errstr, '(a)') "Input trial parameters:"
    call message(AppMsg, Verbose)
    do i = 1, n    

        write(errstr, '(a, i2, a, e12.4, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), " bounds-", bl(i), "-", bu(i)      
        call message(AppMsg, Verbose)

    end do
!write(*, *) "Line 996"


    do j = 1, 1
        call nonlin(x, n, m, func, bl, bu, toFit, f, status)
   

        write(errstr, '(a, i1, a, e14.8)') "Iteration ", j, " Sum of squares of residuals=", DOT_PRODUCT(f, f)
        call message(AppMsg, Verbose)
    end do



    write(errstr, '(a, i7)') "status=", status   
    call message(AppMsg, Verbose)



    call ChiSquared(n, x, lsData%m, toFit, func, da, psfFitData%redChisq)



    deallocate(lsData%x) 
    deallocate(lsData%y) 
    deallocate(lsData%f) 
    deallocate(lsData%weight) 
    nullify(lsData%x) 
    nullify(lsData%y) 
    nullify(lsData%f) 
    nullify(lsData%weight) 




!write(*, *) "Line 1013"
   
end subroutine PSFDeconvolve1


!***********************************************************
!
! This is the psf fitting function for a single object 
!
!**********************************************************
subroutine PSFFunction2(x, n, f, m, ctrl, c)
    real(double), intent(in) :: x(:)
    integer, intent(in) :: n
    real(double), intent(out) :: f(:)
    integer, intent(in) :: m
    integer, intent(inout) :: ctrl
    real(double), pointer :: c(:, :)
    integer i, j
    real(double) :: fit
    real(double) :: x1(3), h, error
    real(double) :: dx, dy, weight

    if (ctrl .eq. -5) then


        x1 = x
        

        ! Partial derivatives with respect to the height
        do i = 1, m            
                            
            x1(1) = lsData%x(i) - x(1)
            x1(2) = lsData%y(i) - x(2)           

            c(i, 3) = GridInterpolate(x1(1), x1(2), psfGridFit) 
  

            ! Partial derivatives with respect to x and y
                   

            fit = x1(3) * GridInterpolate(x1(1), x1(2), psfGridFit)    
  

            f(i) = lsData%f(i) - fit

            do j = 1, 2
           
                h = 0.05
                     
                c(i, j) = Derivative(3, x1, j, h, PSFPixel1, derivative3, error)          

            end do 
        end do

        return




    end if

    if (ctrl .eq. 2) then
ctrl = 0
return
        ! Partial derivatives with respect to the background
        do i = 1, m            
            c(i, 1) = lsData%weight(i)
        end do
        ! Partial derivatives with respect to the height
        do i = 1, m            
            c(i, 3) =  PSFIntegration((lsData%x(i) - x(1)), &
                      (lsData%y(i) - x(2))) * lsData%weight(i)                        


        end do


        ! Partial derivatives with respect to x and y
        do i = 1, m            
            do j = 1, 2
                x1(1) = lsData%x(i) - x(1)
                x1(2) = lsData%y(i) - x(2)
           
                h = max(1.0_double, abs(x1(j))) * eps
                    
                c(i, j) = Derivative(3, x1, j, h, PSFPixel1, derivative3, error) * lsData%weight(i)
                            
            end do 
        end do

       
        return
write(*, *) n
    end if


    do i = 1, m           
         dx = lsData%x(i) - x(1)
         dy = lsData%y(i) - x(2)

                   
         weight = lsData%rMax - sqrt(dx * dx + dy * dy)
         weight = max(0.0_double, weight)  
     

  
        fit = x(3) *  GridInterpolate(dx, dy, psfGridFit) 
        f(i) = (lsData%f(i) - fit) * weight
     
    end do  
          
end subroutine PSFFunction2      



!***********************************************************
! 7 parameter 2-d gaussian model
! Fitting a + b * exp(dx * dx / sigx + dy * dy / sigy)
! x(1) = a
! x(2) = dx, x(3) = dy, x(4) = b, x(5) = sigx, x(6) = sigy
! 
!**********************************************************
subroutine NlGaussianFunction1(x, n, f, m, ctrl, c)
    real(double), intent(in) :: x(:)
    integer, intent(in) :: n
    real(double), intent(out) :: f(:)
    integer, intent(in) :: m
    integer, intent(inout) :: ctrl
    real(double), pointer :: c(:, :)
    integer i
    real(double) :: fit

    real(double) :: h, error
    real(double) :: x1(size(x))
    real(double) :: dx, dy

    if (ctrl .eq. -5) then ! Need to compute the partial derivatives


        ! Compute the partial derivatives with respect to x(5), x(6) and  x(7) 
        x1 = x
        do i = 1, m

            x1(2) = lsData%x(i) - x(2)
            x1(3) = lsData%y(i) - x(3)
         

            fit = GaussianIntegration2(x1(2), x1(3), x(5), x(6), x(7))    
                          
            c(i, 1) = 1.0_double ! Derivative with respect to background
        
     
                        
            h = 0.1_double             
            c(i, 2) = x(4) * Derivative(n, x1, 2, h, GaussianPixel2, derivative3, error)  ! Derivative with respect to x  
           
            h = 0.1_double
            c(i, 3) = x(4) * Derivative(n, x1, 3, h, GaussianPixel2, derivative3, error)  ! Derivative with respect to y  
          

            c(i, 4) = fit        ! Derivative with respect to central gaussian peak

            h = 0.05_double
        
            c(i, 5) = x(4) * Derivative(n, x1, 5, h, GaussianPixel2, derivative3, error)  ! Derivative with respect to sigmax  

            h = 0.05_double
        
            c(i, 6) = x(4) * Derivative(n, x1, 6, h, GaussianPixel2, derivative3, error)  ! Derivative with respect to sigmay

            h = 0.01_double
            c(i, 7) = x(4) * Derivative(n, x1, 7, h, GaussianPixel2, derivative3, error)  ! Derivative with respect to theta  
         

            fit = x(1) + x(4) * fit
            f(i) = fit - lsData%f(i) 
        end do



          
        return
    end if

    if (ctrl .eq. 2) then ! Need to compute the Jacobian
ctrl = 0
return

        ! Compute the partial derivatives with respect to x(5), x(6) and  x(7) 
        x1 = x
        do i = 1, m


            ! The derivative with respect to the background
            c(i, 1) = -lsData%weight(i)

            x1(2) = lsData%x(i) - x(2)
            x1(3) = lsData%y(i) - x(3)
         

            ! The derivatives with respect to dx and dy   
            h = 0.1_double             
            c(i, 2) = x(4) * Derivative(n, x1, 2, h, GaussianPixel2, derivative2, error) &
                      * lsData%weight(i) ! Derivative with respect to x  
           
            h = 0.1_double
            c(i, 3) = x(4) * Derivative(n, x1, 3, h, GaussianPixel2, derivative2, error) &
                      * lsData%weight(i) ! Derivative with respect to y  
          
      



            fit = GaussianIntegration2(x1(2), x1(3), x(5), x(6), x(7))    
                          
                      
            c(i, 4) = -fit * lsData%weight(i)  ! Derivative with respect to the Gaussian peak

            h = 0.05_double
        
            c(i, 5) = x(4) * Derivative(n, x1, 5, h, GaussianPixel2, derivative2, error) &
                      * lsData%weight(i) ! Derivative with respect to sigmax  

            h = 0.05_double
        
            c(i, 6) = x(4) * Derivative(n, x1, 6, h, GaussianPixel2, derivative2, error) &
                      * lsData%weight(i) ! Derivative with respect to sigmay

            h = 0.05_double
            c(i, 7) = x(4) * Derivative(n, x1, 7, h, GaussianPixel2, derivative2, error) &
                      * lsData%weight(i) ! Derivative with respect to theta  
         

        end do
    
        return     
    end if  

  
    ! The fit
    do i = 1, m 
      
         
        fit = x(4) * GaussianIntegration2(dx, dy, x(5), x(6), x(7))    
 
         

        f(i) = (lsData%f(i) - (x(1) + fit)) * lsData%weight(i)

        
    end do  
          
end subroutine NlGaussianFunction1      

!*********************************************************
! 
!
!********************************************************
real(double) Function TwoDGaussianFunction(dx, dy, height, sigmax, sigmaY, theta)
    real(double), intent(in) :: dx, dy, height, sigmax, sigmaY, theta


    
    real(double) :: r, delta, xDash, yDash, arg

    r = sqrt(dx * dx + dy * dy)
    delta = atan2(dy, dx)
  
    xDash = r * cos(delta - theta)
    yDash = r * sin(delta - theta)
    arg = xDash * xDash / (-2.0_double * sigmaX * sigmaX) + &

         yDash * yDash / (-2.0_double * sigmaY * sigmaY)

    if (arg .lt. -40.0_double) then
        TwoDGaussianFunction = 0.0_double
    else
        TwoDGaussianFunction = height * exp(arg)
    end if
end function TwoDGaussianFunction




!***************************************************
! Simple 2-d gaussian - sigmax = sigmay
!**************************************************
real(double) Function GaussianIntegration(dx, dy, sigma)
    real(double), intent(in) :: dx, dy, sigma
    sigmaConstant = -2.0_double * sigma * sigma
    
    GaussianIntegration = DoubleIntegration(dx - 0.5_double, dx + 0.5_double, &
                        dy - 0.5_double, dy + 0.5_double, gaussianFunction1)

end function GaussianIntegration
!***************************************************
!
!**************************************************

real(double) Function GaussianFunction1(dx)
    real(double), intent(in) :: dx
    xSave = dx
    GaussianFunction1 = QGauss1(ylSave, yUSave, ZGaussian)

end function GaussianFunction1

!***************************************************
!
!**************************************************
real(double) Function ZGaussian(dy)
    real(double), intent(in) :: dy
    real(double) :: arg
    
    arg = (xSave * xSave + dy * dy)  / sigmaConstant
    
    if (arg .lt. -40.0_double) then
        ZGaussian = 0.0_double
    else
        ZGaussian = exp(arg)
    end if
end function ZGaussian




!***************************************************
!
!**************************************************
real(double) Function GaussianIntegration2(dx, dy, sigmaX, sigmAY, theta)
    real(double), intent(in) :: dx, dy, sigmaX, sigmaY, theta

    sigmaXConstant = -2.0_double * sigmaX * sigmaX
    sigmaYConstant = -2.0_double * sigmaY * sigmaY
  
    pa = theta

    GaussianIntegration2 = DoubleIntegration(dx - 0.5_double, dx + 0.5_double, &
                        dy - 0.5_double, dy + 0.5_double, gaussianFunction2)

!write(*, *) "Line 1694 GaussianIntegration2 dx=", dx, " dy=", dy, " sigmax=", sigmax, &
!   " sigmay=", sigmay, " theta=", theta
end function GaussianIntegration2
!***************************************************
! Evaluates the double integral in the x-direction
!**************************************************
real(double) Function GaussianFunction2(dx)
    real(double), intent(in) :: dx

    xsave = dx
    dxSquared = dx * dx
    
    GaussianFunction2 = QGauss1(ylSave, yUSave, Gaussian)

!write(*, *) "GaussianFunction2 line 1715 dx=", dx, " yl=", ylSAve, " yu=", yuSAve, " GaussianFunction2=", GaussianFunction2
end function GaussianFunction2


!***************************************************
!
!**************************************************
real(double) Function XGaussian(dx)
    real(double), intent(in) :: dx
    real(double) :: arg
    
    arg = dx * dx / sigmaXConstant
    
    if (arg .lt. -30.0_double) then
        XGaussian = 0.0_double
    else
        XGaussian = exp(arg)
    end if
end function XGaussian


!***************************************************
!
!**************************************************
real(double) Function YGaussian(dy)
    real(double), intent(in) :: dy
    real(double) :: arg
    
    arg = dy * dy / sigmaYConstant
    
    if (arg .lt. -40.0_double) then
        YGaussian = 0.0_double
    else
        YGaussian = exp(arg)
    end if
end function YGaussian

!***************************************************
!
!**************************************************
real(double) Function Gaussian(dy)
    real(double), intent(in) :: dy
    
    real(double) :: r, delta, xDash, yDash, arg

    r = sqrt(dxSquared + dy * dy)
    delta = atan2(dy, xSave)
  
    xDash = r * cos(delta - pa)
    yDash = r * sin(delta - pa)
    arg = xDash * xDash / sigmaXConstant + yDash * yDash / sigmaYConstant

    if (arg .lt. -30.0_double) then
        Gaussian = 0.0_double
    else
        Gaussian = exp(arg)
    end if
end function Gaussian

!***************************************************
! 2-d Gaussian fit  
! x(1) = background
! x(2) = x centre! 
! x(3) = y centre
! x(4) = height
! x(5) sigma x
! x(6) = sigma y
! x(7) = position angle
!**************************************************
subroutine Fit2dGaussian2(source, image, binX, levelImage1, backgroundImage, dx1, fitParameters)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :)
    integer(int16), pointer, optional :: levelImage1(:, :)
    real(single), pointer, optional :: backgroundimage(:, :)
    
    integer, intent(in) :: binX 
    integer, intent(in), optional :: dx1
  

    type(GaussFitParameters), intent(out) :: fitParameters
  
    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(7), x1(7), bl(7), bu(7)
   
  
    logical(boolean) :: toFit(7)

    real(single) :: mode, sigma, fMax 
    integer(int8) :: dx
 
    if (associated(backgroundImage)) then

    end if

    call InitialiseGauss(10)
 
    eps = sqrt(EPSILON(0.0_double))
 
    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
    write(errstr, '(a, i2)')  "Fit2dGaussian2 Using dx=", dx
    call message(AppMsg, Verbose)
     
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)




   ! Get the global background mode
    mode = ComputeBackgroundMode2(image, nx, ny, 2.0, sigma) 

    toFit = .true.
    fmax = maxval(image(il : iu, jl : ju))


    
    bl(1) = max(dble(mode - 10.0), 0.0_double)
    bu(1) = (dble(mode) + 10.0_double)
   
    x(1) = mode
    bl(2) = max(0.0_double, source%xPos - dble(4 / binx))
    bu(2) = min(dble(nx + 1), source%xPos + dble(4 / binx))
    x(2) = source%xPos
    
    bl(3) = max(0.0_double, source%yPos - dble(4 / binx))
    bu(3) = min(dble(ny + 1), source%yPos + dble(4 / binx))
    x(3) = source%yPos
 
    lsData%xcentre = x(2)   
   
    lsData%ycentre = x(3)   
   
    bl(4) =  mode + (fmax - mode) / 3.0_double
    bu(4) =  mode + (fmax - mode) * 3.0_double
    x(4) = fmax - mode
 
    x(5) = max(1.0, source%semiA) 
    bl(5) = 0.3_double
    bu(5) = max(5.0, source%semiA * 3.0)
    x(6) = max(1.0, source%semiB) 
    
    bl(6) = 0.3_double
    bu(6) = max(5.0, source%semiB * 3.0)
      
    x(7) = 0.0_double
    bl(7) = 0.0_double
    bu(7) = PI / 2.0_double 

!write(*, *) "x(7)=", x(7)
   
    lsData%rmax = -1.0

    if (source%extensionFlag .eq. int(0, int16)) then
 
 
       call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 7, x, bl, bu,  toFit, NLGaussianFunction1, status, fitParameters)
    else if (present(levelImage1)) then
 
 
        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 7, x, bl, bu,  toFit, NLGaussianFunction1, &
            status, fitParameters, LEVELIMAGE=levelImage1, SOURCENUMBER=source%number)
    else
 
 
        call GaussDeconvolve(1, image,  &
            il, iu, jl, ju, 7, x, bl, bu,  toFit, NLGaussianFunction1, status, fitParameters)
    end if 


    x1 = x


  
    fitParameters%background = x(1)
    fitParameters%xcentre = x(2)
    fitParameters%ycentre = x(3)
    fitParameters%height = x(4)
    fitParameters%sigmax = max(x(5), x(6))
    fitParameters%sigmay = min(x(5), x(6))

    fitParameters%positionAngle = x(7) * radiansToDegrees

  
end subroutine Fit2DGaussian2


!***************************************************
! 2-D Gaussian fit using a background image for 
! background subtraction
! x(1) = x centre
! x(2) = y centre
! x(3) = height
! x(4) = sigma x
! x(5) = sigma y
! x(6) = position angle
!
!**************************************************
subroutine Fit2dGaussian3(source, image, binX, levelImage1, &
         bckgroundImage, dx1, fitParameters)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :)
    integer(int16), pointer, optional :: levelImage1(:, :)
    real(single), pointer  :: bckgroundimage(:, :)
    
    integer, intent(in) :: binX 

    integer, intent(in), optional :: dx1
    type(GaussFitParameters), intent(out) :: fitParameters
 
 
  
    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(6), x1(6), bl(6), bu(6)
    logical(boolean) :: toFit(6)
    real(single) :: fMin, fMax 
    integer(int8) :: dx
    integer :: n  
   
  
    call InitialiseGauss(10)
 


    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
 
    write(errstr, '(a, i2)')  "Fit2dGaussian3 Using dx=", dx
    call message(AppMsg, Verbose)
     



    nx = ubound(image, 1)
    ny = ubound(image, 2)

   

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)

    
    toFit = .true.
    fmin = minval(image(il : iu, jl : ju) - bckgroundImage(il : iu, jl : ju))
    fmin = max(0.0, fmin)
    fmax = maxval(image(il : iu, jl : ju) - bckgroundImage(il : iu, jl : ju))




    
    bl(1) = max(0.0_double, source%xPos - dble(4 / binx))
    bu(1) = min(dble(nx + 1), source%xPos + dble(4 / binx))

    x(1) = source%xPos
    
    bl(2) = max(0.0_double, source%yPos - dble(4 / binx))
    bu(2) = min(dble(ny + 1), source%yPos + dble(4 / binx))
    x(2) = source%yPos
 
    lsData%xcentre = x(1)   
   
    lsData%ycentre = x(2)   
  
   
    bl(3) =  fmin
    bu(3) =  fmax * 3.0_double
    x(3) = fmax
 
    x(4) = max(1.0, source%semiA) 
    bl(4) = 0.3
    bu(4) = max(source%semiA * 3.0, 8.0)
    x(5) = max(1.0, source%semiB) 
    bl(5) = 0.3
    bu(5) = max(source%semiB * 3.0, 8.0)
  
    
    x(6) = 0.0_double
    bl(6) = 0.0_double
    bu(6) = PI / 2.0_double 

    lsData%rmax = -1.0

    if (source%extensionFlag .eq. 0) then

        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 6, x, bl, bu,  toFit, NLGaussianFunction2, status, fitParameters, bckgroundimage)
    else if (present(levelImage1)) then
        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 6, x, bl, bu,  toFit, NLGaussianFunction2, &
            status, fitParameters, bckgroundImage, levelImage=levelImage1, sourcenumber=source%number)
    else

        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 6, x, bl, bu,  toFit, NLGaussianFunction2, status, fitParameters, bckgroundimage)
    end if 


    x1 = x

n = 6


!if (mod(count, 10) .eq. 0) then  
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, NLGaussianFunction2, da, 10)

!write(3, '(a, f6.2, a, f6.2, a, f6.2, a, f6.2)') "Gaussian3 Fit nsims=10 sigx=", x(4), " +/- ", da(4),  &
!            " Fit sigy=", x(5), " +/- ", da(5)
 
    x = x1
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, NLGaussianFunction2, da, 10)

 


!end if

!    call ChiSquared(n, x, lsData%m, toFit,  NLGaussianFunction2, da, fitParameters%redChisq)


    fitParameters%background = 0.0 
    fitParameters%xcentre = x(1)
    fitParameters%ycentre = x(2)
    fitParameters%height = x(3)
    fitParameters%sigmax = max(x(4), x(5))
    fitParameters%sigmay = min(x(4), x(5))

    fitParameters%positionAngle = x(6) * radiansToDegrees

  
end subroutine Fit2DGaussian3

!***************************************************
! 2-d GaussianFit but not fitting x and y
!
!**************************************************
subroutine Fit2dGaussian4(source, image, binX, levelImage1, &
          backgroundImage, dx1, fitparameters)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :)
    integer(int16), pointer, optional :: levelImage1(:, :)
    real(single), pointer, optional :: backgroundimage(:, :)
    
    integer, intent(in) :: binX 
    integer, intent(in), optional :: dx1
    type(GaussFitParameters), intent(out) :: fitParameters

  
    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(7), x1(7), bl(7), bu(7)
   
  
    logical(boolean) :: toFit(7)

    real(single) :: mode, sigma, fMax 
    integer(int8) :: dx
    integer :: n   

    if (associated(backgroundimage)) then

    end if

    call InitialiseGauss(10)
 

 
    eps = sqrt(EPSILON(0.0_double))

    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
 
    write(errstr, '(a, i2)')  "Fit2dGaussian4 (Fixed coordinates) Using dx=", dx
    call message(AppMsg, Verbose)
     



    nx = ubound(image, 1)
    ny = ubound(image, 2)

    ! Get the global background mode
    mode = ComputeBackgroundMode2(image, nx + 1_int16, ny + 1_int16, 2.0, sigma) 

  

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)


    
    toFit = .true.
    fmax = maxval(image(il : iu, jl : ju))

  
    
    bl(1) = max(dble(mode - 5.0), 0.0_double) * fMax
    bu(1) = (dble(mode) + 5.0_double) * fMax
   
    x(1) = mode * fMax
    bl(2) = source%xPos - dble(4 / binx)
    bu(2) = source%xPos + dble(4 / binx)
    x(2) = source%xPos
    
    bl(3) = source%yPos - dble(4 / binx)
    bu(3) = source%yPos + dble(4 / binx)
    x(3) = source%yPos
 
    lsData%xcentre = x(2)   
   
    lsData%ycentre = x(3)   
  
   
    bl(4) =  fmax / 3.0_double
    bu(4) =  fmax * 3.0_double
    x(4) = fmax
 
    x(5) = max(1.0, source%semiA) 
    bl(5) = max(0.3, source%semiA / 3.0)
    bu(5) = max(7.0, source%semiA * 3.0)
    x(6) = max(1.0, source%semiB) 
    bl(6) = max(0.3, source%semiB / 3.0)
    bu(6) = max(7.0, source%semiB * 3.0)
  
    
    x(7) = 0.0_double
    bl(7) = 0.0_double
    bu(7) = PI / 2.0_double 

    toFit(2) = .false.
    toFit(3) = .false.

    lsData%rmax = -1.0


!do i = 1, 7
!write(*, *) i, " x=", x(i), " bl=", bl(i), " bu=", bu(i)
!end do
    n = 7
    if (source%extensionFlag .eq. int(0, int16)) then
        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, n, x, bl, bu,  toFit, NLGaussianFunction1, status, fitParameters)
    else if (present(levelImage1)) then
        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, n, x, bl, bu,  toFit, NLGaussianFunction1, &
            status, fitParameters, LEVELIMAGE=levelImage1, SOURCENUMBER=source%number)
    else

        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, n, x, bl, bu,  toFit, NLGaussianFunction1, status, fitParameters)
    end if 

    x1 = x

   
!write(*, *) "Line 1601 calling computefiterrors"

!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, NLGaussianFunction1, da, 10)

 
!    call ChiSquared(n, x, lsData%m, toFit, NLGaussianFunction1, da, fitparameters%redChisq)
  


    fitParameters%background = x(1)
    fitParameters%xcentre = source%xpos
    fitParameters%ycentre = source%ypos
    fitParameters%height = x(4)
    fitParameters%sigmax = max(x(5), x(6))
    fitParameters%sigmay = min(x(5), x(6))

    fitParameters%positionAngle = x(7) * radiansToDegrees

end subroutine Fit2DGaussian4


!***************************************************
! 2-D Gaussian fit using a background image for 
! background subtraction and not fitting
! source coords
! x(1) = x centre
! x(2) = y centre
! x(3) = height
! x(4) = sigma x
! x(5) = sigma y
! x(6) = position angle
!
!**************************************************
subroutine Fit2dGaussian5(source, image, binX, levelImage1, &
         bckgroundImage, dx1, fitParameters)
    type(sourceParameters), intent(inout):: source
    real(single), pointer :: image(:, :)
    integer(int16), pointer, optional :: levelImage1(:, :)
    real(single), pointer  :: bckgroundimage(:, :)
    
    integer, intent(in) :: binX 

    integer, intent(in), optional :: dx1
    type(GaussFitParameters), intent(out) :: fitParameters
 
 
  
    integer(int16) :: i0, j0, il, iu, jl, ju 
    integer(int16) :: status, nx, ny
    real(double) :: x(6), x1(6), bl(6), bu(6)
    logical(boolean) :: toFit(6)
    real(single) :: fMin, fMax 
    integer(int8) :: dx
    integer :: n  
  

    call InitialiseGauss(10)
 

    if (present(dx1)) then
        dx = dx1
    else 
        dx = 14 / binx
    end if
  
 
    write(errstr, '(a, i2)')  "Fit2dGaussian5- Fixed background and coordinates Using dx=", dx
    call message(AppMsg, Verbose)
     



    nx = ubound(image, 1)
    ny = ubound(image, 2)

   

    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)

    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)
    
    toFit = .true.
    fmin = minval(image(il : iu, jl : ju) - bckgroundImage(il : iu, jl : ju))

    fmax = maxval(image(il : iu, jl : ju) - bckgroundImage(il : iu, jl : ju))




    
    bl(1) = max(0.0_double, source%xPos - dble(4 / binx))
    bu(1) = min(dble(nx + 1), source%xPos + dble(4 / binx))

    x(1) = source%xPos
    
    bl(2) = max(0.0_double, source%yPos - dble(4 / binx))
    bu(2) = min(dble(ny + 1), source%yPos + dble(4 / binx))
    x(2) = source%yPos
 
    lsData%xcentre = x(1)   
   
    lsData%ycentre = x(2)   
  
   
    bl(3) =  fmin
    bu(3) =  fmax * 3.0_double
    x(3) = fmax
 
    x(4) = max(1.0, source%semiA) 
    bl(4) = 0.3
    bu(4) = max(source%semiA * 3.0, 5.0)
    x(5) = max(1.0, source%semiB) 
    bl(5) = 0.3
    bu(5) = max(source%semiB * 3.0, 5.0)
  
    
    x(6) = 0.0_double
    bl(6) = 0.0_double
    bu(6) = PI / 2.0_double 


    toFit = .true.
    toFit(1) = .false.
    toFit(2) = .false.

    lsData%rmax = -1.0

    if (source%extensionFlag .eq. int(0, int16)) then

        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 6, x, bl, bu,  toFit, NLGaussianFunction2, status, fitParameters, bckgroundimage)
    else if (present(levelImage1)) then
        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 6, x, bl, bu,  toFit, NLGaussianFunction2, &
            status, fitParameters, bckgroundImage, levelImage=levelImage1, sourcenumber=source%number)
    else

        call GaussDeconvolve(1, image, &
            il, iu, jl, ju, 6, x, bl, bu,  toFit, NLGaussianFunction2, status, fitParameters, bckgroundimage)
    end if 


    x1 = x

n = 6


!if (mod(count, 10) .eq. 0) then  
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, NLGaussianFunction2, da, 10)

!write(3, '(a, f6.2, a, f6.2, a, f6.2, a, f6.2)') "Gaussian3 Fit nsims=10 sigx=", x(4), " +/- ", da(4),  &
!            " Fit sigy=", x(5), " +/- ", da(5)
 
    x = x1
!    call computeFitErrors(n, lsData%m, x, bl, bu, toFit, NLGaussianFunction2, da, 10)

 


!end if

!    call ChiSquared(n, x, lsData%m, toFit,  NLGaussianFunction2, da, fitParameters%redChisq)


    fitParameters%background = 0.0 
    fitParameters%xcentre = x(1)
    fitParameters%ycentre = x(2)


    fitParameters%height = x(3)
    fitParameters%sigmax = max(x(4), x(5))
    fitParameters%sigmay = min(x(4), x(5))

    fitParameters%positionAngle = x(6) * radiansToDegrees


end subroutine Fit2DGaussian5

!***********************************************************
! Two dimensional gaussian fitting with no background
! Fitting b * exp(dx * dx / sigx + dy * dy / sigy)
! x(1) = dx
! x(2) = dy, 
! x(3) = height, x(4) = sigx, x(5) = sigy, x(6) = pa
! NB NO BACKGROUND
! 
!**********************************************************
subroutine NlGaussianFunction2(x, n, f, m, ctrl, c)
    real(double), intent(in) :: x(:)
    integer, intent(in) :: n
    real(double), intent(out) :: f(:)
    integer, intent(in) :: m
    integer, intent(inout) :: ctrl
    real(double), pointer :: c(:, :)
    integer i
    real(double) :: fit
    real(double) :: h, error
    real(double) :: x1(size(x))


    if (ctrl .eq. -5) then ! Need to compute the partial derivatives


        x1 = x

        do i = 1, m
           
            ! Compute the partial derivative with respect to dx and dy)  
            x1(1) = lsData%x(i) - x(1)
            x1(2) = lsData%y(i) - x(2)
            h = 0.2_double
                        
            c(i, 1) = x(3) * Derivative(n, x1, 1, h, GaussianPixel3, derivative3, error)
   
            h = 0.2_double                        
            c(i, 2) = x(3) * Derivative(n, x1, 2, h, GaussianPixel3, derivative3, error)
   


            ! Compute the partial derivative with respect to x(3) ! the height 
            fit = GaussianIntegration2(x1(1), x1(2), x(4), x(5), x(6))         

            c(i, 3) = fit    
    
       
            ! Compute the partial derivatives with respect to x(4), x(5) and  x(6) 
        
             
            f(i) = lsData%f(i)  - x1(3) * fit
    
                
    
            h = 0.2_double                                              
            c(i, 4) = x(3) * Derivative(n, x1, 4, h, GaussianPixel3, derivative3, error)
      
            h = 0.2_double                                              
            c(i, 5) = x(3) * Derivative(n, x1, 5, h, GaussianPixel3, derivative3, error)
 
            h = 0.2_double                                              
            c(i, 6) = x(3) * Derivative(n, x1, 6, h, GaussianPixel3, derivative3, error)
      

        end do
         
          
        return
    end if



    if (ctrl .eq. 2) then ! Need to compute the Jacobian
ctrl = 0
return
 
        x1 = x

        do i = 1, m
            

            ! Compute the partial derivative with respect to dx and dy  
            x1(1) = lsData%x(i) - x(1)
            x1(2) = lsData%y(i) - x(2)
            h = 0.1_double
                        
            c(i, 1) = x(3) * Derivative(n, x1, 1, h, GaussianPixel3) * lsData%weight(i)
    
            h = 0.1_double                        
            c(i, 2) = x(3) * Derivative(n, x1, 2, h, GaussianPixel3) * lsData%weight(i)
    
 
            ! Compute the partial derivative with respect to x(3) ! the height 
            c(i, 3) = -GaussianIntegration2(x1(1), x1(2), x(4), x(5), x(6)) * lsData%weight(i)    

            h = 0.05_double
            c(i, 4) = x(3) * Derivative(n, x1, 4, h, GaussianPixel3) * lsData%weight(i)
 
            h = 0.05_double                        
            c(i, 5) = x(3) * Derivative(n, x1, 5, h, GaussianPixel3) * lsData%weight(i)
    
            h = 0.08_double                        
            c(i, 6) = x(3) * Derivative(n, x1, 6, h, GaussianPixel3) * lsData%weight(i)
    
        end do       

        return     
    end if


    do i = 1, m 
      

        fit = x(3) * GaussianIntegration2(lsData%x(i) - x(1), lsData%y(i) - x(2) , x(4), x(5), x(6))    

        f(i) = (lsData%f(i) - fit) * lsData%weight(i)

    

    end do  
          
end subroutine NlGaussianFunction2      

!********************************************************
!  Deconvolution using psf function
!
!
!*******************************************************
subroutine testdeconvolution(source, image, backgroundimage, frameTime, binXIn)
    type(sourceParameters), pointer :: source(:)
    
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    real(double), intent(in) :: frameTime
    integer, intent(in), optional :: binXIn
    integer :: i, j, count = 1
    real(single) :: dx, dy, radius
    integer, parameter :: maxCount = 10
    integer :: binX = 2, binY = 2
    real(single) :: maxRadius
    real(single) ::  psfCounts
    integer :: psfNx = 12, psfNy = 12
 

    type(sourceParameters), pointer :: tempSource(:) => null()
    integer :: sourceIndex(maxcount)
    integer, pointer :: sortIndex(:) => null()
    integer, pointer :: inversesortIndex(:) => null()

    type(sourceParameters), pointer :: tempDetectedSource(:) => null()
    type(sourceParameters), pointer :: detectedsource(:) => null()

    logical, pointer :: recomputed(:) => null()
    integer :: nSources


    if (.not. associated(source)) return
    nSources = size(source)
    if (nsources .lt. 2) return

    write(errstr, '(a)') "PSF multi-star fitting"
    call message(AppMsg, Verbose)

 

    if (present(binxin)) then
        binx = binxIn
        biny = binx
    end if


    psfCounts = -1.0_double

    call CreatePSFImage21(psfNx, psfNy, xpos=1024.5_double, ypos=1024.5_double, &
          ratio=0.5_double, bin=binX, res=1, totalcounts=psfcounts, psfgridfitout=psfgridFit) 

!write(*, *) "TESTDECONVOLUTION LINE 2930 interpolated value at 0, 0=", gridinterpolate(0.0, 0.0, psfgridfit)
!write(*, *) "psfnx=", psfnx, " psfny=", psfny

 

    maxRadius = real(18 / binX)
    allocate(tempsource(1 : maxCount))
    allocate(tempDetectedsource(1 : maxCount))
    allocate(Detectedsource(1 : nSources))
    allocate(recomputed(1 : nSources))

    sortIndex => allocateInt32(1, nSources, "sortIndex")
    inversesortIndex => allocateInt32(1, nSources, "inversesortIndex")


    !*******************************************************
    ! Obtain an index of the source positions in
    ! increasing x-pixel position. NB xSource is not
    ! sorted. sortIndex(i) will give the index of the 
    ! array element that gives the positions in increasing
    ! x-position. This is used for obtaining the distance
    ! from one source to its nearest neighbour. 
    !********************************************************
    call IndexSort(detectedSource%xPos, sortIndex, size(detectedSource))  


    !*********************************************************
    ! Sort the sources into increasing x order
    !*********************************************************
    do i = 1, nsources
        detectedSource(sortIndex(i)) = source(i)
        inverseSortIndex(sortIndex(i)) = i ! To go from sorted to unsorted
    end do 

    recomputed = .false.
 

    !**************************************************************
    ! Loop through all the sources
    !*************************************************************
    do i = 1, nSources
       
if (i .ne. 6) cycle

!if (i .ne. 1601 .and. i .ne. 1) cycle        

        if (detectedSource(i)%extensionFlag .gt. 0) cycle
!        if (detectedSource(i)%confusionFlag .eq. 1) cycle

      
        if (recomputed(i)) cycle    ! Already been used
        recomputed(i) = .true.
        count = 1
        tempSource(count) = detectedSource(i)   ! First source in list of stars to be fitted
        tempDetectedSource(count) = detectedSource(i)     
        sourceIndex(count) = i

        maxRadius = real(18 / binX) + ceiling(log10(detectedSource(i)%significance))
  
        !**************************************************************
        ! 
        ! Find out how many close neighbours this star has to its left
        !**************************************************************
        if (i .gt. 1) then
   
            do j = i - 1, 1, -1
   
!               if (detectedSource(j)%extensionFlag .gt. 0) cycle ! Skip extended sources
    
                dx = detectedSource(i)%xPos - detectedSource(j)%xpos

                if (abs(dx) .gt. maxRadius) exit
                dy = detectedSource(i)%yPos - detectedSource(j)%ypos
                if (abs(dy) .gt. maxRadius) cycle
        
                radius = sqrt(dx * dx + dy * dy)
                if (radius .gt. maxRadius) cycle
                
                count = count + 1   ! Number of stars
                recomputed(j) = .true.
               write(*, *) "Line 2463 i=", i, " j=", j
                tempSource(count) = detectedSource(j)             
   
                tempDetectedSource(count) = detectedSource(j)     

                if (count .eq. maxCount) exit  ! Reached maximum allowed
                sourceIndex(count) = j
             end do


        end if


        !***************************************************************
        ! 
        ! Find out how many close neighbours this star has to its right
        !***************************************************************
        if (i .lt. nSources .and. count .lt. maxCount) then
            do j = i + 1, nSources
      
!               if (detectedSource(j)%extensionFlag .gt. 0) cycle ! Skip extended sources
    
                dx = detectedSource(i)%xPos - detectedSource(j)%xpos

        
!                if (abs(dx) .gt. maxRadius) exit
                dy = detectedSource(i)%yPos - detectedSource(j)%ypos
                if (abs(dy) .gt. maxRadius) cycle
        
                radius = sqrt(dx * dx + dy * dy)
                if (radius .gt. maxRadius) cycle
            
                count = count + 1   ! Number of stars
                recomputed(j) = .true.
            
                tempSource(count) = detectedSource(j)             
   
                tempDetectedSource(count) = detectedSource(j)     

                if (count .eq. maxCount) exit  ! Reached maximum allowed
                sourceIndex(count) = j
            end do
        end if




        !*********************************************************************
        !
        !
        !********************************************************************* 
!        if ( (count .gt. 1 .and. count .lt. 4) .or. (count .eq. 1 .and. tempSource(1)%confusionFlag .eq. 4)) then
     if ( count .gt. 0 .and. count .lt. 4) then

!   if (count .eq. 1) then
            if (count .eq. 2) then
                write(errstr, '(a, i4, a, i2, a)') "Source ", i, " has one neighbour"      
            else
                write(errstr, '(a, i4, a, i2, a)') "Source ", i, " has ", (count - 1), " neighbours"      
         
            end if
            call message(AppMsg, Verbose)



            
            call stardecon_deconvolve3(binX, count, tempSource, &
                           image, backgroundImage, dble(psfCounts))


            do j = 1, count
                write(errstr, '(a, i6)') "Star no ", j
                call message(AppMsg, Verbose)
                  
                write(errstr, '(4(a, f10.2))') " Original detected x=", &
                     tempDetectedSource(j)%xpos, " new ", tempSource(j)%xPos

                call message(AppMsg, Verbose)
           
                write(errstr, '(4(a, f10.2))') " Original detected y=", &
                      tempDetectedSource(j)%ypos, " new ", tempSource(j)%yPos

                call message(AppMsg, Verbose)
           
                write(errstr, '(4(a, f10.2))') " Original raw count rate", &
                      tempDetectedSource(j)%rawCountRate, " new ", tempSource(j)%rawCountRate
      

                call message(AppMsg, Verbose)
                write(errstr, '(4(a, f10.2))') " Original raw background count rate", &
                      tempDetectedSource(j)%rawbackgroundCountRate, " new ", tempSource(j)%rawbackgroundCountRate
      
                call message(AppMsg, Verbose)

                     
                ! Correct counts for coincidence losses
                ! NB The corrected count rates are stored in source
                ! members correctedCountRate, correctedCountRateErr,
                ! and correctedBackgroundRateErr   
                tempSource(j)%apertureRadius = 12.0
                call PointSourceCoincLossCorr(tempSource(j), 12.0, frameTime)     



                ! Compute the signal-to-noise (source counts / sqrt(source + background counts)   
                tempSource(j)%Significance = & 
                         ComputeSourceSignificance(tempSource(j)%corrCountRate, &
                             tempSource(j)%corrBackGroundCountRate, tempSource(j)%exposure)
              
                detectedsource(sourceIndex(j)) = tempSource(j)
                source(inversesortIndex(sourceindex(j))) = tempSource(j)

  
            end do
              
        end if

    end do

    call GridInterpolantDestruct(psfgridFit)
    deallocate(detectedSource, tempSource, tempDetectedSource, recomputed, sortindex, inversesortindex)
    nullify(detectedSource, tempSource, tempDetectedSource, recomputed, sortindex, inversesortindex)
  

end subroutine testdeconvolution

!**************************************************************
! Fits a model consisting of one or more psfs to the image data
!
!**************************************************************
subroutine stardecon_deconvolve3(binX, nStars, source, image, &
                    backgroundImage, psfCounts)
    integer(int32), intent(in) :: binX, nStars
    
    type(SourceParameters), pointer :: source(:)

    real, pointer, dimension(:, :) :: image, backgroundImage
    real(double), intent(in) :: psfCounts  
   
    real(double) :: x(50), bl(50), bu(50)
    
    integer(int32) :: i, j, k, n
    integer(int32) :: il, iu, jl, ju, nx, ny
   
    integer(int32) :: width
 
    real(single) :: mode, fMin, fMax
  
    real(double) :: xCentre, yCentre
    real(single) ::  xMin, xMax, yMin, yMax, dx
    type(GaussFitParameters) :: fitParameters
    integer(int16) :: status
    

    write(errstr, '(a, i2, a)') "Deconvolving ", nStars, " stars"
    call message(AppMsg, Verbose)

    ! Find the centre of the fitting region
    xCentre = 0.0_double
    yCentre = 0.0_double
    xMin = source(1)%xPos
    xMax = source(1)%xPos
    yMin = source(1)%yPos
    yMax = source(1)%yPos

    do i = 1, nStars
        xCentre = xCentre + source(i)%xPos
        yCentre = yCentre + source(i)%ypos
        xmin = min(xMin, source(i)%xPos)
        xmax = max(xMax, source(i)%xPos)
        ymin = min(yMin, source(i)%yPos)
        ymax = max(yMax, source(i)%yPos)

    end do

    xCentre = xCentre / dble(nStars) ! The x centre of the fit region
    yCentre = yCentre / dble(nStars) ! The y centre of the fit region


    do i = 1, nStars
        write(errstr, '(a, i2, a,  f7.2, a, f7.2)') &
            "Centroided x of source ", i, " = ", source(i)%xPos, " y=", source(i)%yPos      
        call message(AppMsg, Verbose)
    end do
   
    nx = size(image, 1)
    ny = size(image, 2)
  

    width = 40 / binX 
    il = max(0, int(xmin - 0.5) - width)
    iu = min(nx - 1, ceiling(xmax - 0.5) + width)
    

    jl = max(0, int(ymin - 0.5) - width)
    ju = min(ny - 1, ceiling(ymax - 0.5) + width)

    mode = minval(backgroundImage(il : iu, jl : ju), backgroundImage(il : iu, jl : ju) .ge. 0.0)

    width = 12 / binX 
    il = max(0, int(xmin - 0.5) - width)
    iu = min(nx - 1, ceiling(xmax - 0.5) + width)
    

    jl = max(0, int(ymin - 0.5) - width)
    ju = min(ny - 1, ceiling(ymax - 0.5) + width)

  
    fMin = minval(image(il : iu, jl : ju), image(il : iu, jl : ju) .ge. 0.0)
 
    fMax = maxval(image(il : iu, jl : ju))


 
  

    !**********************************************************
    ! Background 
    !*********************************************************    
    bl(1) = max(0.0, mode - 2.0 * sqrt(mode))
    bu(1) = max(2.0, mode + 2.0 * sqrt(mode))
    x(1) = mode   ! Estimate of the background

   

    !***********************************************************
    ! The parameters of the stars
    !**********************************************************
    dx = real(4 / binx)

    do j = 1, nStars
        k = (j - 1) * 3 + 2
        x(k) = source(j)%xPos          ! X - coordinate
        bl(k) = x(k) - dx
        bu(k) = x(k) + dx

        fmin = 0.25 * minval(image(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax), &
                 image(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax) .gt. 0.0 )

        fmax = maxval(image(source(j)%imin : source(j)%imax, source(j)%jmin : source(j)%jmax)) - mode
      

        k = k + 1
        x(k) = source(j)%yPos           ! Y - coordinate 
        bl(k) = x(k) - dx
        bu(k) = x(k) + dx
        k = k + 1
        x(k) = fmax
       
        bl(k) = max(fmin, mode)
        bu(k) = min(50000.0, fmax * 3.0) 
    
    end do 

    n = nStars * 3 + 1 ! The total number of parameters



    !*********************************************
    ! Do the actual fitting
    !********************************************
    call Deconvolve(nStars, image, il, iu, jl, ju, n, x, bl, bu, &
                 PsfMultiFitFunction, status, fitparameters)
 

    source(1)%rawBackgroundCountRate = x(1) * PI * real(144  / &
                  (binX * binX)) / source(1)%exposure



    ! Compute the new count-rates from the fitted parameters
   
    do j = 1, nStars
        k = (j - 1) * 3 + 2
     
        source(j)%xPos = x(k)
 
        k = k + 1
        source(j)%yPos = x(k)
  
        ! Raw count-rate
        k = k + 1
   
        source(j)%rawCountRate = x(k) * psfCounts / source(j)%exposure
 
        ! Background count rate
        source(j)%rawBackgroundCountRate = source(1)%rawBackgroundCountRate 


        ! The new signal-to-noise ratio
        source(j)%significance = x(k) * sqrt(PI * real(144 / binx * binx) * source(1)%exposure)  / sqrt(x(k) + x(1))
        source(j)%apertureRadius = 12.0
        
    end do

  
end subroutine stardecon_deconvolve3
!**************************************************
! Deconvolve
!
!
!**************************************************
subroutine Deconvolve(noStars, image,  &
           il, iu, jl, ju, n, x, bl, bu, func, status, fitParameters)
    integer(int32), intent(in) :: noStars
    real(single), pointer :: image(:, :)
  

    integer(int32), intent(in) :: il, iu, jl, ju
    integer(int32), intent(in) :: n
    real(double), intent(inout) :: x(1 : n)
    real(double), intent(inout) :: bl(1 : n), bu(1 : n)
    logical(boolean)  :: toFit(1 : n)
    integer(int16), intent(inout) :: status
    type(GaussFitParameters), intent(out) :: fitParameters

    integer(int16) :: i, j, k, l, m
    real(double), pointer :: f(:)=>Null()
    real(single) :: width, height
    real(double) :: xPos, yPos, minX, maxX, minY, maxY, dx, dy, radius, minRadius
    real(single) :: redchisq
    real(double) :: da(n)

    external func
    
    m = (iu - il + 1) * (ju - jl + 1)
  
    call AllocateLsData(m)

!write(*, *) "Deconvolve line 368 minval image=", minval(image), " max=", maxval(image)

    m = 0
    lsData%noStars = noStars
    width = real(iu - il)
    height = real(ju - jl) 
    minX = bl(2)
    maxX = bu(2)
    minY = bl(3)
    maxY = bu(3)

    do j = 2, noStars
        k = (j - 1) * 3 + 2 
        minX = min(minX, bl(k))
        maxX = max(maxX, bu(k))
        k = k + 1
        minY = min(minY, bl(k))
        maxY = max(maxY, bu(k))
    end do   
   
    minx = minx - real(24 / binx)
    maxx = maxx + real(24 / binx)
    miny = miny - real(24 / binx)
    maxy = maxy + real(24 / binx)


    do i = il, iu

        xPos = dble(i + 1)


        if (xPos .lt. minX .or. xPos .gt. maxX) cycle
        do j = jl , ju

            yPos = dble(j + 1)

            if (yPos .lt. minY .or. yPos .gt. maxY) cycle

            minRadius = 1000.0_double          
            do k = 1, noStars        
                l = (k - 1) * 3 + 2
                dx = xpos - x(l)
                dy = ypos - x(l + 1)
                radius = sqrt(dx * dx + dy * dy)
                minRadius = min(minRadius, radius)
            end do

            if (minRadius .le. real(18 / binx)) then     
                m = m + 1
                lsData%x(m) = xPos
    
                lsData%y(m) = yPos
                lsData%f(m) = dble(image(i, j))
                  

                lsData%weight(m) = 1.0_double / max(1.0_double, dble(image(i, j)))

 lsData%weight(m) = 1.0_double / dble(image(i, j))

!lsData%weight(m) = dble(image(i, j)) / maxval(image(il : iu, jl : ju))

!lsData%f(m) = MultiPSFFunction(bu, lsData%noStars,  lsData%x(m), lsData%y(m))

            end if
            
        end do
    end do



  
    allocate(f(1 : m))

!call fatal("stopped")

   
    toFit = .true.  ! Fit all parameters
!    toFit(1) = .false.    
    write(errstr, '(a)') "Input trial parameters:"
    call message(AppMsg, Verbose)

    do i = 1, n    
        if (tofit(i)) then
            write(errstr, '(a, i4, a, e12.4, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), " bounds-", bl(i), "-", bu(i)      
        else
            write(errstr, '(a, i4, a, e12.4, a)') "x(", i, ")=", x(i), " - Fixed parameter"      
  
        end if

        call message(AppMsg, Verbose)

    end do
  

    do j = 1, 3
        call nonlin(x, n, int(m, int32), func, bl, bu, toFit, f, status)
        if (status .gt. 0) exit
    end do

   
    write(errstr, '(a, e14.8)') "Sum of squares of residuals=", DOT_PRODUCT(f, f)
    call message(AppMsg, Verbose)
  
    write(errstr, '(a, i8)') "status=", status   
    call message(AppMsg, Verbose)

    call ChiSquared(n, x, int(m, int32), toFit, func, da, redChisq)

   
    call deallocatelsdata
    fitParameters%xcentre = 0.0
 
end subroutine Deconvolve


!**************************************************
! Deconvolve
!
!
!**************************************************
subroutine GaussDeconvolve1(noStars, image, &
           il, iu, jl, ju, n, x, &
           bl, bu, toFit, func, status, fitParameters, levelImage, sourceNumber)
    integer(int32), intent(in) :: noStars
    real(single), pointer :: image(:, :)
    integer(int32), intent(in), optional :: sourceNumber
   

    integer(int16), pointer, optional :: levelimage(:, :)

 
    integer(int16), intent(in) :: il, iu, jl, ju
    integer(int32), intent(in) :: n
    real(double), intent(inout) :: x(1 : n)
    real(double), intent(inout) :: bl(1 : n), bu(1 : n)
    logical(boolean)  :: toFit(:)
    integer(int16), intent(inout) :: status
    type(GaussFitParameters), intent(out) :: fitParameters

    integer(int32) :: i, j, m
    real(double), pointer :: f(:)=>Null()
    real(double) :: xPos, yPos
    real(double) :: error 
    
    real(double) :: da(1 : n)
    real(double) :: rmax, dx, dy  

 
    interface
       
    subroutine func(x, n, f, m, ctrl, c)
    use dal
        real(double), intent(in) :: x(:)
        integer, intent(in) :: n
        real(double), intent(out) :: f(:)
        integer, intent(in) :: m
        integer, intent(inout) :: ctrl
        real(double), pointer :: c(:, :)
    end subroutine func

    end interface


    call InitialiseGauss(10)
 
  
    m = (iu - il + 1) * (ju - jl + 1)
    allocate(lsData%x(1 : m)) 
    allocate(lsData%y(1 : m)) 
    allocate(lsData%f(1 : m)) 
    allocate(lsData%weight(1 : m)) 

    lsData%weight = 0.0_double   
    lsData%f = 0.0_double   

    m = 0
    lsData%noStars = noStars
   

    rmax = 0.0_double

    do i = il, iu
        xPos = dble(i) + 1.0_double
        
        do j = jl , ju
            if(present(levelImage)) then
                if (levelImage(i, j) .ne. sourceNumber) cycle
            end if 
            yPos = dble(j) + 1.0_double
                                
            m = m + 1
            lsData%x(m) = xPos
    
            lsData%y(m) = yPos
            dx = xpos - lsData%xcentre
            dy = ypos - lsData%ycentre
            rmax = max(rmax, sqrt(dx * dx + dy * dy))


         
            lsData%f(m) = dble(image(i, j))
 
            error =  lsData%f(m)
           
            
            if(error > 0.0_double) then
                lsData%weight(m) = 1.0_double / error
                lsData%weight(m) = min(1.0_double, lsData%weight(m))
  

       
            else
                lsData%weight(m) = 0.0_double

            end if

     
   
        end do
    end do


    if (m .lt. 10) then

        write(errstr, '(a, i2, a)') "Too few points for fits (", m, ") fit abandoned"
        call message(AppMsg, Verbose)
        return


    end if

    if (lsData%m .lt. 0.0) lsData%m = m
    if (lsData%rmax .lt. 0.0) lsData%rmax = rmax + 0.5_double
    allocate(f(1 : m))


    write(errstr, '(a)') "Input trial parameters:"
    call message(AppMsg, Verbose)
    do i = 1, n    

        write(errstr, '(a, i2, a, e12.4, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), " bounds-", bl(i), "-", bu(i)      
        call message(AppMsg, Verbose)

    end do


    do j = 1, 1

        call nonlin(x, n, m, func, bl, bu, toFit, f, status)
 
 

        write(errstr, '(a, i1, a, e14.8)') "Iteration ", j, " Sum of squares of residuals=", DOT_PRODUCT(f, f)
        call message(AppMsg, Verbose)
    end do



100 format(i3, a, 10e14.4)
    write(errstr, '(a, i7)') "status=", status   
    call message(AppMsg, Verbose)



    call ChiSquared(n, x, m, toFit, func, da, fitParameters%redChisq)



    deallocate(lsData%x) 
    deallocate(lsData%y) 
    deallocate(lsData%f) 
    deallocate(lsData%weight) 
    nullify(lsData%x) 
    nullify(lsData%y) 
    nullify(lsData%f) 
    nullify(lsData%weight) 

end subroutine GaussDeconvolve1


!**************************************************
! 
! GaussDeconvolve2
! 2-dimensional gaussian fit with no background
!
!**************************************************
subroutine GaussDeconvolve2(noStars, image, &
           il, iu, jl, ju, n, x, &
           bl, bu, toFit, func, status, fitParameters, &
           backgroundImage, levelImage, sourceNumber)
    integer(int32), intent(in) :: noStars
    real(single), pointer :: image(:, :)
    integer(int32), intent(in), optional :: sourceNumber
    real(single), pointer :: backgroundimage(:, :)

    integer(int16), pointer, optional :: levelimage(:, :)
   

    integer(int16), intent(in) :: il, iu, jl, ju
    integer(int32), intent(in) :: n
    real(double), intent(inout) :: x(1 : n)
    real(double), intent(inout) :: bl(1 : n), bu(1 : n)
    logical(boolean)  :: toFit(:)
    integer(int16), intent(inout) :: status
    type(GaussFitParameters), intent(out) :: fitParameters

    integer(int32) :: i, j, m
    real(double), pointer :: f(:)=>Null()
    real(double) :: xPos, yPos, error, dx, dy, rmax

    
    real(double) :: da(1 : n)
    real(double), pointer :: c(:, :) => null()
      interface
       
    subroutine func(x, n, f, m, ctrl, c)
    use dal
        real(double), intent(in) :: x(:)
        integer, intent(in) :: n
        real(double), intent(out) :: f(:)
        integer, intent(in) :: m
        integer, intent(inout) :: ctrl
        real(double), pointer :: c(:, :)
    end subroutine func

    end interface
  
    call InitialiseGauss(10)
 
  

    m = (iu - il + 1) * (ju - jl + 1)
    allocate(lsData%x(1 : m)) 
    allocate(lsData%y(1 : m)) 
    allocate(lsData%f(1 : m)) 
    allocate(lsData%weight(1 : m)) 
    lsData%weight = 0.0_double


 write(*, *) "GaussDeconvolve1 line 3371"    

    m = 0
    lsData%noStars = noStars
    rmax = 0.0_double    

    do i = il, iu
        xPos = dble(i) + 1.0_double
        
        do j = jl , ju
            if(present(levelImage)) then
                if (levelImage(i, j) .ne. sourceNumber) cycle
            end if 
            yPos = dble(j) + 1.0_double
                       
            m = m + 1
            lsData%x(m) = xPos
    
            lsData%y(m) = yPos

            dx = xpos - lsData%xcentre
            dy = ypos - lsData%ycentre
            rmax = max(rmax, sqrt(dx * dx + dy * dy))


            ! Subtract the background
            lsData%f(m) = dble(image(i, j) - backgroundImage(i, j))

!   lsData%f(m) = x(3) * GaussianIntegration2(lsData%x(m) - x(1), lsData%y(m) - x(2) , x(4), x(5), x(6))    


            ! The square of the error
            error = dble(image(i, j) + backgroundImage(i, j))
            
            if(error > 0.0_double) then
                lsData%weight(m) = 1.0_double / error

                lsData%weight(m) = min(1.0_double, lsData%weight(m))
       
            else
                m = m - 1
            end if
     
   
        end do
    end do

   if (m .lt. 10) then

        write(errstr, '(a, i2, a)') "Too few points for fits (", m, ") fit abandoned"
        call message(AppMsg, Verbose)
        return


    end if
!write(*, *)  "GaussDeconvolve2 m=", m


    c => allocate2dDouble(1, m, 1, n, "c")

     
 
    lsData%m = m

    if (lsdata%rmax .lt. 0.0) lsdata%rmax = rmax + 0.5_double

 
    allocate(f(1 : m))

    write(errstr, '(a)') "Input trial parameters:"
    call message(AppMsg, Verbose)
    do i = 1, n    

        write(errstr, '(a, i2, a, e12.4, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), " bounds-", bl(i), "-", bu(i)      
        call message(AppMsg, Verbose)

    end do


 
    do j = 1, 1

        f = 0.0_double
        call nonlin(x, n, m, func, bl, bu, toFit, f, status)
   

        write(errstr, '(a, i1, a, e14.8)') "Iteration ", j, " Sum of squares of residuals=", DOT_PRODUCT(f, f)
        call message(AppMsg, Verbose)
    end do


    call ChiSquared(n, x, m, toFit, func, da, fitParameters%redChisq)


    write(errstr, '(a, i7)') "status=", status   
    call message(AppMsg, Verbose)

    deallocate(f, c)
    nullify(f, c)
    deallocate(lsData%x) 
    deallocate(lsData%y) 
    deallocate(lsData%f) 
    deallocate(lsData%weight) 
    nullify(lsData%x) 
    nullify(lsData%y) 
    nullify(lsData%f) 
    nullify(lsData%weight) 

   
end subroutine GaussDeconvolve2
!***********************************************
!
!**********************************************
subroutine ChiSquared(n, x, ndata, toFit, func, dx, redChisq)

    integer, intent(in) :: n
    real(double) :: x(:)
    integer, intent(in) :: ndata
   
    logical(boolean) :: toFit(:)
   
    real(double) :: dx(:)
    real(single), intent(out) :: redChisq
    integer :: i, j, k, l, m
    real(double), pointer :: c(:, :) => null()
    real(double) :: chisq, dy, wt, prob
    real(double), pointer :: covarMatrix(:, :) => null()
    real(double), pointer :: f(:) => null()
    integer ctrl, nf, ndf
  
    interface       
        subroutine func(x, n, f, m, ctrl, c)
            use dal
            real(double), intent(in) :: x(:)
            integer, intent(in) :: n
            real(double), intent(out) :: f(:)
            integer, intent(in) :: m
            integer, intent(inout) :: ctrl
            real(double), pointer :: c(:, :)
        end subroutine func

    end interface

 
   ! Evalue the Jacobian and the function values

    ctrl = -5
    f => allocateDouble(1, nData, "f")
    f = 0.0_double
    c => allocate2dDouble(1, ndata, 1,  n, "c")
    call func(x, n, f, ndata, ctrl, c)



    nf = 0  ! Actual number of fitted parameters
    do i = 1, n
       if(toFit(i)) nf = nf + 1
    end do

    ! Compute chi-squared and the covariance matrix
    covarMatrix => allocate2dDouble(1, n, 1, n, "covarMatrix")
    covarMatrix = 0.0_double
    chisq = 0.0_double
    dx = 0.0

    do i = 1, ndata  ! Loop over all the observations

        dy = f(i) ! The residual between model and fit (weighting removed)


        j = 0

        do l = 1, n ! Loop over all the parameters
            if (toFit(l)) then

                wt = c(i, l) * lsData%weight(i)              
                k = 0 
                j = j + 1

                do m = 1, l
                    if (toFit(m)) then
                        k = k + 1
                         
                        covarMatrix(j, k) = covarMatrix(j, k) + wt * c(i, m)
                          
                    end if
                end do    
                
            end if
        end do    
        chisq = chisq + dy * dy * lsData%weight(i) 

    end do


    ! Fill in the symmetric side
    do j = 2, n
        do k = 1, j - 1
            covarMatrix(k, j) = covarMatrix(j, k)
        end do
    end do

   
    write(errstr, '(a)') "Solution"
    call message(AppMsg, Verbose)
   
    k  = 0
    do i = 1, n

        if (toFit(i)) then
            k = k + 1
            if (covarMatrix(k, k) .ge. 0.0) then
                dx(i) = sqrt(covarMatrix(k, k))
            else
                dx(i) = -1.0
            end if
 

            write(errstr, '(a, i2, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), "+/-", dx(i)
            call message(AppMsg, Verbose)
        else
            write(errstr, '(a, i2, a, e12.4, a)') "x(", i, ")=", x(i), " -fixed parameter"
            call message(AppMsg, Verbose)
        end if
    end do    



    ! Compute the probability
    ndf = ndata - nf
    prob = Gammaq(0.5_double * dble(ndf), 0.5_double * chisq) * 100.0_double
    redChisq =  chisq / real(ndf)


    write(errstr, '(a, e12.4, a, i4, a, i2)') "Reduced Chi-squared=", redChisq, &
        " No of degrees of freedom=", ndf, " no of fitted parameters=", nf
    call message(AppMsg, Verbose)


 
    deallocate(c, f, covarMatrix)
    nullify(c, f, covarMatrix)    



end subroutine ChiSquared
!*************************************************************
!
!
!*************************************************************
subroutine computeFitErrors(n, m, x, bl, bu, toFit, func, dx, k1)
    
    integer(int32), intent(in) :: n, m

    external func
    real(double) :: x(:), dx(:)
    integer, intent(in), optional :: k1
    real(double) :: mean
 
    integer(int32) :: i, j, k, index

 
    real(double) :: bl(1 : n), bu(1 : n)
    logical(boolean)  :: toFit(:)
    integer(int16) :: status
    real(double), pointer :: f(:)=>Null(), x1(:) => null()
    real(double), pointer :: params(:, :) => null()
    real(single) :: number, m1

    type (ImageData) :: tempLsData 

    allocate(templsData%x(1 : m)) 
    allocate(templsData%y(1 : m)) 
    allocate(templsData%f(1 : m)) 
    allocate(templsData%weight(1 : m)) 

    if (present(k1)) then 
        k = k1
    else
        k = min(10, m)
    end if 

    params => allocate2dDouble(1, n, 1, k, "params") 
    
    templsData = lsData
    f => allocateDouble(1, m, "f")
    x1 => allocateDouble(1, k, "x1")

    do j = 1, k  ! Each simulation
!        write(errstr, '(a, i4)') "Simulation ", j

!        call message(AppMsg, Verbose)
!write(*, *) " m=", m, " n=", n
!write(*, '(a, 7f10.2)') "x=", x
!write(*, '(a, 7f10.2)') "bl=", bl
!write(*, '(a, 7f10.2)') "bu=", bu
 

        ! Generate a new data set for each simulation
        m1 = real(m - 1) + 0.0001
        do i = 1, m
            call random_number(number)

            index = int(number * m1) + 1
    
            lsdata%x(i) = templsData%x(index) ! The new x data set
            lsdata%y(i) = templsData%y(index) ! The new y data set
            lsdata%f(i) = templsData%f(index) ! The new f data set
            lsdata%weight(i) = templsData%weight(index) ! The new weight data set

        end do

        ! Do the fit
       
        x1(1 : n) = x(1 : n)
        f = 0.0_double        
        call nonlin(x1, n, m, func, bl, bu, toFit, f, status)
        params(1 : n, j) = x1(1 : n)

!    do i = 1, n
!        write(errstr, '(a, i2, a, f10.4)') "x(", i, ")=", params(i, j)
!        call message(AppMsg, Verbose)
!    end do    
   
         
    end do
  

    ! Compute the errors
    do j = 1, n
        x1(1 : k) = params(j, 1 : k)


!        call ComputeStatistics(k, params(1 : k, j), mean, dx(j))

    call ComputeStatistics(k, x1, mean, dx(j))

!        do i = 1, k
!            write(*, *) i, "  x1=", x1(i), " params=", params(j, k)
!        end do
!        write(*, *) j, " mean=", mean, " k=",  k
        dx(j) = sqrt(dx(j))
    end do
 
    deallocate(f, params)
    nullify(f, params)


    nullify(templsData%x)
    nullify(templsData%y)
    nullify(templsData%f)
    write(errstr, '(a, i2)') "Monte carlo errors- No of simulations=", k
    call message(AppMsg, Verbose)
    

    do i = 1, n
        write(errstr, '(a, i2, a, e12.4, a, e12.4)') "x(", i, ")=", x(i), "+/-", dx(i)
        call message(AppMsg, Verbose)
    end do    
    
    
end subroutine computeFitErrors

!*******************************************
! Integrates 2-dimensional gaussian over
! one pixel (fits includes background but
! no position angle)
!******************************************
real(double) Function GaussianPixel1(x)
    real(double) :: x(:)
 
    GaussianPixel1 =  x(1) + x(4) * &
            GaussianIntegration(x(2), x(3), x(5))

end function GaussianPixel1

!*******************************************
! Integrates 2-dimensional gaussian over
! one pixel (fits includes background and
! position angle
!******************************************
real(double) Function GaussianPixel2(x)
    real(double) :: x(:)

!write(*, *) "Gaussian pixel 2 x=", x 
    GaussianPixel2 =  GaussianIntegration2(x(2), x(3), x(5), x(6), x(7))

!write(*, *) "GaussianPixel2=", GaussianPixel2
end function GaussianPixel2

!*******************************************
! Integrates 2-dimensional gaussian over
! one pixel (background assumed to be zero)
!******************************************
real(double) Function GaussianPixel3(x)
    real(double) :: x(:)
 
    GaussianPixel3 = GaussianIntegration2(x(1), x(2), x(4), x(5), x(6))

end function GaussianPixel3




!**************************************************************
! 
!*************************************************************
subroutine AllocateLsData(m)
    integer(int16), intent(in) :: m
    
    lsData%x => allocateDouble(1_int16, m, "x")
    lsData%y => allocateDouble(1_int16, m, "y")
    lsData%f => allocateDouble(1_int16, m, "f")   
    lsData%weight => allocateDouble(1_int16, m, "weight")
    lsData%fit => allocateDouble(1_int16, m, "fit")

 
end subroutine AllocateLsData
!**************************************************************
! 
!*************************************************************
subroutine DeAllocateLsData

    deallocate(lsData%x, lsData%y, lsData%f, lsData%weight, lsData%fit)
    nullify(lsData%x, lsData%y, lsData%f, lsData%weight, lsData%fit)



end subroutine DeAllocateLsData

 


end module stardecomp_mod





























